#ifndef INC_genTasks_h
#define INC_genTasks_h
/*	@(#)genTasks.h	1.2 9/20/91
 *	Author:	Roger A. Cole
 *	Date:	05-23-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00	05-23-90	rac	initial version
 * .01	06-19-91	rac	installed in SCCS
 * .02 	09-05-91	joh	updated for v5 vxWorks
 *				(taskLib.h now does not include sigLib.h)
 *
 * make options
 *	-DvxWorks	makes a version for VxWorks
 */
/*+/mod***********************************************************************
* TITLE	genTasks.h - definitions for using tasks with either SunOS or VxWorks
*
* DESCRIPTION
*	This file contains various definitions needed for setting
*	up and using tasks and semaphores.  Although the definitions provided
*	here aren't strictly system independent, they can be used in a
*	restricted mode so that system independence is achieved.  Presently
*	the definitions support tasks and semaphores under SunOS and VxWorks.
*	With each definition, "gotcha's" affecting SunOS usage are listed.
*
*	For SunOS, these tools are based on the Lightweight Process library,
*	referred to as the LWP library.  The term 'task' as used here is meant
*	to be synonymous with the LWP term 'thread'.
*
*       When using the LWP library, a special problem is the use of system
*       calls, including those for I/O.  SunOS provides a non-blocking I/O
*       library which addresses this problem.  Unfortunately, the non-blocking
*       I/O library doesn't set well with Channel Access, so another solution
*       must be used for programs which use Channel Access.
*
*       The tools provided automatically include 'errno' as part of the task
*       context, so that each task has its own copy.
*
*       This header file automatically includes the necessary LWP header files
*       and defines the necessary VxWorks symbols and typedefs.  In some
*       cases, symbols and typedefs are defined here which aren't available
*       with VxWorks--for portability, the user must supply those when
*       compiling for VxWorks.
*
*       To link when using the LWP with nonblocking I/O library, specify:
*               -llwp -lnbio -llwp              (yes, all three!!)
*	(Note that Channel Access doesn't operate properly when you use
*	the nonblocking I/O library.)
*
* BUGS
* o	In some cases, a source code may use <vxWorks.h> even though it is
*	compiling for SunOS.  Some conflicts in #defines will occur.
*
* SEE ALSO:
* o	genDefs.h	'master' include file
* o	genTaskSubr.c	code corresponding to some of the definitions here
*
*-***************************************************************************/

#ifndef INC_genDefs_h
#   include <genDefs.h>
#endif

#ifdef vxWorks
/*----------------------------------------------------------------------------
*    includes for VxWorks compile
*---------------------------------------------------------------------------*/
#    include <vxWorks.h>
#    include <stdioLib.h>
#    include <taskLib.h>
#    include <sigLib.h>
/* several symbols are used in the code for SunOS vs. VxWorks compatibility */
    typedef int TASK_ID;
/*#   define SIG_DFL (FUNCPTR)NULL*/
#   define SELF taskIdSelf()
#else
/*----------------------------------------------------------------------------
*    includes for Sun compile
*---------------------------------------------------------------------------*/
#   include <stdio.h>
#ifndef WIN32
#   include <lwp/lwp.h>
#   include <lwp/stackdep.h>
#endif WIN32
/*----------------------------------------------------------------------------
*    definitions for compatibility with VxWorks, to allow running under
*    UNIX.
*---------------------------------------------------------------------------*/
    typedef int STATUS;
    typedef void (*FUNCPTR)();
#endif

/*----------------------------------------------------------------------------
*       ---------------------------------------------------------------------
*                               SEMAPHORES
*       ---------------------------------------------------------------------
*       These tools provide functionality similar to the VxWorks semaphore
*       library.  See VxWorks Programmer's Guide, chapter on Architecture; and
*       VxWorks Reference Manual, documentation for semLib and individual
*       routines in semLib.
*
*       SEM_ID  defines a semaphore
*
*       VOID semInit(semId)     initialize a semaphore, to empty state
*       SEM_ID semId;
*               Initializes a semaphore to the empty state.  Since the
*               semaphore is empty, a semTake() blocks until a semGive() has
*               been done.  (GOTCHA: with LWP, the task which creates a
*               semaphore can do a semTake() even though a semGive() hasn't
*               been done.  This is detected by semTake(), so you'll find out
*               if you haven't coded in such a way as to avoid this
*               possibility.)
*
*       VOID semGive(semId)     fill a semaphore
*       SEM_ID semId;
*               Fills a semaphore, so that a task waiting for the semaphore
*               can proceed.
*
*       VOID semTake(semId)     empty a semaphore, waiting if not full
*       SEM_ID semId;
*               Empties a semaphore.  If the semaphore isn't full, the task
*               will wait until the semaphore is full.  Any task which takes
*               a semaphore must give it before dying.  (GOTCHA: with LWP,
*               a task which has taken a semaphore can take it again, even
*               though an intervening semGive() hasn't been done.  This is
*               detected by semTake(), so you'll find out if you haven't
*               coded in such a way as to avoid this possibility.)
*
*       STATUS semClear(semId)  empty a semaphore, proceeding if not full
*       SEM_ID semId;
*               Empties a semaphore if it is full.  If the semaphore isn't
*               full, ERROR is returned and the task is allowed to proceed.
*               If the semaphore was full, OK is returned.  (GOTCHA: with
*               LWP, semClear has the same problem as semTake(), with the
*               same resolution.)
*
*---------------------------------------------------------------------------*/
#ifndef vxWorks
#   define SEM_ID mon_t
#   define semInit(semId) (void) mon_create(&semId); (void) mon_enter(semId)
#   define semGive(semId) (void) mon_exit(semId)

    void semTake();
    STATUS semClear();
#endif


/*----------------------------------------------------------------------------
*       ---------------------------------------------------------------------
*                               TASK MANIPULATION
*       ---------------------------------------------------------------------
*       TASK_ID taskId;
*
*       TASK_ID taskSpawn(name, prio, options, stackSize, entryAddress,
*                       arg1, ... arg10)
*
*       STATUS taskSleep(taskId, seconds, usec);
*       STATUS taskSuspend(taskId);
*       STATUS taskResume(taskId);
*       STATUS taskDelete(taskId);
*	STATUS taskPrioritySet(taskId, newPriority);
*
*	FOR COMPATIBILITY:
*
*	o  To check if taskSpawn() failed, use GenTaskNull(taskId) on the
*	   resultant taskId; this macro is true if taskId is NULL.
*	o  Handle priorities special, since with SunOS LWP larger priorities
*	   are favored while VxWorks favors smaller priorities.  The macro
*	   GenLowerPriority(basePrio, delta) returns a priority which is less
*	   favored than 'basePrio' by 'delta' amount.
*       o  When you would use a task id of 0 with a VxWorks task call, use
*          SELF instead.
*	o  Instead of using taskDelay(), use taskSleep().  taskSleep() is
*	   available under both SunOS and VxWorks, and allows specifying
*	   delay time in seconds and micro-seconds, rather than in ticks,
*	   which gives some additional independence.
*
*
*---------------------------------------------------------------------------*/
STATUS taskSleep();

#ifndef vxWorks
    typedef thread_t TASK_ID;

    TASK_ID taskSpawn();
#   ifndef VX_STDIO
#       define VX_STDIO 1
#       define VX_FP_TASK 2
#       define VX_UNBREAKABLE 4
#       define VX_DEALLOC_STACK 8
#   endif
    STATUS taskPrioritySet();
    STATUS taskResume();
    STATUS taskSuspend();
    STATUS taskDelete();
#   define GenTaskNull(taskId) (SAMETHREAD(taskId, THREADNULL))
#   define GenLowerPriority(prio,delta) (prio-(delta))
#else
#   define GenTaskNull(taskId) (taskId <= 0)
#   define GenLowerPriority(prio,delta) (prio+(delta))
#endif

/*----------------------------------------------------------------------------
*       ---------------------------------------------------------------------
*                               LWP LIBRARY
*       ---------------------------------------------------------------------
*       This file provides some tools to help in using the LWP library.
*       For information on the LWP library specifically, see: SunOS System
*       Services Overview: Chapter 6--Lightweight Processes; and SunOs
*       Reference Manual, Section 3L: Lightweight Process Library.
*
*       lwpThreadStat(message, ptId)    print status of thread(s)
*       char *message;          message to print to identify status info
*       thread_t* ptId;         pointer to thread ID, or NULL for all threads
*               The status of all LWP threads or a specific LWP thread is
*               printed.  'message' is printed in conjunction with the status
*               to allow distinguishing if several calls to lwpThreadStat()
*               have been made.
*
*       lwpMonStat(message)
*       char *message;          message to print to identify status info
*               The status of all LWP monitors is printed.
*               'message' is printed in conjunction with the status
*               to allow distinguishing if several calls to lwpMonStat()
*               have been made.
*
*
*---------------------------------------------------------------------------*/
#ifndef vxWorks
    lwpThreadStat();
    lwpMonStat();
#endif

#endif	/* INC_genTasks_h */
