/************************************************************************/
/*									*/
/*			RESOURCE TABLE DEFINITION			*/
/*									*/
/*	This table contains the list of resources and the states of	*/
/*	their connections to their partners. The table is created	*/
/*	"on the fly" by the dbServer task. It is used by channel	*/
/*	access to determine if this IOC should respond to queries	*/
/*	for any given point. It is also used by the scan tasks to	*/
/*	determine if they should run a given chain of points.		*/
/*									*/
/************************************************************************/

/****** MOD LOG
01/08/93	GMU	410-581-0422	uchenick@tate.com
		Original entry
02/11/93	GMU
		Originally, there was not much stuff in this table that
		was used to keep track of the TCP link. As that has grown,
		code was stuck into dbServerTask (dbserver.c) to keep all
		of the channel data for each resource the same. At this
		point it is worthwhile considering having a structure per
		link and having each resource point to it. This can happen
		during a clean up operation because those data fields are
		from everyone but dbServer.
******/

/************************************************************************/
/*									*/
/*			TABLE STRUCTURE					*/
/*									*/
/*	All network type information in this table is to be kept	*/
/*	in the local machine's byte order. The reason for this is	*/
/*	that part of the connection algorithm requires that we		*/
/*	compare Internet addresses. If we kept that field in the	*/
/*	table in network byte order, this might give funny results	*/
/*	if we ever had a mix of little and big endian IOCs as 		*/
/*	redundant partners (shudder!). Now, because one piece of	*/
/*	network information is to be kept in local machine byte		*/
/*	order, then we should be consistent and keep all of it 		*/
/*	way.								*/
/*									*/
/************************************************************************/

/*
	First, a list of lock sets. Defined here to avoid forward reference
*/

#define	LOCKS_PER_LIST	63

typedef struct lockList
	{
	short		locksUsed;
	void		*lockSets[LOCKS_PER_LIST];
	struct lockList	*next;
	} LOCK_LIST;

typedef struct restab
	{
	char	resourceName[16];	/* Resource name, null terminated */
	short	driverPort;		/* port created by driver */
	short	localDbPort;		/* port created by dbServer */
	struct in_addr partnerInet;	/* Partner's Inet Address */
	short	partnerPort;		/* port created by partner's driver */
	short	partnerConnected;	/* BOOLEAN: TCP connection made? */
	int	partnerFd;		/* socket for driver <-> driver */
	short	dbServerPort;		/* port created by partner dbServer */
	short	dbServerConnected;	/* BOOLEAN: TCP connection made? */
	int	dbServerFd;		/* socket for dbServer <-> dbServer */
	short	private;		/* BOOLEAN: Resource nonredundant? */
	short	caAnswer;		/* BOOLEAN: Should ca respond? */
	int	bcastRate;		/* Task iterations between bcasts */
	int	bcastCounter;		/* Iterations since last bcast */
	long	bcastsMade;		/* Broadcast attempts made */
	int	redAlarmFd;		/* Redundant alarm update fd */
	unsigned long lastRxHeartbeat;	/* Time we last got a message */
	unsigned long lastTxHeartbeat;	/* Time we last sent a message */
	void	*driverData;		/* Device specific data */
	struct restab *link;		/* Next resource in the chain */
	LOCK_LIST	*lockSets;	/* List of lock sets involving this resource */
	unsigned long disconnectTime;	/* Ticks when TCP socket was created */
	unsigned short localState;	/* State of our driver */
	unsigned short partnerState;	/* Partner's declared state */
	unsigned short command;		/* Current command, 0 = empty */
	} RES_TABLE;

/************************************************************************/
/*									*/
/*		VALUES FOR THE DRIVER STATE FIELDS			*/
/*									*/
/************************************************************************/

#define	RES_MSTR	1		/* Resource is in master state */
#define RES_SLV		2		/* Resource is in slave state */
#define	RES_SLVWANT	3		/* Resource wants to be in slave state */
#define RES_SLVWANTCA	4		/* Wants to be slave, yield CA conn */
#define RES_ISOLATED	5		/* Not connected */
#define RES_CHANGED	0x8000		/* Status has been changed */

/************************************************************************/
/*									*/
/*		VALUES FOR THE COMMAND FIELD				*/
/*									*/
/************************************************************************/

#define RES_RELINQUISH	1		/* Relinquish the resource */
#define RES_CANCELREL	2		/* Cancel the relinquish command */
#define RES_TAKE	3		/* Take the resource */

/************************************************************************/
/*									*/
/*		WHERE TO FIND THE TABLE IN MEMORY			*/
/*									*/
/************************************************************************/

extern RES_TABLE *resTableHead;		/* Addr of the first entry */

/************************************************************************/
/*									*/
/*			RESOURCE TABLE UTILITY ROUTINES			*/
/*									*/
/*	All of these routines return integers. RES_OK means that	*/
/*	the call succeeded. The others are explained along with their	*/
/*	definitions.							*/
/*									*/
/************************************************************************/

/****** STATUS RETURNS *****/

#define RES_OK		0
#define	RES_NOTYET	-1	/* Resource table not built yet */
#define RES_NOTFOUND	-2	/* Couldn't find what you wanted */
#define	RES_BADADDR	-3	/* Caller passed a null pointer */
#define RES_NOCONN	-4	/* When fd was requested, connection is broken */
#define	RES_NOMORE	-5	/* When finding next entry, there are no more */
#define RES_LOCKERR	-6	/* Lock set over/under flow */

#define	MAXGROUPS	16

int resGetByName (char *name, RES_TABLE **resEntry);
/******
	This routine finds a resource table entry with a given name.

	Syntax:
		resGetByName(name, &resEntry)
		char		*name;
		RES_TABLE	**resEntry;

	Arguments:
		name is the resource name that is wanted, passed as a null
		terminated string. resEntry is a pointer to a resource
		table entry.
	Returns:
		RES_OK, RES_NOTYET, or RES_NOTFOUND
******/

int resDeclarePrivate (RES_TABLE *resEntry);
/******
	This routine declares that a resource table entry describes a
	non-redundant resource. This will keep dbServer from trying to
	find a partner for it. The routine will also set the caAnswer
	structure member.

	Syntax:
		resDeclarePrivate(resEntry)
		RES_TABLE	*resEntry;
	Arguments:
		resEntry is a pointer to the resource table entry to be
		declared non-redundant.
	Returns:
		RES_OK or RES_BADADDR
******/

int resRegisterDriver (RES_TABLE *resEntry, short tcpPort, int fd, int rate);
/******
	This routine is called by a driver to declare that it is willing
	to participate in the redundancy scheme. The routine posts the
	driver's TCP port number and the socket fd. The driver also gives
	the rate at which broadcasts should be made when looking for a
	redundant partner.

	Syntax:
		resRegisterDriver(resEntry, port, fd, rate)
		RES_TABLE	*resEntry;
		short		port;
		int		fd;
		float		rate;
	Arguments:
		resEntry is a pointer to the resource table entry for
		which the driver is registering. port is the TCP port
		number for the driver <-> driver link. fd is the file
		descriptor for that socket. rate is the amount of time,
		in seconds, between broadcasts. This will be rounded off
		to the scheduled execution rate of the dbServer task.
	Returns:
		RES_OK or RES_BADADDR
******/

int regGetBcastsMade (RES_TABLE *resEntry, long *attempts);
/******
	This routine is called to get the number of broadcast messages
	that have been transmitted in an attempt to find a partner for
	this particular resource. This information is useful for the
	driver task at startup time so he can determine if his redundant
	partner is present.

	Syntax:
		resGetBcastsMade(resEntry, &attempts)
		RES_TABLE	*resEntry;
		long		attempts;
	Arguments:
		resEntry is a pointer to the resource table entry that
		is of interest to the caller. The number of broadcast
		messages that have been transmitted is stored in attempts.
	Returns:
		RES_OK or RES_BADADDR
******/

int resDriverConnected (RES_TABLE *resEntry);
/******
	This routine is called by a driver task to declare that he has
	established a TCP connection with his redundant partner. This
	routine will not only set the partnerConnected member of the
	table entry addressed in the call, but it will also set that
	field in every table row that contains identical networking
	information (i.e., partner Internet address and TCP port number).

	Syntax:
		resDriverConnected(resEntry)
		RES_TABLE	*resEntry
	Arguments:
		resEntry points to the table row for which the TCP
		connection has been made.
	Returns:
		RES_OK or RES_BADADDR
*/
int resDriverBroken (RES_TABLE *resEntry);
/******
	This routine is called by a driver task to declare that the TCP
	link with his redundant partner has been terminated. This routine
	will clear the partnerConnected member of the table row passed,
	but will also clear it in every other table row that contains
	identical networking information (i.e., partner Internet address
	and TCP port number).

	Syntax:
		resDriverBroken(resEntry)
		RES_TABLE	*resEntry;
	Arguments:
		resEntry points to the table row for which the TCP
		connection was terminated.
	Returns:
		RES_OK or RES_BADADDR
******/

int resGetDriverFd (RES_TABLE *resEntry, int *fd);
/******
	This routine is called to get the fd of the TCP socket used
	for communications between drivers.

	Syntax:
		resGetDriverFd(resEntry, &fd)
		RES_TABLE	*resEntry;
		int		fd;
	Arguments:
		resEntry points to a specific table row. The fd of the
		TCP link between drivers is returned into fd.
	Returns:
		RES_OK, RES_BADADDR, or RES_NOCONN if the partnerConnected
		member of the structure is clear (i.e., the connection
		was never established).
******/

int resFindEquivDriver (RES_TABLE *resEntry, RES_TABLE **currPointer);
/******
	This routine is called to find the next resource table entry
	that has an equivalent driver <-> driver link. An equivalent
	link is one with the same Internet address and TCP port #.
	The user passes a pointer to a table entry, or zero to start
	at the head of the table. This routine will not return the
	starting pointer (it tries to find the NEXT table entry!),
	but it will return the search target. See the example.

	Syntax:
		resFindEquivDriver(resEntry, &currPointer)
		RES_TABLE	*resEntry;
		RES_TABLE	*currPointer;
	Arguments:
		resEntry is the table entry that contains the desired
		network information. currPointer points to the table
		row that has already been processed.
	Example:
		RES_TABLE	*resEntry;
		RES_TABLE	*currPointer = 0;
		resEntry = address of a table row
		while (resFindEquivDriver(resEntry, &currPointer) == RES_OK)
			{ do something using currPointer}
	Returns:
		RES_OK or RES_NOMORE when the address of all suitable
		table rows have been returned.
******/

int resFindEquivDbServer(RES_TABLE *resEntry, RES_TABLE **startEntry);
/******
	This routine is very similar to resFindEquivDriver with the
	exception that it finds a link with the same dbServer link
	instead of the same driver link.

	Syntax:
		resFindEquivDbServer(resEntry, &currPointer)
		RES_TABLE	*resEntry;
		RES_TABLE	*currPointer;
	Arguments:
		resEntry points to the table row that contains the
		desired network information. currPointer points to the
		table row that has already been processed.
	Returns:
		RES_OK or RES_NOMORE when the address of all suitalbe
		table rows have been returned.
******/

int resSetCaAnswer (RES_TABLE *resEntry);
/******
	This routine is called to set the caAnswer structure member for
	a specific table entry.

	Syntax:
		resSetCaAnswer(resEntry)
		RES_TABLE	*resEntry;
	Arguments:
		resEntry points to the table row where the caAnswer
		member is to be set.
	Returns:
		RES_OK or RES_BADADDR
******/

int resClearCaAnswer (RES_TABLE *resEntry);
/******
	This routine is called to clear the caAnswer structure member
	for a specific table entry.

	Syntax:
		resClearCaAnswer(resEntry)
		RES_TABLE	*resEntry;
	Arguments:
		resEntry points to the table row where the caAnswer
		member is to be cleared.
	Returns:
		RES_OK or RES_BADADDR
******/

int resGetCaAnswer (RES_TABLE *resEntry, short *answer);
/******
	This routine is called to get the caAnswer structure member
	for a specific table entry.

	Syntax:
		resGetCaAnswer(resEntry, &answer)
		RES_TABLE	*resEntry;
		short		answer;
	Arguments:
		resEntry points to a specific table row. The caAnswer
		structure member is written to answer.
	Returns:
		RES_OK or RES_BADADDR
******/

int resFindPartnerFd (int fd, RES_TABLE **found);
/******
	This routine is called to find a table entry with a given
	fd for the driver <-> driver TCP link. It would normally
	be used by a driver to process an I/O completion event on
	that link.

	Syntax:
		resFindPartnerFd(fd, &found)
		int		fd;
		RES_TABLE	*found;
	Arguments:
		Search for the first table entry where partnerFd
		is equal to fd. The address of the table row is
		returned in found.
	Returns:
		RES_OK, RES_NOTYET, or RES_NOTFOUND
******/

int resFindServerFd (int fd, RES_TABLE **found);
/******
	This routine is called to find a table entry with a given
	fd for the dbServer <-> dbServer TCP link. It would only
	be used by dbServer to process an I/O completion event on
	that link.

	Syntax:
		resFindServerFd(fd, &found)
		int		fd;
		RES_TABLE	*found;
	Arguments:
		Search for the first table entry where dbServerFd
		is equal to fd. The address of the table row is
		returned in found.
	Returns:
		RES_OK, RES_NOTYET, or RES_NOTFOUND
******/

int resPutDriverData (RES_TABLE *found, void *driverAddr);
/******
	This routine is called to store an address in the driverData
	structure member. A driver may want to use this facility to
	store the address of a structure that has the detailed status
	of this resource.

	Syntax:
		resPutDriverData(resEntry, driverData)
		RES_TABLE	*resEntry;
		void		*driverData;
	Arguments:
		resEntry points to the specific table row. The pointer
		to the driver's structure is stored in driverData.
	Returns:
		RES_OK or RES_BADADDR
******/

int resGetDriverData (RES_TABLE *found, void **callerAddr);
/******
	This routine is called to fetch the address that was stored
	in the driverData structure member.

	Syntax:
		resGetDriverData(resEntry, &driverData)
		RES_TABLE	*resEntry;
		void		*driverData;
	Arguments:
		resEntry points to the specific table row. The user's
		driverData address is returned in driverData.
	Returns:
		RES_OK or RES_BADADDR
******/

int resPutLocalState (RES_TABLE *found, unsigned short newState);
/******
	This routine is called to set the state of the localState field
	to one of its defined values.

	Syntax:
		resPutLocalState(resEntry, state)
		RES_TABLE	*resEntry;
		unsigned short	state;
	Arguments:
		resEntry points to the specific table row. The value to
		be set is contained in state.
	Returns:
		RES_OK or RES_BADADDR
******/

int resPutPartnerState (RES_TABLE *found, unsigned short newState);
/******
	This routine is called to set the state of the partnerState field
	to one of its defined values.

	Syntax:
		resPutPartnerState(resEntry, state)
		RES_TABLE	*resEntry;
		unsigned short	state;
	Arguments:
		resEntry points to the specific table row. The value to
		be set is contained in state.
	Returns:
		RES_OK or RES_BADDADR
******/

int resGetStates (RES_TABLE *found, unsigned short *local, unsigned short *partner);
/******
	This routine gets the localState field, the partnerState field,
	or both of them.

	Syntax:
		resGetStates(resEntry, &localState, &partnerState)
		RES_TABLE	*resEntry;
		unsigned short	localState
		unsigned short	partnerState
	Arguments:
		resEntry points to the specific table row. localState and
		partnerState are pointers to unsigned shorts where the
		appropriate values are to be stored. If either one of these
		pointers are zero, then the corresponding field will not be
		stored.
	Returns:
		RES_OK or RES_BADADDR. RES_BADADDR is not returned for a
		null pointer for either localState or partnerState.
******/

int resDump (void);
/******
	This routine prints the resource table on the default console.

	Syntax:
		resDump()
	Arguments:
		none
	Returns:
		Zero, always.
******/


int resTakeAll (void);
/******
	This routine traverses the entire resource table, setting the
	private and caAnswer members. This is provided for systems
	that have drivers that don't understand redundancy.

	Syntax:
		resTakeAll()
	Arguments:
		none
	Returns:
		Zero, always.
******/

int resPutCommand (RES_TABLE *found, unsigned short newCommand);
/******
	This routine puts the given command into the resource table.

	Syntax:
		resPutCommand(resEntry, command)
		RES_TABLE	*resEntry;
		unsigned short	command;
	Arguments:
		resEntry points to the specific entry in the resource
		table. The command to be inserted is in command.
	Returns:
		RES_OK or RES_BADADDR
******/

int resGetCommand (RES_TABLE *found, unsigned short *commandOut);
/******
	This routine gets the command from a specific resource table
	entry.

	Syntax:
		resGetCommand(resEntry, &command)
		RES_TABLE	*resEntry;
		unsigned short	command;
	Arguments:
		resEntry points to a specific table row. The current
		command field is written into command.
	Returns:
		RES_OK or RES_BADADDR
******/

int resAppendLock (RES_TABLE *resEntry, void *lockSet);
int resLockInit (void);
void resFreeChains (void);

