/****
*
*	File name :		meter.c
*	Author :		Steve Johnson
*	Creation date :	September 10, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		meter_init
*		meter_connect
*		meter_poll
*		meter_write
*		meter_read
*		meter_get_id
*		meter_read_table
*		meter_shutdown
*		meter_exit
*
*	Siemens Meter Protocol implementation
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sstimer.h>
#include <sslevel1.h>
#include <ssaudit.h>
#include "meter.h"

#define METER_GROUP			"meter_main"

#define METER_STX			0x02
#define METER_EOT			0x04
#define METER_ACK			0x06
#define METER_BLOCK_ID		0x0C
#define METER_NACK			0x15
#define METER_ETB			0x17
#define METER_ESC			0x1B
#define METER_FILLER		0x45
#define METER_X_IDLE		0x07
#define METER_X_WRITE_PW	0x0C
#define METER_X_NEW_TIME	0x0D
#define METER_X_GET_ID		0x0E
#define METER_X_HANGUP		0x10
#define METER_X_READ_PW		0x13
#define METER_X_COMMAND		'X'
#define METER_Y_COMMAND		'Y'
#define METER_Z_COMMAND		'Z'
#define GOOD_STATUS			0
#define REQUEST_SIZE		15
#define RESPONSE_SIZE		16
#define BLOCK_SIZE			267
#define BLOCK_ACK_SIZE		5
#define RESET_TIMEOUT		500
#define COMMAND_TIMEOUT		1200
#define TABLE_TIMEOUT		3000
#define DELAY_TIMEOUT		200
#define USER_TIMEOUT		6000
#define HANGUP_TIMEOUT		1200
#define TABLE_3_TIMEOUT		10000
#define OFFLINE_TIMEOUT		60000
#define MAX_ATTEMPTS		3
#define MAX_DIFF_COUNT		2
#define MAX_SKIP_COUNT		10
#define DEFAULT_PASSWORD	"0000"
#define PASSWORD_SIZE		4
#define UNIT_ID_UNKNOWN		(-1)

typedef enum {
	STATE_IDLE					= 1,
	STATE_OFFLINE				= 2,
	STATE_NEED_RESET			= 3,
	STATE_NEED_ID				= 4,
	STATE_NEED_HANGUP			= 5,
	STATE_W4_ID					= 6,
	STATE_W4_TABLE				= 7,
	STATE_W4_BLOCK				= 8,
	STATE_W4_RESET				= 9,
	STATE_W4_HANGUP				= 10,
	STATE_W4_ESCAPE				= 11,
	STATE_W4_USER_X_RESPONSE	= 12,
	STATE_W4_USER_Y_RESPONSE	= 13,
	STATE_W4_USER_Z_RESPONSE	= 14,
	STATE_W4_USER_BLOCK			= 15,
	STATE_W4_USER_ACK			= 16,
	STATE_W4_MV90_COMMAND		= 17,
	STATE_W4_MV90_BLOCK			= 18,
	STATE_W4_MV90_ACK			= 19,
} METER_STATE;

typedef struct {
	QUEUE_DEF		link;
	UCHAR			table_number;
	long			last_update;
	int				needs_update;
	int				completed;
	int				read_only;
	int				diff_count;
	int				skip_count;
	size_t			last_block_size;
	size_t			table_size;
	size_t			byte_count;
	USHORT			block_count;
	USHORT			block_index;
	USHORT			blocks_required;
	UCHAR			*data_locked;
	UCHAR			*data_current;
	UCHAR			*data_temp;
} TABLE_DEF;

typedef struct {
	QUEUE_DEF		link;
	QUEUE_DEF		root_table;
	LEVEL1_HANDLE	lower_handle;
	TABLE_DEF		*current_table;
	int				ready_for_mv90;
	int				needs_snapshot;
	int				locked;
	long			unit_id;
	long			hangup_timeout;
	long			command_timeout;
	long			reset_timeout;
	long			table_timeout;
	long			delay_timeout;
	long			user_timeout;
	long			table_3_timeout;
	long			offline_timeout;
	int				max_diff_count;
	int				max_skip_count;
	long			timeout;
	long			user_last_command;
	long			start_tick;
	long			read_delay_tick;
	long			last_table_3;
	long			last_cache_time;
	USHORT			port;
	int				passthru;
	int				max_retries;
	int				retry_count;
	int				data_size;
	int				interval_time;
	int				channel_count;
	int				max_storage_channels;
	int				max_stor_k;
	int				mem_size;
	int				new_read_password;
	int				new_write_password;
	int				new_time_stamp;
	size_t			interval_count;
	size_t			interval_size;
	METER_STATE		state;
	size_t			user_block;
	size_t			user_index;
	size_t			user_table_index;
	UCHAR			user_last_size;
	TABLE_DEF		*user_table;
	UCHAR			time_stamp [6];
	size_t			read_index;
	size_t			read_buffer_size;
	size_t			request_size;
	size_t			response_size;
	size_t			block_ack_size;
	size_t			block_data_size;
	size_t			send_buffer_size;
	size_t			user_read_size;
	size_t			user_send_size;
	UCHAR			*read_buffer;
	UCHAR			*send_buffer;
	UCHAR			request [REQUEST_SIZE];
	UCHAR			response [RESPONSE_SIZE];
	UCHAR			block_ack [BLOCK_ACK_SIZE];
	UCHAR			block_data [BLOCK_SIZE];
	UCHAR			user_read_data [BLOCK_SIZE];
	UCHAR			user_send_data [BLOCK_SIZE];
	UCHAR			sequence;
	char			write_password [PASSWORD_SIZE];
	char			read_password [2];
	char			saved_password [PASSWORD_SIZE];
	char			password_dir [MAX_STRING];
	char			name [MAX_STRING];
} HANDLE_DEF;

static QUEUE_DEF	root_handle;
static long			reset_timeout;
static long			command_timeout;
static long			table_timeout;
static long			delay_timeout;
static long			user_timeout;
static long			hangup_timeout;
static long			table_3_timeout;
static long			offline_timeout;
static int			max_attempts;
static int			max_diff_count;
static int			max_skip_count;
static int			console_on;
static char			password_dir [MAX_STRING];

static HANDLE_DEF	*add_handle (char *name);
static void			delete_handle (HANDLE_DEF *);
static TABLE_DEF	*find_table_by_number (HANDLE_DEF *, int);
static USHORT		get_crc (size_t, UCHAR *);
static void			build_ack_request (HANDLE_DEF *, USHORT);
static void			build_escape_request (HANDLE_DEF *, USHORT);
static void			build_id_response (HANDLE_DEF *);
static void			build_hangup_response (HANDLE_DEF *);
static void			build_y_response (HANDLE_DEF *);
static void			build_ack_response (HANDLE_DEF *);
static void			build_nack_response (HANDLE_DEF *);
static int			table_add (HANDLE_DEF *, UCHAR, int);
static void			check_data (HANDLE_DEF *);
static void			check_timers (HANDLE_DEF *);
static void			check_tables (HANDLE_DEF *);
static int			check_response (HANDLE_DEF *);
static int			check_block (HANDLE_DEF *);
static void			read_finished (HANDLE_DEF *);
static void			user_finished (HANDLE_DEF *);
static void			header_finished (HANDLE_DEF *);
static void			block_finished (HANDLE_DEF *);
static void			parse_table_0 (HANDLE_DEF *, TABLE_DEF *);
static void			parse_table_11 (HANDLE_DEF *, TABLE_DEF *);
static void			parse_table_3 (HANDLE_DEF *, TABLE_DEF *);
static int			check_status (HANDLE_DEF *);
static void			state_timeout (HANDLE_DEF *);
static void			set_state_idle (HANDLE_DEF *);
static void			set_state_offline (HANDLE_DEF *);
static void			send_reset (HANDLE_DEF *);
static void			send_id_request (HANDLE_DEF *);
static void			send_y_request (HANDLE_DEF *, TABLE_DEF *);
static void			send_hangup (HANDLE_DEF *);
static METER_STATUS	user_control (HANDLE_DEF *);
static METER_STATUS	send_user_buffer (HANDLE_DEF *);
static int			send_command (HANDLE_DEF *, size_t, UCHAR *);
static int			get_blocks_required (HANDLE_DEF *, UCHAR *, UCHAR *);
static void			get_date_time (UCHAR *, TIME_DEF *);
static long			get_bcd (UCHAR);
static void			put_bcd (int, UCHAR *);
static void			insert_table_12 (HANDLE_DEF *, TABLE_DEF *);
static void			resend_last (HANDLE_DEF *);
static int			check_password (HANDLE_DEF *, int, int);
static void			refresh_table (HANDLE_DEF *, int);
static int			update_table_1 (TABLE_DEF *);
static int			all_tables_current (HANDLE_DEF *);
static void			read_passwords (HANDLE_DEF *);
static void			save_passwords (HANDLE_DEF *);
static void 		create_file (char *);
static void			close_handle (HANDLE_DEF *);
static void			lock_tables (HANDLE_DEF *);
static void			log_console (char *, int, char *, int);

METER_STATUS meter_init (void)
{
	TRACE ("meter_init, entry point");
	queue_init (&root_handle);
	console_on = ini_get_int (METER_GROUP, "CONSOLE_ON", FALSE);
	reset_timeout = ini_get_long (METER_GROUP
							    , "RESET_TIMEOUT"
							    , RESET_TIMEOUT);
	command_timeout = ini_get_long (METER_GROUP
								  , "COMMAND_TIMEOUT"
								  , COMMAND_TIMEOUT);
	hangup_timeout = ini_get_long (METER_GROUP
								 , "HANGUP_TIMEOUT"
								 , HANGUP_TIMEOUT);
	table_timeout = ini_get_long (METER_GROUP
							    , "TABLE_TIMEOUT"
							    , TABLE_TIMEOUT);
	delay_timeout = ini_get_long (METER_GROUP
							    , "DELAY_TIMEOUT"
							    , DELAY_TIMEOUT);
	user_timeout = ini_get_long (METER_GROUP
							   , "USER_TIMEOUT"
							   , USER_TIMEOUT);
	table_3_timeout = ini_get_long (METER_GROUP
								  , "TABLE_3_TIMEOUT"
								  , TABLE_3_TIMEOUT);
	offline_timeout = ini_get_long (METER_GROUP
								  , "OFFLINE_TIMEOUT"
								  , OFFLINE_TIMEOUT);
	max_diff_count = ini_get_int (METER_GROUP
								, "MAX_DIFF_COUNT"
								, MAX_DIFF_COUNT);
	max_skip_count = ini_get_int (METER_GROUP
								, "MAX_SKIP_COUNT"
								, MAX_SKIP_COUNT);
	if (!ini_get_str (METER_GROUP, "PASSWORD_DIR", password_dir)) {
		strcpy (password_dir, "/ata0/tis4000d/");
	}
	max_attempts = ini_get_int (METER_GROUP, "MAX_ATTEMPTS", MAX_ATTEMPTS);
	return METER_NORMAL;
}

METER_STATUS meter_connect (METER_HANDLE *user_handle, char *name)
{
	int				status;
	HANDLE_DEF		*handle;
	char			device_name [MAX_STRING];
	char			password [MAX_STRING];

	TRACE ("meter_connect, entry point");
	TRACE_S ("meter_connect, meter_name = (%s)", name);
	handle = add_handle (name);
	if (!handle) {
		ERROR_TRACE ("meter_connect, add_handle failed");
		*user_handle = (METER_HANDLE) NULL;
		return METER_NO_MEMORY;
	}
	handle -> passthru = ini_get_int (name, "PASSTHRU", FALSE);
	handle -> reset_timeout = ini_get_long (name
										  , "RESET_TIMEOUT"
										  , reset_timeout);
	handle -> command_timeout = ini_get_long (name
										    , "COMMAND_TIMEOUT"
										    , command_timeout);
	handle -> hangup_timeout = ini_get_long (name
										   , "HANGUP_TIMEOUT"
										   , hangup_timeout);
	handle -> table_timeout = ini_get_long (name
										  , "TABLE_TIMEOUT"
										  , table_timeout);
	handle -> delay_timeout = ini_get_long (name
										 , "DELAY_TIMEOUT"
										 , delay_timeout);
	handle -> user_timeout = ini_get_long (name
									     , "USER_TIMEOUT"
									     , user_timeout);
	handle -> table_3_timeout = ini_get_long (name
											, "TABLE_3_TIMEOUT"
											, table_3_timeout);
	handle -> offline_timeout = ini_get_long (name
										    , "OFFLINE_TIMEOUT"
										    , offline_timeout);
	handle -> max_diff_count = ini_get_int (name
										   , "MAX_DIFF_COUNT"
										   , max_diff_count);
	handle -> max_skip_count = ini_get_int (name
										   , "MAX_SKIP_COUNT"
										   , max_skip_count);
	if (!ini_get_str (name, "PASSWORD_DIR", handle -> password_dir)) {
		strcpy (handle -> password_dir, password_dir);
	}
	handle -> max_retries = ini_get_int (name, "MAX_ATTEMPTS", max_attempts);

	handle -> port = ini_get_int (name, "TCP_PORT", 0);
	if (ini_get_str (name, "PASSWORD", password)) {
		memcpy (handle -> read_password, password, 2);
		memcpy (handle -> write_password, password, 4);
	} else {
		memcpy (handle -> read_password, DEFAULT_PASSWORD, 2);
		memcpy (handle -> write_password, DEFAULT_PASSWORD, 4);
	}
	read_passwords (handle);
	if (!ini_get_str (name, "DEVICE_NAME", device_name)) {
		strcpy (device_name, name);
	}
	handle -> lower_handle = NULL;
	status = level1_connect (&handle -> lower_handle, device_name);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("meter_connect, level1_connect = %d", status);
		log_console ("Connect to meter port failed", handle -> port, "", 0);
		handle -> lower_handle = (LEVEL1_HANDLE) NULL;
		set_state_offline (handle);
	} else {
		send_reset (handle);
	}
	table_add (handle, 0, TRUE);
	*user_handle = (METER_HANDLE *) handle;
	return METER_NORMAL;
}

METER_STATUS meter_poll (METER_HANDLE *return_handle)
{
	HANDLE_DEF		*handle;
	
	handle = (HANDLE_DEF *) queue_next (&root_handle);
	while (handle) {
		if (handle -> user_read_size) {
			TRACE_I ("meter_poll, user_read_size = %d", handle -> user_read_size);
			*return_handle = handle;
			return METER_NORMAL;
		}
		if (handle -> user_table) {
			if ((handle -> last_cache_time + handle -> user_timeout) < (long) GetTickCount ()) {
				ERROR_TRACE_I ("meter_poll, cache expired port = %d", handle -> port);
				handle -> user_table = NULL;
			}
		}
		switch (handle -> state) {
		case STATE_IDLE:
			if (handle -> user_send_size) {
				send_user_buffer (handle);
			} else {
				check_tables (handle);
			}
			break;
		case STATE_NEED_HANGUP:
			send_hangup (handle);
			break;
		case STATE_NEED_RESET:
			send_reset (handle);
			break;
		case STATE_W4_ID:
		case STATE_W4_HANGUP:
		case STATE_W4_TABLE:
		case STATE_W4_BLOCK:
		case STATE_W4_USER_X_RESPONSE:
		case STATE_W4_USER_Z_RESPONSE:
		case STATE_W4_USER_Y_RESPONSE:
		case STATE_W4_USER_BLOCK:
		case STATE_W4_USER_ACK:
			check_data (handle);
			break;
		case STATE_NEED_ID:
		case STATE_W4_RESET:
		case STATE_W4_ESCAPE:
		case STATE_W4_MV90_COMMAND:
		case STATE_W4_MV90_BLOCK:
		case STATE_W4_MV90_ACK:
		case STATE_OFFLINE:
			check_timers (handle);
			break;
		default:
			ERROR_TRACE_I ("meter_poll, unhandled state = %d", handle -> state);
			break;
		}
		handle = (HANDLE_DEF *) queue_next (&handle -> link);
	}
	return METER_NO_DATA;
}

METER_STATUS meter_write (METER_HANDLE	user_handle
						, size_t		buffer_size
						, void			*buffer)
{
	TABLE_DEF	*table;
	HANDLE_DEF	*handle;
	UCHAR		*request;

	TRACE ("meter_write, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_write, NULL handle");
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	if (buffer_size > sizeof (handle -> user_send_data)) {
		ERROR_TRACE_I ("meter_write, buffer_size = %d", buffer_size);
		ERROR_TRACE_I ("meter_write, sizeof (user_send_data) = %d"
					 , sizeof (handle -> user_send_data));
		ERROR_TRACE_B ("meter_write", buffer_size, buffer);
		return METER_OVERFLOW;
	}
	handle -> user_send_size = buffer_size;
	memcpy (handle -> user_send_data, buffer, buffer_size);
	request = (UCHAR *) handle -> user_send_data;
	handle -> new_read_password = FALSE;
	handle -> new_read_password = FALSE;
	handle -> new_time_stamp = FALSE;
	if (!handle -> ready_for_mv90) {
		TRACE_I ("meter_write, tables not cached command = 0x%X", request [1]);
		log_console ("Tables not Cached", handle -> port, "Command", request [1]);
		handle -> user_send_size = 0;
		return METER_NORMAL;
	}
	switch (request [1]) {
	case METER_Y_COMMAND:
		log_console ("Y_Command\n", handle -> port, "Table:", request [3]);
		TRACE ("meter_write, METER_Y_COMMAND");
		TRACE_I ("meter_write, Y table_number = %d", request [3]);
		if (!check_password (handle, METER_Y_COMMAND, 0)) {
			TRACE ("meter_write, password failed");
			handle -> user_send_size = 0;
			return METER_NORMAL;
		}
		handle -> user_table = (TABLE_DEF *) NULL;
		if (handle -> passthru) {
			return user_control (handle);
		}
		if (request [3] == 1) {
			TRACE ("meter_write, request for table 1 goes through");
			return user_control (handle);
		}
		build_y_response (handle);
		handle -> user_send_size = 0;
		return METER_NORMAL;
	case METER_X_COMMAND:
		log_console ("X_Command", handle -> port, "Command:", request [4]);
		handle -> user_table = (TABLE_DEF *) NULL;
		if (!check_password (handle, METER_X_COMMAND, request [4])) {
			TRACE ("meter_write, password failed");
			handle -> user_send_size = 0;
			return METER_NORMAL;
		}
		switch (request [4]) {
		case METER_X_WRITE_PW:
			TRACE ("meter_write, write password X_command");
			memcpy (handle -> saved_password
				  , &request [5]
				  , PASSWORD_SIZE);
			audit_password ("Quad4 Meter Write Password Changed");
			log_console ("X_Command - Write Password Changed", handle -> port, "", 0);
			handle -> new_write_password = TRUE;
			break;
		case METER_X_READ_PW:
			TRACE ("meter_write, read password X_command");
			memcpy (handle -> saved_password
				  , &request [5]
				  , 2);
			audit_password ("Quad4 Meter Read Password Changed");
			log_console ("X_Command - Read Password Changed", handle -> port, "", 0);
			handle -> new_read_password= TRUE;
			break;
		case METER_X_IDLE:
			TRACE ("meter_write, idle command");
			audit_password ("Quad4 Set to Idle - Table 12 and 15 dumped");
			log_console ("X_Command - Set to Idle", handle -> port, "", 0);
			table = find_table_by_number (handle, 12);
			if (!table) {
				ERROR_TRACE ("build_y_response, no table 12");
				handle -> user_send_size = 0;
				return METER_NORMAL;
			}
			table -> completed = FALSE;
			table -> needs_update = TRUE;
			table = find_table_by_number (handle, 15);
			if (!table) {
				ERROR_TRACE ("build_y_response, no table 15");
				handle -> user_send_size = 0;
				return METER_NORMAL;
			}
			table -> completed = FALSE;
			table -> needs_update = TRUE;
			break;
		case METER_X_GET_ID:
			TRACE ("meter_write, get_id X_COMMAND");
			if (handle -> passthru) {
				return user_control (handle);
			}
			build_id_response (handle);
			handle -> user_send_size = 0;
			return METER_NORMAL;
		case METER_X_HANGUP:
			TRACE ("meter_write, hangup X_COMMAND");
			if ((handle -> passthru) || (handle -> state == STATE_W4_MV90_COMMAND)) {
				return user_control (handle);
			}
			build_hangup_response (handle);
			handle -> user_send_size = 0;
			return METER_NORMAL;
		case METER_X_NEW_TIME:
			TRACE ("meter_write, new time stamp");
			handle -> new_time_stamp = TRUE;
			break;
		default:
			TRACE_I ("meter_write, METER_X_COMMAND 0x%X", request [4]);
			TRACE_I ("meter_write, port = %d", handle -> port);
			TRACE_B ("meter_write", buffer_size, buffer);
		}
		TRACE_I ("X_Command from mv90 command 0x%X\n", request [4]);
		audit_write ("Quad4 X command from MV90");
		return user_control (handle);
	case METER_Z_COMMAND:
		log_console ("Z_Command", handle -> port, "Table:", request [3]);
		TRACE_I ("meter_write, Z_Command table = %d\n", request [3]);
		TRACE_I ("meter_write, port = %d", handle -> port);
		TRACE_B ("meter_write", buffer_size, buffer);
		handle -> user_table = (TABLE_DEF *) NULL;
		audit_write ("Quad4 Z command - table changed");
		if (!check_password (handle, METER_Z_COMMAND, 0)) {
			TRACE ("meter_write, password failed");
			handle -> user_send_size = 0;
			return METER_NORMAL;
		}
		TRACE ("meter_write, sending Z_COMMAND");
		refresh_table (handle, request [3]);
		return user_control (handle);
	case METER_ACK:
		TRACE ("meter_write, METER_ACK");
		TRACE_I ("meter_write, ACK block_number = %d", request [2]);
		if ((handle -> passthru) || (handle -> state == STATE_W4_MV90_ACK)) {
			return user_control (handle);
		}
		build_ack_response (handle);
		handle -> user_send_size = 0;
		return METER_NORMAL;
	case METER_NACK:
		TRACE ("meter_write, METER_NACK");
		if ((handle -> passthru) || (handle -> state == STATE_W4_MV90_ACK)) {
			return user_control (handle);
		}
		build_nack_response (handle);
		handle -> user_send_size = 0;
		return METER_NORMAL;
	case METER_BLOCK_ID:
		TRACE ("meter_write, METER_BLOCK_ID");
		return user_control (handle);
	case METER_ESC:
		TRACE ("meter_write, METER_ESC");
		handle -> user_table = (TABLE_DEF *) NULL;
		if ((handle -> passthru) || (handle -> state == STATE_W4_MV90_ACK)) {
			return user_control (handle);
		} else {
			TRACE ("meter_write, escape ignored");
		}
		handle -> user_send_size = 0;
		return METER_NORMAL;
	}
	TRACE_I ("meter_write, user_send_data [1] = 0x%X", request [1]);
	handle -> user_table = (TABLE_DEF *) NULL;
	handle -> user_send_size = 0;
	return METER_NOT_IMPLEMENTED;
}

METER_STATUS meter_read (METER_HANDLE	user_handle
						, size_t		buffer_size
						, void			*buffer
						, size_t		*return_size)
{
	HANDLE_DEF	*handle;

	TRACE ("meter_read, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_read, NULL handle");
		*return_size = 0;
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	if (handle -> user_read_size > buffer_size) {
		ERROR_TRACE_I ("meter_read, user_read_size = %d", handle -> user_read_size);
		ERROR_TRACE_I ("meter_read, buffer_size = %d", buffer_size);
		handle -> user_read_size = 0;
		*return_size = 0;
		return METER_OVERFLOW;
	}
	memcpy (buffer, handle -> user_read_data, handle -> user_read_size);
	*return_size = handle -> user_read_size;
	handle -> user_read_size = 0;
	TRACE_B ("meter_read", *return_size, buffer);
	return METER_NORMAL;
}

METER_STATUS meter_lock (METER_HANDLE user_handle)
{
	HANDLE_DEF		*handle;

	TRACE ("meter_lock, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_lock, NULL handle");
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	if (handle -> locked) {
		ERROR_TRACE ("meter_lock, already locked");
		return METER_LOCK_ERROR;
	}
	if (all_tables_current (handle)) {
		lock_tables (handle);
	}
	handle -> locked = TRUE;
	return METER_NORMAL;
}

METER_STATUS meter_unlock (METER_HANDLE user_handle)
{
	HANDLE_DEF	*handle;

	TRACE ("meter_unlock, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_unlock, NULL handle");
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	if (!handle -> locked) {
		ERROR_TRACE ("meter_unlock, already unlocked");
		return METER_LOCK_ERROR;
	}
	handle -> locked = FALSE;
	return METER_NORMAL;
}

UCHAR *meter_table_addr (USHORT user_port, USHORT table_number)
{
	HANDLE_DEF	*handle;
	TABLE_DEF	*table;

	handle = (HANDLE_DEF *) queue_next (&root_handle);
	while (handle) {
		TRACE_I ("meter_table_addr, port = %d", handle -> port);
		if (handle -> port == user_port) {
			table = find_table_by_number (handle, table_number);
			if (table) {
				if (table -> completed) {
					return (UCHAR *) table -> data_current;
				} else {
					return (UCHAR *) NULL;
				}
			}
		}
		handle = (HANDLE_DEF *) queue_next (&handle -> link);
	}
	return (UCHAR *) NULL;
}

void meter_debug_on (void)
{
	console_on = TRUE;
}

void meter_debug_off (void)
{
	console_on = FALSE;
}


METER_STATUS meter_read_table (METER_HANDLE		user_handle
							 , int				table_number
							 , size_t			buffer_size
							 , void				*buffer
							 , size_t			*return_size)
{
	HANDLE_DEF	*handle;
	TABLE_DEF	*table;

	TRACE ("meter_read_table, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_read_table, NULL handle");
		*return_size = 0;
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	table = find_table_by_number (handle, table_number);
	if (!table) {
		ERROR_TRACE_I ("meter_read_table, no table = %d", table_number);
		ERROR_TRACE_S ("meter_read_table, handle = %s", handle -> name);
		*return_size = 0;
		return METER_BAD_HANDLE;
	}
	if (table -> table_size > buffer_size) {
		ERROR_TRACE ("meter_read_table, buffer overflow");
		ERROR_TRACE_I ("meter_read_table, table_size = %d", table -> table_size);
		ERROR_TRACE_I ("meter_read_table, buffer_size = %d", buffer_size);
		*return_size = 0;
		return METER_OVERFLOW;
	}
	if (!table -> table_size) {
		ERROR_TRACE ("meter_read_table, table empty");
		*return_size = 0;
		return METER_TABLE_EMPTY;
	}
	TRACE_I ("meter_read_table, table_size = %d", table -> table_size);
	memcpy (buffer, table -> data_current, table -> table_size);
	*return_size = table -> table_size;
	return METER_NORMAL;
}


METER_STATUS meter_get_id (METER_HANDLE user_handle, long *meter_id)
{
	HANDLE_DEF	*handle;

	TRACE ("meter_get_id, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_get_id, NULL handle");
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	*meter_id = handle -> unit_id;
	TRACE_L ("meter_get_id, returning id = 0x%X", *meter_id);
	return METER_NORMAL;
}

METER_STATUS meter_shutdown (METER_HANDLE user_handle)
{
	HANDLE_DEF 		*handle;

	TRACE ("meter_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("meter_shutdown, NULL handle");
		return METER_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	close_handle (handle);
	delete_handle (handle);
	return METER_NORMAL;
}

METER_STATUS meter_exit (void)
{
	HANDLE_DEF	*handle;

	TRACE ("meter_exit, entry point");
	while (queue_next (&root_handle)) {
		TRACE ("meter_exit, shutdown next handle");
		handle = (HANDLE_DEF *) queue_next (&root_handle);
		meter_shutdown ((METER_HANDLE *) handle);
	}
	return METER_NORMAL;
}

/****
*
****/

static HANDLE_DEF *add_handle (char *name)
{
	HANDLE_DEF	*handle;

	TRACE_S ("add_handle, meter name = %s", name);
	handle = (HANDLE_DEF *) malloc (sizeof (HANDLE_DEF));
	if (!handle) {
		ERROR_TRACE_I ("add_handle, malloc size = %d", sizeof (HANDLE_DEF));
		return NULL;
	}
	strcpy (handle -> name, name);
	handle -> read_buffer = (UCHAR *) NULL;
	handle -> ready_for_mv90 = FALSE;
	handle -> needs_snapshot = FALSE;
	handle -> locked = FALSE;
	handle -> read_index = 0;
	handle -> sequence = 1;
	handle -> interval_time = 0;
	handle -> interval_count = 0;
	handle -> interval_size = 0;
	handle -> user_last_command = 0;
	handle -> last_cache_time = 0;
	handle -> user_table = NULL;
	handle -> channel_count = 0;
	handle -> max_storage_channels = 0;
	handle -> max_stor_k = 0;
	handle -> last_table_3 = 0;
	handle -> unit_id = UNIT_ID_UNKNOWN;
	handle -> data_size = 0;
	handle -> new_read_password = FALSE;
	handle -> new_write_password = FALSE;
	handle -> new_time_stamp = FALSE;
	handle -> user_send_size = 0;
	handle -> user_read_size = 0;
	handle -> current_table = (TABLE_DEF *) NULL;
	queue_init (&handle -> root_table);
	queue_append (&root_handle, &handle -> link);
	return handle;
}

/****
*
****/

static void delete_handle (HANDLE_DEF *handle)
{
	TABLE_DEF		*table;

	TRACE ("delete_handle (meter), entry point");
	while (queue_next (&handle -> root_table)) {
		table = (TABLE_DEF *) queue_get (&handle -> root_table);
		if (table -> data_locked) {
			free (table -> data_locked);
		}
		if (table -> data_current) {
			free (table -> data_current);
		}
		if (table -> data_temp) {
			free (table -> data_temp);
		}
		free (table);
	}
	queue_remove (&root_handle, &handle -> link);
	free (handle);
}

static int check_password (HANDLE_DEF *handle, int command, int request)
{
	TRACE_B ("check_password, write", PASSWORD_SIZE, handle -> write_password);
	TRACE_B ("check_password, read", 2, handle -> read_password);
	TRACE_B ("check_password, user", PASSWORD_SIZE, &handle -> user_send_data [9]);
	switch (command) {
	case METER_Z_COMMAND:
		if (!memcmp (handle -> write_password
				  , &handle -> user_send_data [9]
				  , PASSWORD_SIZE)) {
			return TRUE;
		}
		audit_bad_password ("Quad4 Meter - Bad Table Write Password Attempt");
		if (console_on) {
			printf ("Quad4: (%s) Bad Table Write Password", handle -> name);
		}
		TRACE ("check_password, Z/X command failed");
		TRACE_B ("check_password, write table", PASSWORD_SIZE, handle -> write_password);
		TRACE_B ("check_password, user", PASSWORD_SIZE, &handle -> user_send_data [9]);
		return FALSE;
	case METER_X_COMMAND:
		switch (request) {
		case METER_X_HANGUP:
		case METER_X_GET_ID:
			if (!memcmp (handle -> read_password
					   , &handle -> user_send_data [9]
					   , 2)) {
				return TRUE;
			}
			if (!memcmp (handle -> write_password
					   , &handle -> user_send_data [9]
					   , 2)) {
				return TRUE;
			}			
			break;
		default:
			if (!memcmp (handle -> write_password
					  , &handle -> user_send_data [9]
					  , PASSWORD_SIZE)) {
				return TRUE;
			}
		}
		audit_bad_password ("Quad4 Meter - Bad Write Password Attempt");
		if (console_on) {
			printf ("Quad4: (%s) Bad Write Password", handle -> name);
		}
		TRACE ("check_password, Z/X command failed");
		TRACE_B ("check_password, write", PASSWORD_SIZE, handle -> write_password);
		TRACE_B ("check_password, user", PASSWORD_SIZE, &handle -> user_send_data [9]);
		return FALSE;
	case METER_Y_COMMAND:
		if (!memcmp (handle -> read_password
				   , &handle -> user_send_data [9]
				   , 2)) {
			return TRUE;
		}
		if (!memcmp (handle -> write_password
				   , &handle -> user_send_data [9]
				   , 2)) {
			return TRUE;
		}
	  	audit_bad_password ("Quad4 Meter - Bad Read Password Attempt");
		if (console_on) {
			printf ("Quad4: (%s) Bad Read Password", handle -> name);
		}
		TRACE ("check_password, Y command failed");
		TRACE_B ("check_password, write", PASSWORD_SIZE, handle -> write_password);
		TRACE_B ("check_password, read", 2, handle -> read_password);
		TRACE_B ("check_password, user", PASSWORD_SIZE, &handle -> user_send_data [9]);
		return FALSE;
	}
	ERROR_TRACE ("check_password, unknown path");
	return FALSE;
}

static int table_add (HANDLE_DEF *handle, UCHAR table_number, int read_only)
{
	TABLE_DEF	*table;

	TRACE_I ("table_add, entry point table = %d", table_number);
	table = (TABLE_DEF *) malloc (sizeof (TABLE_DEF));
	if (!table) {
		ERROR_TRACE_I ("table_add, malloc = %d", sizeof (TABLE_DEF));
		return METER_NO_MEMORY;
	}
	table -> data_locked = (UCHAR *) NULL;
	table -> data_current = (UCHAR *) NULL;
	table -> data_temp = (UCHAR *) NULL;
	table -> table_number = table_number;
	table -> completed = FALSE;
	table -> diff_count = 0;
	table -> skip_count = 0;
	table -> read_only = read_only;
	table -> needs_update = TRUE;
	queue_append (&handle -> root_table, &table -> link);
	return METER_NORMAL;
}

static void build_hangup_response (HANDLE_DEF *handle)
{
	UCHAR		sequence;
	USHORT		crc;
	UCHAR		*request;

	TRACE ("build_hangup_response, return from cache");
	request = handle -> user_send_data;
	if (request [0] != METER_STX) {
		ERROR_TRACE_I ("build_hangup_response, METER_STX = 0x%X", request [0]);
		return;
	}
	if (request [1] != METER_X_COMMAND) {
		ERROR_TRACE_I ("build_hangup_response, METER_X_COMM = 0x%X", request [1]);
		return;
	}
	if (request [4] != METER_X_HANGUP) {
		ERROR_TRACE_I ("build_hangup_response, command = 0x%X", request [4]);
		return;
	}
	sequence = request [2];
	handle -> user_read_data [0] = METER_STX;
	handle -> user_read_data [1] = METER_ACK;
	handle -> user_read_data [2] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [3] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [4] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [5] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [6] = 0; /* status */
	handle -> user_read_data [7] = sequence;
	handle -> user_read_data [8] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [9] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [10] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [11] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [12] = 0;
	handle -> user_read_data [13] = 0;
	crc = get_crc (13, &handle -> user_read_data [1]);
	handle -> user_read_data [14] = (UCHAR) (crc >> 8);
	handle -> user_read_data [15] = (UCHAR) crc;
	handle -> user_read_size = 16;
	log_console ("Hangup cached", handle -> port, "UnitID:", handle -> unit_id);
}

static void build_id_response (HANDLE_DEF *handle)
{
	UCHAR		sequence;
	USHORT		crc;
	UCHAR		*request;

	TRACE ("build_id_response, return from cache");
	request = handle -> user_send_data;
	if (request [0] != METER_STX) {
		ERROR_TRACE_I ("build_id_response, METER_STX = 0x%X", request [0]);
		return;
	}
	if (request [1] != METER_X_COMMAND) {
		ERROR_TRACE_I ("build_id_response, METER_X_COMM = 0x%X", request [1]);
		return;
	}
	if (request [4] != METER_X_GET_ID) {
		ERROR_TRACE_I ("build_id_response, command = 0x%X", request [4]);
		return;
	}
	sequence = request [2];
	handle -> user_read_data [0] = METER_STX;
	handle -> user_read_data [1] = METER_ACK;
	handle -> user_read_data [2] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [3] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [4] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [5] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [6] = 0; /* status */
	handle -> user_read_data [7] = sequence;
	handle -> user_read_data [8] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [9] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [10] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [11] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [12] = 0;
	handle -> user_read_data [13] = 0;
	crc = get_crc (13, &handle -> user_read_data [1]);
	handle -> user_read_data [14] = (UCHAR) (crc >> 8);
	handle -> user_read_data [15] = (UCHAR) crc;
	handle -> user_read_size = 16;
	log_console ("Get ID cached", handle -> port, "UnitID:", handle -> unit_id);
}

static void build_y_response (HANDLE_DEF *handle)
{
	USHORT		crc;
	UCHAR		sequence;
	size_t		max_table_index;
	UCHAR		table_number;
	TABLE_DEF	*table;
	UCHAR		*request;

	TRACE ("build_y_response, return from cache");
	request = handle -> user_send_data;
	if (request [0] != METER_STX) {
		ERROR_TRACE_I ("build_y_response, METER_STX = 0x%X", request [0]);
		return;
	}
	if (request [1] != METER_Y_COMMAND) {
		ERROR_TRACE_I ("build_y_response, request [1] = 0x%X", request [1]);
		ERROR_TRACE_I ("build_y_response, METER_Y_COMMAND = 0x%X", METER_Y_COMMAND);
		return;
	}
	sequence = request [2];
	table_number = request [3];
	table = find_table_by_number (handle, table_number);
	if (!table) {
		ERROR_TRACE_I ("build_y_response, no table = %d", table_number);
		return;
	}
	if (!table -> completed) {
		ERROR_TRACE_I ("build_y_response, table %d not loaded", table_number);
		return;
	}
	handle -> user_table_index = 0;
	handle -> user_table = table;
	handle -> last_cache_time = GetTickCount ();
	handle -> user_index = (request [4] << 8) | request [5];
	handle -> user_block = (request [6] << 8) | request [7];
	handle -> user_last_size = request [8];
	TRACE_I ("build_y_response, sequence = %d", sequence);
	TRACE_I ("build_y_response, table_number = %d", table_number);
	TRACE_I ("build_y_response, user_index = %d", handle -> user_index);
	TRACE_I ("build_y_response, user_block = %d", handle -> user_block);
	TRACE_I ("build_y_response, user_last_size = %d", handle -> user_last_size);
	if (handle -> user_block > table -> block_count) {
		handle -> user_block = request [7];
		ERROR_TRACE_I ("build_y_response, overflow blocks = %d", handle -> user_block);
		log_console ("MV90 table overflow adjusting", handle -> port, "Table", table_number);
	}
	if (handle -> user_block > table -> block_count) {
		handle -> user_block = table -> block_count;
		log_console ("MV90 table overflow", handle -> port, "Table:", table_number);
		ERROR_TRACE_I ("build_y_response, overflow blocks = %d", handle -> user_block);
	}
	if (!handle -> user_block) {
		handle -> user_block = table -> block_count;
		TRACE_I ("build_y_response, override zero blocks = %d", handle -> user_block);
	}
	max_table_index = 256 * (table -> block_count - handle -> user_block);
	if (max_table_index + 256 > table -> table_size) {
		ERROR_TRACE ("build_y_repsponse, table size exceeded");
		ERROR_TRACE_I ("build_y_repsponse, size = %d", max_table_index);
		ERROR_TRACE_I ("build_y_repsponse, block_count = %d", table -> block_count);
		ERROR_TRACE_I ("build_y_repsponse, block_count = %d", handle -> user_block);
		ERROR_TRACE_I ("build_y_repsponse, table size = %d", table -> table_size);
		return;
	}
	if (!handle -> user_last_size) {
		handle -> user_last_size = handle -> user_table -> last_block_size;
	}
	TRACE_I ("build_y_repsponse, index = %d", handle -> user_table_index);
	handle -> user_read_data [0] = METER_STX;
	handle -> user_read_data [1] = METER_ACK;
	handle -> user_read_data [2] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [3] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [4] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [5] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [6] = 0; /* status */
	handle -> user_read_data [7] = sequence;
	handle -> user_read_data [8] = table_number;
	handle -> user_read_data [9] = (UCHAR) (handle -> user_index >> 8);
	handle -> user_read_data [10] = (UCHAR) (handle -> user_index & 0xFF);
	handle -> user_read_data [11] = (UCHAR) (handle -> user_block >> 8);
	handle -> user_read_data [12] = (UCHAR) (handle -> user_block & 0xFF);
	handle -> user_read_data [13] = handle -> user_last_size;
	crc = get_crc (13, &handle -> user_read_data [1]);
	handle -> user_read_data [14] = (UCHAR) (crc >> 8);
	handle -> user_read_data [15] = (UCHAR) crc;
	handle -> user_read_size = 16;
	log_console ("Table header from cache", handle -> port, "Table:", table_number);
}

static void build_ack_response (HANDLE_DEF *handle)
{
	USHORT		crc;
	UCHAR		block_number;
	TABLE_DEF	*table;
	UCHAR		*request;

	TRACE ("build_ack_response, return from cache");
	request = handle -> user_send_data;
	if (request [0] != METER_STX) {
		ERROR_TRACE_I ("build_ack_response, METER_STX = 0x%X", request [0]);
		return;
	}
	if (request [1] != METER_ACK) {
		ERROR_TRACE_I ("build_ack_response, METER_ACK = 0x%X", request [1]);
		return;
	}
	table = handle -> user_table;
	if (!table) {
		ERROR_TRACE_I ("build_ack_response, no active table port = %d", handle -> port);
		return;
	}
	TRACE_I ("build_ack_response, table_number = %d", table -> table_number);
	block_number = request [2];
	if (block_number != ((UCHAR) (handle -> user_block & 0xFF))) {
		ERROR_TRACE ("build_ack_response, mv90 expecting wrong block");
		ERROR_TRACE_I ("build_ack_response, user_block = 0x%X", handle -> user_block);
		ERROR_TRACE_I ("build_ack_response, block_number = 0x%X", block_number);
		return;
	}
	TRACE_I ("build_ack_response, block_number = %d", block_number);
	TRACE_I ("build_ack_response, user_block_number = %d", handle -> user_block);
	if (!handle -> user_block) {
		TRACE ("build_ack_response, acknowlege last block");
		if (table -> table_number == 12) {
			refresh_table (handle, 1);
		}
		handle -> user_table = (TABLE_DEF *) NULL;
		return;
	}
	if (handle -> user_table_index + 256 > table -> table_size) {
		ERROR_TRACE ("build_ack_repsponse, table size exceeded");
		ERROR_TRACE_I ("build_ack_repsponse, size = %d", handle -> user_table_index);
		ERROR_TRACE_I ("build_ack_repsponse, table size = %d", table -> table_size);
		return;
	}
	--handle -> user_block;
	handle -> user_read_data [0] = METER_STX;
	handle -> user_read_data [1] = METER_BLOCK_ID;
	handle -> user_read_data [2] = (UCHAR) (handle -> unit_id >> 24);
	handle -> user_read_data [3] = (UCHAR) (handle -> unit_id >> 16);
	handle -> user_read_data [4] = (UCHAR) (handle -> unit_id >>  8);
	handle -> user_read_data [5] = (UCHAR) (handle -> unit_id & 0xFF);
	handle -> user_read_data [6] = (UCHAR) (handle -> user_block >> 8);
	handle -> user_read_data [7] = (UCHAR) (handle -> user_block & 0xFF);
	TRACE_I ("build_ack_repsponse, index = %d", handle -> user_table_index);
	if (table -> table_number == 1) {
		if (!update_table_1 (table)) {
			ERROR_TRACE ("build_ack_response, table 1 crossover");
			return;
		}
	}
	if (handle -> user_index) {
		if (handle -> user_index + handle -> user_last_size > 256) {
			ERROR_TRACE ("build_ack_response, overflow");
			ERROR_TRACE_I ("build_ack_response, user_index", handle -> user_index);
			ERROR_TRACE_I ("build_ack_response, last_size", handle -> user_last_size);
			return;
		}
		TRACE ("build_ack_response, partial read");
		TRACE_I ("build_ack_response, user_index", handle -> user_index);
		TRACE_I ("build_ack_response, last_size", handle -> user_last_size);
		memcpy (&handle -> user_read_data [8]
			  , &table -> data_locked [handle -> user_index]
			  , handle -> user_last_size);
		memset (&handle -> user_read_data [handle -> user_index + handle -> user_last_size]
			   , 0x45
			   , 256 - (handle -> user_index + handle -> user_last_size));
	} else {
		memcpy (&handle -> user_read_data [8]
			  , &table -> data_locked [handle -> user_table_index]
			  , 256);
	}
	handle -> user_table_index += 256;
	crc = get_crc (263, &handle -> user_read_data [1]);
	handle -> user_read_data [264] = (UCHAR) (crc >> 8);
	handle -> user_read_data [265] = (UCHAR) crc;
	if (!handle -> user_block) {
		handle -> user_read_data [266] = METER_EOT;
	} else {
		handle -> user_read_data [266] = METER_ETB;
	}
	handle -> last_cache_time = GetTickCount ();
	handle -> user_read_size = 267;
	log_console ("Table Block from cache", handle -> port, "Table:", table -> table_number);
}

static void build_nack_response (HANDLE_DEF *handle)
{
	int			block_number;
	UCHAR		*request;

	ERROR_TRACE ("build_nack_response, return from cache");
	request = handle -> user_send_data;
	if (request [0] != METER_STX) {
		ERROR_TRACE_I ("build_nack_response, METER_STX = 0x%X", request [0]);
		return;
	}
	if (request [1] != METER_NACK) {
		ERROR_TRACE_I ("build_nack_response, METER_NACK = 0x%X", request [1]);
		return;
	}
	block_number = handle -> request [2];
	TRACE_I ("build_nack_response, block_number = %d", block_number);
	if (!block_number) {
		TRACE ("build_nack_response, acknowlege last block");
		return;
	}
	++handle -> user_block;
	build_ack_response (handle);
}

static void build_ack_request (HANDLE_DEF *handle, USHORT block_number)
{
	USHORT		crc;

	handle -> block_ack [0] = METER_STX;
	handle -> block_ack [1] = METER_ACK;
	handle -> block_ack [2] = (UCHAR) block_number;
	crc = get_crc (2, &handle -> block_ack [1]);
	handle -> block_ack [3] = (UCHAR) (crc >> 8);
	handle -> block_ack [4] = (UCHAR) crc;
	handle -> block_ack_size = 5;
}

static void build_escape_request (HANDLE_DEF *handle, USHORT block_number)
{
	USHORT		crc;

	handle -> block_ack [0] = METER_STX;
	handle -> block_ack [1] = METER_ESC;
	handle -> block_ack [2] = (UCHAR) block_number;
	crc = get_crc (2, &handle -> block_ack [1]);
	handle -> block_ack [3] = (UCHAR) (crc >> 8);
	handle -> block_ack [4] = (UCHAR) crc;
	handle -> block_ack_size = 5;
}

static int update_table_1 (TABLE_DEF *table)
{
	long			tick_diff;
	long			diff_seconds;
	DATE_TIME_DEF	date_time;
	TIME_DEF		time_only;
	UCHAR			last_time [6];
	UCHAR			new_time [6];

	ERROR_TRACE ("update_table_1, not finished");
	tick_diff = GetTickCount () - table -> last_update;
	if (tick_diff > (30L * 60000L)) {
		ERROR_TRACE ("update_table_1, way out of date");
		return FALSE;
	}
	table -> last_update = GetTickCount ();
	memcpy (last_time, table -> data_current, 6);
	ERROR_TRACE_L ("update_table_1, tick_diff = %ld", tick_diff);
	ERROR_TRACE_B ("update_table_1, last", 6, last_time);
	diff_seconds = tick_diff / 1000L;
	ERROR_TRACE_I ("udpate_table_1, diff_seconds = %d", diff_seconds);
	date_time . second = get_bcd (last_time [5]);
	date_time . minute = get_bcd (last_time [4]);
	date_time . hour = get_bcd (last_time [3]);
	date_time . year = get_bcd (last_time [2]);
	date_time . day = get_bcd (last_time [1]);
	date_time . month= get_bcd (last_time [0]);
	if (99 == date_time . year) {
		date_time . year = 1999;
	} else {
		date_time . year = 2000 + date_time . year;
	}
	time_date_to_time (&date_time, &time_only);
	TRACE_L ("update_table_1, before tv_sec = %ld", time_only . tv_sec);
	time_only . tv_sec += diff_seconds;
	TRACE_L ("update_table_1, after tv_sec = %ld", time_only . tv_sec);
	time_time_to_date (&time_only, &date_time);
	put_bcd (date_time . second, &new_time [5]);
	put_bcd (date_time . minute, &new_time [4]);
	put_bcd (date_time . hour, &new_time [3]);
	put_bcd (date_time . year % 100, &new_time [2]);
	put_bcd (date_time . day, &new_time [1]);
	put_bcd (date_time . month, &new_time [0]);
	ERROR_TRACE_B ("update_table_1, new", 6, new_time);
	memcpy (table -> data_current, new_time, 6);
	return TRUE;
}

static METER_STATUS user_control (HANDLE_DEF *handle)
{
	TRACE ("user_control, entry point");
	switch (handle -> state) {
	case STATE_IDLE:
	case STATE_W4_USER_X_RESPONSE:
	case STATE_W4_USER_Y_RESPONSE:
	case STATE_W4_USER_Z_RESPONSE:
	case STATE_W4_USER_BLOCK:
	case STATE_W4_USER_ACK:
	case STATE_W4_MV90_COMMAND:
	case STATE_W4_MV90_BLOCK:
	case STATE_W4_MV90_ACK:
		if (!send_user_buffer (handle)) {
			ERROR_TRACE ("user_control, send_user_buffer failed");
			handle -> user_send_size = 0;
			return METER_IO_ERROR;
		}
		return METER_NORMAL;
	case STATE_W4_HANGUP:
	case STATE_W4_TABLE:
	case STATE_W4_BLOCK:
	case STATE_W4_ID:
		ERROR_TRACE_I ("user_control, waiting state = %d", handle -> state);
		return METER_NORMAL;
	default:
		ERROR_TRACE_I ("user_control, interrupt state = %d", handle -> state);
		if (handle -> current_table) {
			send_reset (handle);
		} else {
			send_hangup (handle);
		}
		return METER_NORMAL;
	}
}

static void check_data (HANDLE_DEF *handle)
{
	int			status;
	size_t		read_count;

	if (!handle -> lower_handle) {
		check_timers (handle);
		return;
	}
	status = level1_poll (handle -> lower_handle, &read_count);
	if (LEVEL1_NO_DATA == status) {
		check_timers (handle);
		return;
	} else if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("check_data, level1_poll = %d", status);
		set_state_offline (handle);
		status = level1_shutdown (handle -> lower_handle);
		if (LEVEL1_NORMAL != status) {
			ERROR_TRACE_I ("close_handle, level1_shutdown = %d", status);
		}
		handle -> lower_handle = (LEVEL1_HANDLE) NULL;
		return;
	}
	status = level1_read (handle -> lower_handle
						, handle -> read_buffer_size - handle -> read_index
						, &handle -> read_buffer [handle -> read_index]
						, &read_count);
	if ((LEVEL1_NO_DATA != status) && (LEVEL1_NORMAL != status)) {
		ERROR_TRACE_I ("check_data, level1_read = %d", status);
		ERROR_TRACE_I ("check_data, cisco timeout = %d", handle -> state);
		ERROR_TRACE_I ("check_data, timeout = %d", handle -> user_timeout);
		set_state_offline (handle);
		status = level1_shutdown (handle -> lower_handle);
		if (LEVEL1_NORMAL != status) {
			ERROR_TRACE_I ("close_handle, level1_shutdown = %d", status);
		}
		handle -> lower_handle = (LEVEL1_HANDLE) NULL;
		return;
	}
	handle -> read_index += read_count;
	if (handle -> read_index >= handle -> read_buffer_size) {
		TRACE_B ("check_data, full"
				, handle -> read_index
				, handle -> read_buffer);
		read_finished (handle);
	}
}

static void check_timers (HANDLE_DEF *handle)
{
	switch (handle -> state) {
	case STATE_IDLE:
		return;
	case STATE_W4_ID:
	case STATE_OFFLINE:
	case STATE_NEED_ID:
	case STATE_W4_RESET:
	case STATE_W4_HANGUP:
	case STATE_W4_ESCAPE:
	case STATE_W4_TABLE:
	case STATE_W4_BLOCK:
	case STATE_W4_USER_X_RESPONSE:
	case STATE_W4_USER_Y_RESPONSE:
	case STATE_W4_USER_Z_RESPONSE:
	case STATE_W4_USER_BLOCK:
	case STATE_W4_USER_ACK:
	case STATE_W4_MV90_BLOCK:
	case STATE_W4_MV90_ACK:
		if ((handle -> start_tick + handle -> timeout) < (long) GetTickCount ()) {
			TRACE_I ("check_timers, state = %d", handle -> state);
			state_timeout (handle);
		}
		break;
	case STATE_W4_MV90_COMMAND:
		if ((handle -> user_last_command + handle -> user_timeout)
		< (long) GetTickCount ()) {
			TRACE ("check_timers, user timed out");
			send_hangup (handle);
		}
		break;
	default:
		break;
	}
}

static void check_tables (HANDLE_DEF *handle)
{
	TABLE_DEF		*table;

	if ((handle -> read_delay_tick + handle -> delay_timeout) < (long) GetTickCount ()) {
		table = (TABLE_DEF *) queue_next (&handle -> root_table);
		while (table) {
			if (table -> needs_update) {
				TRACE ("check_tables, update the table");
				switch (table -> table_number) {
				case 12:
				case 3:
				case 1:
					TRACE_I ("check_tables, update table %d", table -> table_number);
					send_y_request (handle, table);
					return;
				}
				if (table -> skip_count >= handle -> max_skip_count) {
					TRACE ("check_tables, reset the skip count");
					table -> skip_count = 0;
					send_y_request (handle, table);
					return;
				} else if (table -> diff_count >= handle -> max_diff_count) {
					TRACE ("check_tables, skip the table");
					++table -> skip_count;
					table -> needs_update = FALSE;
				} else {
					send_y_request (handle, table);
					return;
				}
			}
			table = (TABLE_DEF *) queue_next (&table -> link);
		}
		if ((handle -> last_table_3 + handle -> table_3_timeout) < (long) GetTickCount ()) {
			TRACE ("check_tables, table3 expired");
			refresh_table (handle, 3);
			send_hangup (handle);
			return;
		}
 		close_handle (handle);
		handle -> state = STATE_IDLE;
		handle -> read_buffer_size = 0;
		handle -> ready_for_mv90 = TRUE;
		if (!handle -> locked) {
			if (handle -> needs_snapshot) {
				handle -> needs_snapshot = FALSE;
				lock_tables (handle);
			}
		}
	}
}

static void state_timeout (HANDLE_DEF *handle)
{
	TRACE_B ("state_timeout", handle -> read_index, handle -> read_buffer);
	TRACE_L ("state_timeout, timedout after %ld mS", handle -> timeout);
	TRACE_L ("state_timeout, start = %ld", handle -> start_tick);
	TRACE_L ("state_timeout, now = %ld", GetTickCount ());
	TRACE_I ("state_timeout, state = %d", handle -> state);
	handle -> read_index = 0;
	switch (handle -> state) {
	case STATE_W4_ID:
	case STATE_W4_HANGUP:
	case STATE_W4_TABLE:
	case STATE_W4_BLOCK:
	case STATE_W4_USER_X_RESPONSE:
	case STATE_W4_USER_Y_RESPONSE:
	case STATE_W4_USER_Z_RESPONSE:
	case STATE_W4_USER_BLOCK:
	case STATE_W4_USER_ACK:
		TRACE ("state_timeout, resend last buffer");
		resend_last (handle);
		break;
	case STATE_NEED_ID:
		if (handle -> user_send_size) {
			send_user_buffer (handle);
		} else {
			send_id_request (handle);
		}
		break;
	case STATE_OFFLINE:
	case STATE_W4_RESET:
		send_hangup (handle);
		break;
	case STATE_W4_ESCAPE:
		if (handle -> user_send_size) {
			send_user_buffer (handle);
		} else {
			set_state_idle (handle);
		}
		break;
	case STATE_W4_MV90_COMMAND:
	case STATE_W4_MV90_BLOCK:
	case STATE_W4_MV90_ACK:
		send_hangup (handle);
		break;
	default:
		ERROR_TRACE_I ("state_timeout, unhandled state = %d", handle -> state);
		send_hangup (handle);
	}
}

static void resend_last (HANDLE_DEF *handle)
{
	TRACE_I ("resend_last, retry_count = %d", handle -> retry_count);
	if (handle -> retry_count < handle -> max_retries) {
		++handle -> retry_count;
		TRACE_B ("resend_last"
					 , handle -> send_buffer_size
					 , handle -> send_buffer);
		TRACE_I ("resend_last, state = %d", handle -> state);
		if (!send_command (handle
						, handle -> send_buffer_size
						, handle -> send_buffer)) {
			ERROR_TRACE ("state_timeout, send_command falied");
			return;
		}
		handle -> read_index = 0;
		handle -> start_tick = GetTickCount ();
	} else {
		TRACE ("resend_last, retries exhausted");
		close_handle (handle);
		handle -> state = STATE_NEED_RESET;
		log_console ("Max Retries", handle -> port, "Retries:", handle -> max_retries);
	}
}

static METER_STATUS send_user_buffer (HANDLE_DEF *handle)
{
	TRACE ("send_user_buffer, entry point");
	if (!handle -> user_send_size) {
		ERROR_TRACE ("send_user_buffer, user_send_size is zero");
		return METER_BAD_HANDLE;
	}
	if (!send_command (handle
					 , handle -> user_send_size
					 , handle -> user_send_data)) {
		ERROR_TRACE ("send_user_buffer, send_command failed");
		handle -> user_send_size = 0;
		return METER_IO_ERROR;
	}
	handle -> user_send_size = 0;

	handle -> start_tick = GetTickCount ();
	handle -> user_last_command = GetTickCount ();
	handle -> retry_count = 0;
	handle -> read_index = 0;
	handle -> read_buffer = handle -> user_read_data;
	switch (handle -> user_send_data [1]) {
	case METER_X_COMMAND:
		TRACE ("send_user_buffer, X_COMMAND");
		handle -> state = STATE_W4_USER_X_RESPONSE;
		handle -> timeout = handle -> command_timeout;
		handle -> read_buffer_size = RESPONSE_SIZE;
		break;
	case METER_Y_COMMAND:
		TRACE ("send_user_buffer, Y_COMMAND");
		handle -> state = STATE_W4_USER_Y_RESPONSE;
		handle -> timeout = handle -> command_timeout;
		handle -> read_buffer_size = RESPONSE_SIZE;
		break;
	case METER_Z_COMMAND:
		ERROR_TRACE ("send_user_buffer, Z_COMMAND");
		handle -> state = STATE_W4_USER_Z_RESPONSE;
		handle -> timeout = handle -> command_timeout;
		handle -> read_buffer_size = RESPONSE_SIZE;
		break;
	case METER_ACK:
		TRACE ("send_user_buffer, METER_ACK");
		if ((!handle -> user_send_data [2]) && (!handle -> user_table)) {
			TRACE ("send_user_buffer, ack to last block");
			handle -> state = STATE_W4_MV90_COMMAND;
			handle -> timeout = handle -> user_timeout;
			handle -> read_buffer_size = 0;
		} else {
			TRACE ("send_user_buffer, more blocks to come");
			handle -> state = STATE_W4_USER_BLOCK;
			handle -> timeout = handle -> table_timeout;
			handle -> read_buffer_size = BLOCK_SIZE;
		}
		break;
	case METER_BLOCK_ID:
		ERROR_TRACE ("send_user_buffer, BLOCK_ID");
		handle -> state = STATE_W4_USER_ACK;
		handle -> timeout = handle -> user_timeout;
		handle -> read_buffer_size = BLOCK_ACK_SIZE;
		break;
	case METER_ESC:
		handle -> state = STATE_W4_MV90_COMMAND;
		handle -> timeout = handle -> user_timeout;
		handle -> read_buffer_size = 0;
		break;
	case METER_NACK:
		TRACE ("send_user_buffer, METER_NACK");
		handle -> state = STATE_W4_USER_BLOCK;
		handle -> timeout = handle -> table_timeout;
		handle -> read_buffer_size = BLOCK_SIZE;
		break;
	default:
		ERROR_TRACE_I ("send_user_buffer, unknown request = 0x%X"
					 , handle -> user_send_data [1]);
	}
	return METER_NORMAL;
}

static void send_id_request (HANDLE_DEF *handle)
{
	USHORT		crc;

	TRACE ("send_id_request, entry point");
	handle -> request [0] = METER_STX;
	handle -> request [1] = METER_X_COMMAND;
	handle -> request [2] = handle -> sequence;
	handle -> request [3] = 0; /* Get ID command high byte */
	handle -> request [4] = 0xE; /* Get ID command low byte */
	handle -> request [5] = 0; /**/
	handle -> request [6] = 0; /**/
	handle -> request [7] = 0; /**/
	handle -> request [8] = 0; /**/
	memcpy (&handle -> request [9], handle -> write_password, PASSWORD_SIZE);
	crc = get_crc (12, &handle -> request [1]);
	handle -> request [13] = (UCHAR) (crc >> 8);
	handle -> request [14] = (UCHAR) crc;
	handle -> request_size = 15;
	if (!send_command (handle, handle -> request_size, handle -> request)) {
		ERROR_TRACE ("send_id_request, send_command failed");
		return;
	}

	handle -> state = STATE_W4_ID;
	handle -> start_tick = GetTickCount ();
	handle -> timeout = handle -> command_timeout;
	handle -> retry_count = 0;
	handle -> read_index = 0;
	handle -> read_buffer = handle -> response;
	handle -> read_buffer_size = sizeof (handle -> response);
}

static void send_hangup (HANDLE_DEF *handle)
{
	USHORT		crc;

	TRACE ("send_hangup, entry point");
	handle -> request [0] = METER_STX;
	handle -> request [1] = METER_X_COMMAND;
	handle -> request [2] = handle -> sequence;
	handle -> request [3] = 0; /* Get ID command high byte */
	handle -> request [4] = 0x10; /* Get ID command low byte */
	handle -> request [5] = 0; /**/
	handle -> request [6] = 0; /**/
	handle -> request [7] = 0; /**/
	handle -> request [8] = 0; /**/
	memcpy (&handle -> request [9], handle -> write_password, PASSWORD_SIZE);
	crc = get_crc (12, &handle -> request [1]);
	handle -> request [13] = (UCHAR) (crc >> 8);
	handle -> request [14] = (UCHAR) crc;
	handle -> request_size = 15;
	if (!send_command (handle, handle -> request_size, handle -> request)) {
		ERROR_TRACE ("send_hangup, send_command failed");
		return;
	}

	handle -> state = STATE_W4_HANGUP;
	handle -> start_tick = GetTickCount ();
	handle -> timeout = handle -> command_timeout;
	handle -> retry_count = 0;
	handle -> read_index = 0;
	handle -> read_buffer = handle -> response;
	handle -> read_buffer_size = sizeof (handle -> response);
}

static void send_y_request (HANDLE_DEF *handle, TABLE_DEF *table)
{
	USHORT		crc;

	TRACE_I ("send_y_request, entry point table = %d", table -> table_number);
	handle -> request [0] = METER_STX;
	handle -> request [1] = METER_Y_COMMAND;
	handle -> request [2] = handle -> sequence;
	handle -> request [3] = (UCHAR) table -> table_number;
	handle -> request [4] = 0; /* displacment high byte */
	handle -> request [5] = 0; /* displacment low byte */
	handle -> request [6] = 0; /* number of blocks high byte */
	handle -> request [7] = 0; /* number of blocks low byte */
	handle -> request [8] = 0; /* bytes in last block ??? */
	memcpy (&handle -> request [9], handle -> write_password, PASSWORD_SIZE);
	crc = get_crc (12, &handle -> request [1]);
	handle -> request [13] = (UCHAR) (crc >> 8);
	handle -> request [14] = (UCHAR) crc;
	handle -> request_size = 15;

	if (!send_command (handle, handle -> request_size, handle -> request)) {
		ERROR_TRACE ("send_y_request, send_command failed");
		table -> table_size = 0;
		return;
	}
	handle -> current_table = table;
	handle -> state = STATE_W4_TABLE;
	handle -> start_tick = GetTickCount ();
	handle -> timeout = handle -> command_timeout;
	handle -> retry_count = 0;
	handle -> read_index = 0;
	handle -> read_buffer = handle -> response;
	handle -> read_buffer_size = sizeof (handle -> response);
	return;
}

static void send_reset (HANDLE_DEF *handle)
{
	USHORT		crc;

	TRACE ("send_reset, entry point");
	handle -> request [0] = METER_STX;
	handle -> request [1] = METER_ESC;
	handle -> request [2] = 0;
	crc = get_crc (2, &handle -> request [1]);
	handle -> request [3] = (UCHAR) (crc >> 8);
	handle -> request [4] = (UCHAR) crc;
	handle -> request_size = 5;
	if (!send_command (handle, handle -> request_size, handle -> request)) {
		log_console ("Meter unavailable", handle -> port, "", 0);
		TRACE ("send_reset, send_command state to offline");
	} else {
		handle -> timeout = handle -> reset_timeout;
		handle -> state = STATE_W4_RESET;
		handle -> start_tick = GetTickCount ();
		handle -> read_index = 0;
	}
	handle -> retry_count = 0;
	handle -> read_buffer = NULL;
	handle -> read_buffer_size = 0;
}

static USHORT get_crc (size_t buffer_size, UCHAR *buffer)
{
	size_t		bit;
	size_t		index;
	ULONG		crc;

	crc = 0;
	for (index = 0; index < buffer_size; ++index) {
		for (bit = 0; bit < 8; ++bit) {
			crc ^= ((0x01 & (buffer [index] >> (7 - bit))) ? 0x8000 : 0);
			crc = crc << 1;
			if (crc & 0x10000L) {
				crc = (crc & 0xFFFFL) ^ 0x8005;
			}
		}
	}
	return (USHORT) (crc & 0xFFFF);
}

static int send_command (HANDLE_DEF *handle, size_t buffer_size, UCHAR *buffer)
{
	int		status;

	TRACE_B ("send_command", buffer_size, buffer);
	if (!handle -> lower_handle) {
		status = level1_connect (&handle -> lower_handle, handle -> name);
		if (LEVEL1_NORMAL != status) {
			TRACE_I ("send_command, level1_connect = %d", status);
			handle -> lower_handle = (LEVEL1_HANDLE) NULL;
			set_state_offline (handle);
			return FALSE;
		}
		taskDelay (50);
	}
	handle -> send_buffer = buffer;
	handle -> send_buffer_size = buffer_size;
	status = level1_write (handle -> lower_handle, buffer_size, buffer);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("send_command, level1_write = %d", status);
		set_state_offline (handle);
		status = level1_shutdown (handle -> lower_handle);
		if (LEVEL1_NORMAL != status) {
			ERROR_TRACE_I ("send_command, level1_shutdown = %d", status);
		}
		handle -> lower_handle = (LEVEL1_HANDLE) NULL;
		return FALSE;
	}
	return TRUE;
}

static void close_handle (HANDLE_DEF *handle)
{
	int		status;

	if (handle -> lower_handle) {
 		TRACE ("close_handle, shutdown lower_handle");
		send_hangup (handle);
		status = level1_shutdown (handle -> lower_handle);
		if (LEVEL1_NORMAL != status) {
			ERROR_TRACE_I ("close_handle, level1_shutdown = %d", status);
		}
		handle -> lower_handle = (LEVEL1_HANDLE) NULL;
	}
	handle -> sequence = 1;
}

static void read_finished (HANDLE_DEF *handle)
{
	switch (handle -> state) {
	case STATE_IDLE:
		ERROR_TRACE ("read_finished, superflous line noise");
		handle -> read_index = 0;
		return;
	case STATE_W4_ID:
		TRACE ("read_finished, STATE_W4_ID");
		if (check_response (handle)) {
			handle -> unit_id = handle -> response [8] << 24;
			handle -> unit_id |= handle -> response [9] << 16;
			handle -> unit_id |= handle -> response [10] << 8;
			handle -> unit_id |= handle -> response [11] & 0xFF;
			set_state_idle (handle);
		} else {
			ERROR_TRACE ("read_finished, STATE_W4_ID check_response");
			resend_last (handle);
		}
		return;
	case STATE_W4_HANGUP:
		TRACE ("read_finished, STATE_W4_HANGUP");
		if (check_response (handle)) {
			if (UNIT_ID_UNKNOWN == handle -> unit_id) {
				handle -> read_index = 0;
				handle -> state = STATE_NEED_ID;
				handle -> start_tick = GetTickCount ();
				handle -> timeout = handle -> hangup_timeout;
				handle -> read_buffer = NULL;
				handle -> read_buffer_size = 0;
			} else {
				set_state_idle (handle);
			}
		} else {
			ERROR_TRACE ("read_finished, STATE_W4_HANGUP check_response");
			resend_last (handle);
		}
		return;
	case STATE_W4_TABLE:
		TRACE ("read_finished, STATE_W4_TABLE");
		if (check_response (handle)) {
			if (!handle -> user_send_size) {
				header_finished (handle);
			} else {
				TRACE ("read_finished, user_control pending");
				send_reset (handle);
			}
		} else {
			ERROR_TRACE ("read_finished, STATE_W4_TABLE check_response");
			resend_last (handle);
		}
		return;
	case STATE_W4_BLOCK:
		TRACE ("read_finished, STATE_W4_BLOCK");
		if (check_block (handle)) {
			if (!handle -> user_send_size) {
				block_finished (handle);
			} else {
				TRACE ("read_finished, user_control pending");
				send_reset (handle);
			}
		} else {
			send_reset (handle);
		}
		return;
	case STATE_W4_USER_X_RESPONSE:
	case STATE_W4_USER_Y_RESPONSE:
	case STATE_W4_USER_Z_RESPONSE:
	case STATE_W4_USER_BLOCK:
	case STATE_W4_USER_ACK:
		TRACE_I ("read_finished, user_control response = %d", handle -> state);
		user_finished (handle);
		return;
	default:
		ERROR_TRACE_I ("read_finished, unknown state = %d", handle -> state);
		set_state_idle (handle);
		return;
	}
}

static int check_response (HANDLE_DEF *handle)
{
	long		unit_id;
	USHORT		recv_crc;
	USHORT		user_crc;
	UCHAR		*response;

	TRACE ("check_response, entry point");
	response = handle -> response;
	if (METER_STX != response [0]) {
		ERROR_TRACE_I ("check_response, response [0] = 0x%X", response [0]);
		return FALSE;
	}
	if (METER_NACK == response [1]) {
		ERROR_TRACE_I ("check_response, NACK response [1] = 0x%X", response [1]);
		return FALSE;
	}
	if (METER_ACK != response [1]) {
		ERROR_TRACE_I ("check_response, response [1] = 0x%X", response [1]);
		return FALSE;
	}
	unit_id = handle -> response [2] << 24;
	unit_id |= handle -> response [3] << 16;
	unit_id |= handle -> response [4] << 8;
	unit_id |= handle -> response [5] & 0xFF;
	if (UNIT_ID_UNKNOWN != handle -> unit_id) {
		if (unit_id != handle -> unit_id) {
			ERROR_TRACE_I ("check_response, unit_id = %d", unit_id);
			ERROR_TRACE_I ("check_response, handle unit_id = %d", handle -> unit_id);
		}
	}
	if (!check_status (handle)) {
		ERROR_TRACE_I ("check_response, status response [6] = 0x%X", response [6]);
		return FALSE;
	}
	if (response [7] != handle -> sequence) {
		ERROR_TRACE_I ("check_response, response [7] = %d", response [7]);
		ERROR_TRACE_I ("check_response, sequence = %d", handle -> sequence);
		return FALSE;
	}
	recv_crc = get_crc (13, &response [1]);
	user_crc = (response [14] << 8) | response [15];
	if (recv_crc != user_crc) {
		ERROR_TRACE_I ("check_response, recv_crc = 0x%X", recv_crc);
		ERROR_TRACE_I ("check_response, user_crc = 0x%X", user_crc);
	}
	++handle -> sequence;
	return TRUE;
}

static int check_block (HANDLE_DEF *handle)
{
	UCHAR		*response;
	USHORT		recv_crc;
	USHORT		user_crc;
	USHORT		recv_block_number;
	long		unit_id;
	TABLE_DEF	*table;
			 
	TRACE ("check_block, entry point");
	response = handle -> block_data;
	if (METER_STX != response [0]) {
		ERROR_TRACE_I ("check_block, response [0] = 0x%X", response [0]);
		return FALSE;
	}
	if (METER_BLOCK_ID != response [1]) {
		ERROR_TRACE_I ("check_block, response [1] = 0x%X", response [1]);
		return FALSE;
	}
	unit_id = handle -> response [2] << 24;
	unit_id |= handle -> response [3] << 16;
	unit_id |= handle -> response [4] << 8;
	unit_id |= handle -> response [5] & 0xFF;
	if (UNIT_ID_UNKNOWN != handle -> unit_id) {
		if (unit_id != handle -> unit_id) {
			ERROR_TRACE_I ("check_block, unit_id = %d", unit_id);
			ERROR_TRACE_I ("check_block, handle unit_id = %d", handle -> unit_id);
			return FALSE;
		}
	}
	recv_crc = get_crc (263, &response [1]);
	user_crc = (response [264] << 8) | response [265];
	if (recv_crc != user_crc) {
		ERROR_TRACE_I ("check_block, recv_crc = 0x%X", recv_crc);
		ERROR_TRACE_I ("check_block, user_crc = 0x%X", user_crc);
		return FALSE;
	}
	table = handle -> current_table;
	if (!table) {
		ERROR_TRACE ("check_block, no table active");
		return FALSE;
	}
	if (!table -> data_current) {
		ERROR_TRACE ("check_block, data_current is NULL");
		return FALSE;
	}
	if (!table -> data_temp) {
		ERROR_TRACE ("check_block, data_temp is NULL");
		return FALSE;
	}
	recv_block_number = (USHORT) ((response [6] << 8) | response [7]);
	if (recv_block_number != table -> block_index) {
		ERROR_TRACE ("check_block, block_number mismatch");
		ERROR_TRACE_I ("check_block, table = %d", table -> table_number);
		ERROR_TRACE_I ("check_block, reply = 0x%X", recv_block_number);
		ERROR_TRACE_I ("check_block, block_index = 0x%X", table -> block_index);
		return FALSE;
		return FALSE;
	}
	if ((METER_EOT != response [266]) && (METER_ETB != response [266])) {
		ERROR_TRACE_I ("check_block, response [266] = 0x%X", response [266]);
		return FALSE;
	}
	return TRUE;
}

static void user_finished (HANDLE_DEF *handle)
{
	TRACE ("user_finished, entry point");
	handle -> user_read_size = handle -> read_index;
	TRACE_I ("user_finished, read_index = %d", handle -> read_index);
	handle -> read_index = 0;
	handle -> user_send_size = 0;
	handle -> user_last_command = GetTickCount ();
	switch (handle -> state) {
	case STATE_W4_USER_X_RESPONSE:
		TRACE ("user_finished, STATE_W4_USER_X_RESPONSE");
		handle -> state = STATE_W4_MV90_COMMAND;
		break;
	case STATE_W4_USER_Y_RESPONSE:
		TRACE ("user_finished, STATE_W4_USER_Y_RESPONSE");
		handle -> state = STATE_W4_MV90_ACK;
		break;
	case STATE_W4_USER_Z_RESPONSE:
		TRACE ("user_finished, STATE_W4_USER_X_RESPONSE");
		handle -> state = STATE_W4_MV90_BLOCK;
		break;
	case STATE_W4_USER_BLOCK:
		TRACE ("user_finished, STATE_W4_USER_BLOCK");
		handle -> state = STATE_W4_MV90_ACK;
		break;
	case STATE_W4_USER_ACK:
		TRACE ("user_finished, STATE_W4_USER_ACK");
		handle -> state = STATE_W4_MV90_COMMAND;
		break;
	default:
		ERROR_TRACE_I ("user_finished, invalid state = %d", handle -> state);
		handle -> state = STATE_W4_MV90_COMMAND;
	}
	handle -> current_table = (TABLE_DEF *) NULL;
	handle -> timeout = handle -> user_timeout;
	handle -> read_buffer = NULL;
	handle -> read_buffer_size = 0;
	if (handle -> new_read_password) {
		memcpy (handle -> read_password, handle -> saved_password, 2);
		save_passwords (handle);
		handle -> new_read_password = FALSE;
	} else if (handle -> new_write_password) {
		memcpy (handle -> write_password, handle -> saved_password, PASSWORD_SIZE);
		save_passwords (handle);
		handle -> new_write_password = FALSE;
	}
}

static void header_finished (HANDLE_DEF *handle)
{
	UCHAR		table_number;
	USHORT		start_offset;
	size_t		block_count;
	UCHAR		byte_count;
	TABLE_DEF	*table;

	TRACE ("header_finished, entry point");
	table = handle -> current_table;
	table_number = handle -> response [8];
	if (table_number != table -> table_number) {
		ERROR_TRACE_I ("header_finished, table_number = %d", table_number);
		ERROR_TRACE_I ("header_finished, requested_table = %d", table -> table_number);
		table -> table_size = 0;
		return;
	}
	start_offset = (USHORT) ((handle -> response [9] << 8) | handle -> response [10]);
	block_count = ((handle -> response [11] << 8) | handle -> response [12]);
	if (!block_count) {
		ERROR_TRACE ("header_finished, zero block count");
		table -> table_size = 0;
		return;
	}
	byte_count = handle -> response [13];
	if (byte_count) {
		table -> last_block_size = byte_count;
	} else {
		table -> last_block_size = 256;
	}
	table -> block_count = block_count;
	table -> block_index = block_count;
	table -> blocks_required = block_count;
	if (table -> data_current) {
		if (table -> table_size != block_count * 256) {
			ERROR_TRACE_I ("header_finished, table_size = %d", table -> table_size);
			ERROR_TRACE_I ("header_finished, block_count * 256 = %d", block_count * 256);
			return;
		}
	} else {
		table -> table_size = block_count * 256;
		table -> data_locked = (UCHAR *) malloc (table -> table_size);
		if (!table -> data_locked) {
			ERROR_TRACE_I ("header_finished, malloc = %d", table -> table_size);
			table -> table_size = 0;
			return;
		}
		table -> data_current = (UCHAR *) malloc (table -> table_size);
		if (!table -> data_current) {
			ERROR_TRACE_I ("header_finished, malloc = %d", table -> table_size);
			free (table -> data_locked);
			table -> table_size = 0;
			return;
		}
		table -> data_temp = (UCHAR *) malloc (table -> table_size);
		if (!table -> data_temp) {
			ERROR_TRACE_I ("header_finished, malloc #2 = %d", table -> table_size);
			free (table -> data_locked);
			free (table -> data_current);
			table -> data_locked = NULL;
			table -> data_current = NULL;
			table -> table_size = 0;
			return;
		}
	}
	TRACE_S ("header_finished, meter = %s", handle -> name);
	TRACE_I ("header_finished, table_size = %d", table -> table_size);
	TRACE_I ("header_finished, table_number = %d", table_number);
	TRACE_I ("header_finished, start_offset = %d", start_offset);
	TRACE_I ("header_finished, block_count = %d", table -> block_count);
	TRACE_I ("header_finished, byte_count = %d", byte_count);
	TRACE_I ("header_finished, blocks_required = %d", table -> blocks_required);
	TRACE_I ("header_finished, last_block_size = %d", table -> last_block_size);
	build_ack_request (handle, table -> block_index);
	if (!send_command (handle, handle -> block_ack_size, handle -> block_ack)) {
		ERROR_TRACE ("header_finished, send_command failed");
		table -> table_size = 0;
		return;
	}
	--table -> block_index;
	table -> byte_count = 0;

	handle -> state = STATE_W4_BLOCK;
	handle -> start_tick = GetTickCount ();
	handle -> timeout = handle -> table_timeout;
	handle -> retry_count = 0;
	handle -> read_index = 0;
	handle -> read_buffer = handle -> block_data;
	handle -> read_buffer_size = sizeof (handle -> block_data);
	return;
}

static void block_finished (HANDLE_DEF *handle)
{
	TABLE_DEF		*table;
	size_t			table_index;
	size_t			data_size;
	UCHAR			*block_data;
	UCHAR			*new_date_time;
	UCHAR			*old_date_time;

	TRACE ("block_finished, entry point");
	block_data = handle -> block_data;
	table = handle -> current_table;
	data_size = handle -> read_index - 11;
	table_index = ((table -> block_count - 1) - table -> block_index) * 256;
	if (table_index + data_size > table -> table_size) {
		ERROR_TRACE ("block_finished, user buffer overflow");
		ERROR_TRACE_I ("block_finished, table = %d", table -> table_number);
		ERROR_TRACE_I ("block_finished, index = %d", table_index);
		ERROR_TRACE_I ("block_finished, block_size = %d", data_size);
		ERROR_TRACE_I ("block_finished, total_size = %d", table -> table_size);
		set_state_idle (handle);
		return;
	}
	TRACE_I ("block_finished, table_index = %d", table_index);
	TRACE_I ("block_finished, data_size = %d", data_size);
	memcpy (&table -> data_temp [table_index]
		  , &block_data [8]
		  , data_size);
	table -> byte_count += data_size;

	TRACE_I ("block_finished, block_index %d", table -> block_index);
	TRACE_I ("block_finished, table %d block loaded", table -> table_number);
	if ((table -> table_number == 12)
	&&  (table -> completed)
	&&  (table -> blocks_required == table -> block_count)) {
		TRACE ("block_finished, table 12 check date stamp");
		new_date_time = &table -> data_temp [2 * handle -> max_storage_channels];
		old_date_time = &table -> data_current [2 * handle -> max_storage_channels];
		if (!memcmp (new_date_time, old_date_time, 6)) {
			TRACE ("block_finished, table 12 up to date");
			table -> blocks_required = 0;
		} else {
			TRACE ("block_finished, only load part of table 12");
			table -> blocks_required = get_blocks_required (handle
														  , new_date_time
														  , old_date_time);
		}
	} else if ((table -> table_number == 15)
	&&         (table -> completed)
	&&  	   (table -> blocks_required == table -> block_count)) {
		TRACE ("block_finished, table 15 check date stamp");
		new_date_time = table -> data_temp;
		old_date_time = table -> data_current;
		if (!memcmp (new_date_time, old_date_time, 6)) {
			TRACE ("block_finished, table 15 up to date");
			table -> blocks_required = 0;
		} else {
			TRACE ("block_finished, reload first 12 blocks of table 15");
			table -> blocks_required = 12;
		}
	} else {
		--table -> blocks_required;
	}
	TRACE_I ("block_finished, blocks_required = %d", table -> blocks_required);
	TRACE_I ("block_finished, block_index = %d", table -> block_index);
	if ((!table -> blocks_required) && (table -> block_index)) {
		build_escape_request (handle, table -> block_index);
	} else {
		build_ack_request (handle, table -> block_index);
	}
	if (!send_command (handle, handle -> block_ack_size, handle -> block_ack)) {
		ERROR_TRACE ("block_finished, send_command failed");
		ERROR_TRACE_I ("block_finished, table = %d", table -> table_number);
		set_state_idle (handle);
		return;
	}
	if (console_on) {
		printf ("Meter (%s) table %d, block %d updated\n"
				, handle -> name
				, table -> table_number
				, table -> block_index);
	}
	if (table -> blocks_required) {
		--table -> block_index;
		handle -> state = STATE_W4_BLOCK;
		handle -> start_tick = GetTickCount ();
		handle -> timeout = handle -> table_timeout;
		handle -> retry_count = 0;
		handle -> read_index = 0;
		handle -> read_buffer = handle -> block_data;
		handle -> read_buffer_size = sizeof (handle -> block_data);
		TRACE_I ("block_finished, wait for next block %d", table -> block_index);
		return;
	}
	if (console_on) {
		printf ("Meter table %d updated\n", table -> table_number);
	}
	handle -> read_delay_tick = GetTickCount ();
	if ((table -> table_number == 12) && (table -> completed)) {
		table -> needs_update = FALSE;
		table -> last_update = GetTickCount ();
		if (handle -> interval_count) {
			insert_table_12 (handle, table);
			handle -> needs_snapshot = TRUE;
		} else {
			TRACE ("block_finished, no data to insert");
		}
		handle -> state = STATE_W4_ESCAPE;
		handle -> current_table = (TABLE_DEF *) NULL;
		handle -> start_tick = GetTickCount ();
		handle -> timeout = handle -> reset_timeout;
		handle -> retry_count = 0;
		handle -> read_index = 0;
		handle -> read_buffer = NULL;
		handle -> read_buffer_size = 0;
		return;
	}
	TRACE_I ("block_finished, size = %d", table -> table_size);
	if (table -> completed) {
		if (!memcmp (table -> data_current
				   , table -> data_temp
				   , table -> table_size)) {
			++table -> diff_count;
			TRACE_I ("block_finished, diff_count = %d", table -> diff_count);
		} else {
			table -> diff_count = 0;
		}
	}
	memcpy (table -> data_current, table -> data_temp, table -> table_size);
	table -> completed = TRUE;
	handle -> needs_snapshot = TRUE;
	table -> needs_update = FALSE;
	switch (table -> table_number) {
	case 0:
		parse_table_0 (handle, table);
		break;
	case 3:
		parse_table_3 (handle, table);
		break;
	case 11:
		parse_table_11 (handle, table);
		break;
	default:
		TRACE_I ("block_finished, table %d complete", table -> table_number);
	}
	table -> last_update = GetTickCount ();
	set_state_idle (handle);
}

static int get_blocks_required (HANDLE_DEF		*handle
							  , UCHAR			*new_date_time
							  , UCHAR			*old_date_time)
{
	long		required;
	long		difference;
	TIME_DEF	last_time;
	TIME_DEF	this_time;

	TRACE ("get_blocks_required, entry point");
	get_date_time (old_date_time, &last_time);
	get_date_time (new_date_time, &this_time);
	TRACE_L ("get_blocks_required, last_time = %d", last_time . tv_sec);
	TRACE_L ("get_blocks_required, this_time = %d", this_time . tv_sec);
	difference = this_time . tv_sec - last_time . tv_sec;
	TRACE_L ("get_blocks_required, difference = %d", difference);
	if (!difference) {
		return 0;
	}
	if (!handle -> interval_time) {
		ERROR_TRACE ("get_blocks_required, interval undefined");
		return 0;
	}
	required = difference / (60 * handle -> interval_time);
	TRACE_L ("get_blocks_required, required = %ld", required);
	handle -> interval_count = required;
	handle -> interval_size = (handle -> data_size * handle -> channel_count) / 8;
	return 1 + ((required * handle -> interval_size) / 256);
}

static void get_date_time (UCHAR *bcd_date_time, TIME_DEF *this_time)
{
	DATE_TIME_DEF	date_time;

	date_time . month = get_bcd (bcd_date_time [0]);
	TRACE_L ("get_date_time, month = %ld", date_time . month);
	date_time . day = get_bcd (bcd_date_time [1]);
	TRACE_L ("get_date_time, day = %ld", date_time . day);
	if (bcd_date_time [2] == 0x99) {
		date_time . year = 1999;
	} else {
		date_time . year = 2000 + get_bcd (bcd_date_time [2]);
	}
	TRACE_L ("get_date_time, year = %ld", date_time . year);
	date_time . hour = get_bcd (bcd_date_time [3]);
	TRACE_L ("get_date_time, hour = %ld", date_time . hour);
	date_time . minute = get_bcd (bcd_date_time [4]);
	TRACE_L ("get_date_time, minute = %ld", date_time . minute);
	date_time . second = 0;
	time_date_to_time (&date_time, this_time);
}

static long get_bcd (UCHAR bcd_value)
{
	return ((bcd_value >> 4) * 10) + (bcd_value & 0x0F);
}

static void put_bcd (int value, UCHAR *bcd_value)
{
	*bcd_value = ((value / 10) << 4) | (value % 10);
}

static void insert_table_12 (HANDLE_DEF *handle, TABLE_DEF *table)
{
	size_t		count;
	size_t		insert_size;
	size_t		shift_count;
	size_t		header_size;
	size_t		in_offset;
	size_t		out_offset;
	size_t		data_end;
	UCHAR		*table_12_data;
	UCHAR		*table_12_temp;

	TRACE ("insert_table_12, entry point");
	insert_size = handle -> interval_count * handle -> interval_size;
	TRACE_I ("insert_table_12, shift_size = %d", insert_size);
	if (insert_size > table -> byte_count) {
		ERROR_TRACE ("insert_table_12, table underflow");
		ERROR_TRACE_I ("insert_table_12, byte_count = %d", table -> byte_count);
		ERROR_TRACE_I ("insert_table_12, interval_count = %d", handle -> interval_count);
		return;
	}
	header_size = 6 + (2 * handle -> max_storage_channels);
	data_end = header_size + (1024 * handle -> mem_size);
	if (data_end > table -> table_size) {
		ERROR_TRACE ("insert_table_12, table overflow");
		ERROR_TRACE_I ("insert_table_12, data_end = %d", data_end);
		ERROR_TRACE_I ("insert_table_12, table_size = %d", table -> table_size);
		return;
	}
	TRACE_I ("insert_table_12, data_end = %d", data_end);
	TRACE_I ("insert_table_12, interval_count = %d", handle -> interval_count);

	shift_count = (1024 * handle -> mem_size) - insert_size;
	TRACE_I ("insert_table_12, insert_size = %d", insert_size);
	TRACE_I ("insert_table_12, shift_count = %d", shift_count);
	if (shift_count > (table -> table_size - header_size)) {
		ERROR_TRACE ("insert_table_12, shift overflow");
		ERROR_TRACE_I ("insert_table_12, shift_count = %d", shift_count);
		ERROR_TRACE_I ("insert_table_12, table_size = %d", table -> table_size);
		ERROR_TRACE_I ("insert_table_12, header_size = %d", header_size);
		return;
	}
	in_offset = (data_end - insert_size) - 1;
	out_offset = data_end - 1;
	table_12_data = table -> data_current;
	table_12_temp = table -> data_temp;
	for (count = 0; count < shift_count; ++count) {
		table_12_data [out_offset] = table_12_data [in_offset];
		--in_offset;
		--out_offset;
	}
	TRACE_I ("insert_table_12, in_offset = %d", in_offset);
	TRACE_I ("insert_table_12, out_offset = %d", out_offset);
	if (in_offset != (header_size - 1)) {
		ERROR_TRACE_I ("insert_table_12, in_offset = %d", in_offset);
	}
	memcpy (table_12_data, table_12_temp, header_size + insert_size);
}

static void parse_table_0 (HANDLE_DEF *handle, TABLE_DEF *table_0)
{
	int			bit;
	int			byte;
	UCHAR		table_number;
	UCHAR		*table_bits;
	TABLE_DEF	*table;

	handle -> max_storage_channels = table_0 -> data_temp [48];
	TRACE_I ("parse_table_0, max_storage = %d", handle -> max_storage_channels);
	handle -> max_stor_k = table_0 -> data_temp [84];
	TRACE_I ("parse_table_0, max_stor_k = %d", handle -> max_stor_k);
	table_bits = &table_0 -> data_temp [10];
	table_number = 0;
	for (byte = 0; byte < 32; ++byte) {
		TRACE_I ("table_bits [x] = 0x%X", table_bits [byte]);
		for (bit = 0; bit < 8; ++bit) {
			if (table_bits [byte] & (1 << bit)) {
				TRACE_I ("parse_table_0, table_number = %d", table_number);
				if (table_number) {
					table = find_table_by_number (handle, table_number);
					if (!table) {
						table_add (handle, table_number, FALSE);
					}
				}
			}
			++table_number;
		}
	}
}

static void parse_table_3 (HANDLE_DEF *handle, TABLE_DEF *table_3)
{
	TABLE_DEF	*table;
	ULONG		ten_minutes;
	UCHAR		new_time_stamp [6];
	UCHAR		old_time_stamp [6];

	TRACE ("parse_table_3, entry point");
	handle -> mem_size = table_3 -> data_temp [22];
	memcpy (new_time_stamp, &table_3 -> data_temp [16], 6);
	memcpy (old_time_stamp, handle -> time_stamp, 6);
	memcpy (handle -> time_stamp, new_time_stamp, 6);
	handle -> last_table_3 = GetTickCount ();
	if (memcmp (old_time_stamp, new_time_stamp, 6)) {
		TRACE ("parse_table_3, time_stamp_changed");
		table = (TABLE_DEF *) queue_next (&handle -> root_table);
		while (table) {
			table -> needs_update = TRUE;
			table = (TABLE_DEF *) queue_next (&table -> link);
		}
	} else {
		TRACE_B ("parse_table_3, new_stamp", 6, new_time_stamp);
		TRACE_B ("parse_table_3, old_stamp", 6, old_time_stamp);
		ten_minutes = (10L * 60000);
		if ((table_3 -> last_update + ten_minutes) < GetTickCount ()) {
			ERROR_TRACE ("parse_table_3, table 3 is stale");
			table = (TABLE_DEF *) queue_next (&handle -> root_table);
			while (table) {
				table -> needs_update = TRUE;
				table = (TABLE_DEF *) queue_next (&table -> link);
			}
		}
	}
	TRACE_I ("parse_table_3, mem_size = %d", handle -> mem_size);
	TRACE_B ("parse_table_3", 6, handle -> time_stamp);
}

static void parse_table_11 (HANDLE_DEF *handle, TABLE_DEF *table)
{
	TRACE ("parse_table_11, entry point");
	handle -> data_size = table -> data_temp [0];
	handle -> interval_time = table -> data_temp [2];
	handle -> channel_count = table -> data_temp [4];
	TRACE_I ("parse_table_11, data_size = %d", handle -> data_size);
	TRACE_I ("parse_table_11, interval_time = %d", handle -> interval_time);
	TRACE_I ("parse_table_11, channel_count = %d", handle -> channel_count);
}

static void set_state_idle (HANDLE_DEF *handle)
{
	handle -> read_index = 0;
	handle -> state = STATE_IDLE;
	handle -> current_table = (TABLE_DEF *) NULL;
	handle -> read_buffer = NULL;
	handle -> read_buffer_size = 0;
	handle -> read_delay_tick = GetTickCount ();
}

static void set_state_offline (HANDLE_DEF *handle)
{
	handle -> read_index = 0;
	handle -> state = STATE_OFFLINE;
	handle -> start_tick = GetTickCount ();
	handle -> timeout = handle -> offline_timeout;
}

static void refresh_table (HANDLE_DEF *handle, int table_number)
{
	TABLE_DEF		*table;

	TRACE ("refresh_table, entry point");
	table = (TABLE_DEF *) queue_next (&handle -> root_table);
	while (table) {
		if (table -> table_number == table_number) {
			handle -> read_delay_tick = 0;
			table -> needs_update = TRUE;
			return;
		}
 		table = (TABLE_DEF *) queue_next (&table -> link);
 	}
}

static int all_tables_current (HANDLE_DEF *handle)
{
	TABLE_DEF		*table;

	TRACE ("all_tables_current, entry point");
	table = (TABLE_DEF *) queue_next (&handle -> root_table);
	while (table) {
		if (table -> needs_update) {
			return FALSE;
		}
 		table = (TABLE_DEF *) queue_next (&table -> link);
 	}
	return TRUE;
}

static void lock_tables (HANDLE_DEF *handle)
{
	TABLE_DEF		*table;

	TRACE ("lock_tables, entry point");
	table = (TABLE_DEF *) queue_next (&handle -> root_table);
	while (table) {
		memcpy (table -> data_locked, table -> data_current, table -> table_size);
		table = (TABLE_DEF *) queue_next (&table -> link);
	}
}

static TABLE_DEF *find_table_by_number (HANDLE_DEF *handle, int number)
{
	TABLE_DEF	*table;

	table = (TABLE_DEF *) queue_next (&handle -> root_table);
	while (table) {
		if ((UCHAR) number == table -> table_number) {
			return table;
		}
		table = (TABLE_DEF *) queue_next (&table -> link);
	}
	return (TABLE_DEF *) NULL;
}

static void save_passwords (HANDLE_DEF *handle)
{
	char		file_name [MAX_STRING];
	TFILE		pw_file;

	TRACE ("save_passwords, entry point");
	sprintf (file_name, "%sX%d.x", handle -> password_dir, handle -> port);
	create_file (file_name);
	pw_file = TOPENW (file_name);
	if (TERROR (pw_file)) {
		ERROR_TRACE_S ("save_password, open failed on (%s)", file_name);
		return;
	}
	TWRITE (pw_file, handle -> write_password, PASSWORD_SIZE);
	TWRITE (pw_file, handle -> read_password, 2);
	TCLOSE (pw_file);
}

static void read_passwords (HANDLE_DEF *handle)
{
	char		file_name [MAX_STRING];
	char		both_passwords [MAX_STRING];
	TFILE		pw_file;

	TRACE ("read_passwords, entry point");
	sprintf (file_name, "X%d.x", handle -> port);
	pw_file = TOPENR (file_name);
	if (TERROR (pw_file)) {
		TRACE_S ("read_passwords, no file = (%s)", file_name);
		return;
	}
	TREAD (pw_file, both_passwords, MAX_STRING);
	memcpy (handle -> write_password, both_passwords, PASSWORD_SIZE);
	memcpy (handle -> read_password, &both_passwords [4], 2);
	TRACE_B ("read_passwords, read", 2, handle -> read_password);
	TRACE_B ("read_passwords, write", 4, handle -> write_password);
	TCLOSE (pw_file);
}

static int check_status (HANDLE_DEF *handle)
{
	TRACE_I ("check_status, code = 0x%X", handle -> response [6]);
	if (handle -> response [6] & 0x01) {
		ERROR_TRACE_S ("check_status, RAM failure Meter(%s)", handle -> name);
	}
	if (handle -> response [6] & 0x02) {
		ERROR_TRACE_S ("check_status, Framing Error Meter(%s)", handle -> name);
		return FALSE;
	}
	if (handle -> response [6] & 0x04) {
		ERROR_TRACE_S ("check_status, Overrun Error Meter(%s)", handle -> name);
		return FALSE;
	}
	if (handle -> response [6] & 0x08) {
		ERROR_TRACE_S ("check_status, Internal Error Meter(%s)", handle -> name);
		return FALSE;
	}
	if (handle -> response [6] & 0x10) {
		TRACE_S ("check_status, ROM failure Meter(%s)", handle -> name);
	}
	if (handle -> response [6] & 0x20) {
		ERROR_TRACE_S ("check_status, CRC error Meter(%s)", handle -> name);
		return FALSE;
	}
	if (handle -> response [6] & 0x40) {
		ERROR_TRACE_S ("check_status, Command Error Meter(%s)", handle -> name);
	}
	if (handle -> response [6] & 0x80) {
		ERROR_TRACE_S ("check_status, Password Error Meter(%s)", handle -> name);
		return FALSE;
	}
	return TRUE;
}

static void log_console (char *message, int port, char *header, int value)
{
	time_t	time_stamp;
	char	*c_string;

	time (&time_stamp);
	c_string = ctime (&time_stamp);
	if (console_on) {
		printf ("Quad4:(%s), port = %d, %s = %d, %s"
				, message
				, port
				, header, value
				, c_string);
	}
}

static void create_file (char *file_name)
{
	FILE	*stupid_systems;

	stupid_systems = fopen (file_name, "w");
	fclose (stupid_systems);
}

