/*buffPoolMgt.c,v 1.1.1.1 1999/08/03 14:41:58 tis4000 Exp */

/****************************************************************/
/*								*/
/*	BUFFER POOL MANAGEMENT FOR CHANNEL ACQUISITION		*/
/*								*/
/*	All of the control blocks in <channelControl.h> are	*/
/*	acquired and released during normal operation. These	*/
/*	routines manage pools of these buffers. What we do	*/
/*	is acquire chunks of these blocks from the operating	*/
/*	system and never return the memory. The reason for	*/
/*	this is to prevent fragmentation and also reduce the	*/
/*	number of trips that we make through memory allocators.	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef WIN32
#include <winsock.h>
#else
#ifdef vxWorks
#include <vxWorks.h>
#include <sockLib.h>
#include <inetLib.h>
#include <netinet/tcp.h>
#else not vxWorks
#include <sys/types.h>
#include <tcp.h>
#include <netdb.h>
#include <netdb.h>
#include <netinet/in.h>
#endif vxWorks
#endif WIN32
#include <channelControl.h>
#if defined(AXP) || defined(WIN32)
#include <stdlib.h>
#endif AXP

#ifdef vxWorks
#include <string.h>
#include <stdlib.h>
#endif	vxWorks


/****************************************************************/
/*								*/
/*			TUNING PARAMETERS			*/
/*								*/
/*	Whenever we allocate structures from the operating	*/
/*	system, we allocate a bunch of them. These variables	*/
/*	control the size of the chunks.				*/
/*								*/
/****************************************************************/

#define	CHB_ALLOC	20
#define	CAB_ALLOC	20
#define	ACB_ALLOC	 2
#define NSQ_ALLOC	10

/******
	Pointers to the internal memory pools
******/

static CHB	*chbPool = 0;
static CAB	*cabPool = 0;
static ACB	*acbPool = 0;
static NSQ	*nsqPool = 0;


/****************************************************************/
/*								*/
/*		ALLOCATE A CHANNEL HEADER BLOCK (CHB)		*/
/*								*/
/****************************************************************/

CHB	*newCHB(void)
{
    CHB		*currChb;
    int		index;
#ifdef POINTER64
    int		tokenVal;
#endif

    /*
    If the pool is empty, fill it up again.
    */

    if (!chbPool)
	{
	chbPool = (CHB *) malloc(CHB_ALLOC * sizeof(CHB));
	if (!chbPool)
	    return (0);
	currChb = chbPool;
	index = CHB_ALLOC;
	while (index--)
	    {
	    if (index)
		currChb -> pNextIOC1 = (struct CHDR *)(currChb + 1);
	    else
		currChb -> pNextIOC1 = 0;
	    currChb++;
	    }
	}

    /*
    Take the one off the front of the pool and clear it.
    */

#ifdef POINTER64
    if ((tokenVal = TokenFromPointer64 (chbPool)) == 0)
	return 0;
#endif
    currChb = chbPool;
    chbPool = (CHB *) currChb -> pNextIOC1;
    memset((void *) currChb, 0, sizeof(CHB));
#ifdef POINTER64
    currChb->chbToken = tokenVal;
#endif
    return (currChb);
}


/****************************************************************/
/*								*/
/*		FREE A CHANNEL HEADER BLOCK (CHB)		*/
/*								*/
/****************************************************************/

void deleteCHB(oldChb)
CHB	*oldChb;
{
    /*
    Push this one onto the front of the pool
    */

    oldChb -> pNextIOC1 = (struct CHDR *) chbPool;
    chbPool = oldChb;
}


/****************************************************************/
/*								*/
/*	ALLOCATE A CHANNEL APPLICATION BLOCK (CAB)		*/
/*								*/
/****************************************************************/

CAB	*newCAB(void)
{
    CAB		*currCab;
    int		index;
#ifdef POINTER64
    int		tokenVal;
#endif

    /*
    If the pool is empty, fill it up again.
    */

    if (!cabPool)
	{
	cabPool = (CAB *) malloc(CAB_ALLOC * sizeof(CAB));
	if (!cabPool)
	    return (0);
	currCab = cabPool;
	index = CAB_ALLOC;
	while (index--)
	    {
	    if (index)
		currCab -> pNextChan = currCab + 1;
	    else
		currCab -> pNextChan = 0;
	    currCab++;
	    }
	}

    /*
    Take the one off the front of the pool and clear it.
    */

#ifdef POINTER64
    if ((tokenVal = TokenFromPointer64 (cabPool)) == 0)
	return 0;
#endif
    currCab = cabPool;
    cabPool = currCab -> pNextChan;
    memset((void *) currCab, 0, sizeof(CAB));
#ifdef POINTER64
    currCab->cabToken = tokenVal;
#endif
    return (currCab);
}


/****************************************************************/
/*								*/
/*		FREE A CHANNEL APPLICATION BLOCK (CAB)		*/
/*								*/
/****************************************************************/

void deleteCAB(oldCab)
CAB	*oldCab;
{
    /*
    Push this one onto the front of the pool
    */

    oldCab -> pNextChan = cabPool;
    cabPool = oldCab;
}


/****************************************************************/
/*								*/
/*	ALLOCATE AN APPLICATION CONTROL BLOCK (ACB)		*/
/*								*/
/****************************************************************/

ACB	*newACB(void)
{
    ACB		*currAcb;
    int		index;
#ifdef POINTER64
    int		tokenVal;
#endif

    /*
    If the pool is empty, fill it up again.
    */

    if (!acbPool)
	{
	acbPool = (ACB *) malloc(ACB_ALLOC * sizeof(ACB));
	if (!acbPool)
	    return (0);
	currAcb = acbPool;
	index = ACB_ALLOC;
	while (index--)
	    {
	    if (index)
		currAcb -> pNextACB = currAcb + 1;
	    else
		currAcb -> pNextACB = 0;
	    currAcb++;
	    }
	}

    /*
    Take the one off the front of the pool and clear it.
    */

#ifdef POINTER64
    if ((tokenVal = TokenFromPointer64 (acbPool)) == 0)
	return 0;
#endif
    currAcb = acbPool;
    acbPool = currAcb -> pNextACB;
    memset((void *) currAcb, 0, sizeof(ACB));
#ifdef POINTER64
    currAcb->acbToken = tokenVal;
#endif
    return (currAcb);
}


/****************************************************************/
/*								*/
/*	FREE AN APPLICATION CONTROL BLOCK (ACB)			*/
/*								*/
/****************************************************************/

void deleteACB(oldAcb)
ACB	*oldAcb;
{
    /*
    Push this one onto the front of the pool
    */

    oldAcb -> pNextACB = (struct APCB *) acbPool;
    acbPool = oldAcb;
}


/****************************************************************/
/*								*/
/*	ALLOCATE AN NAME SERVER QUEUE ELEMENT (NSQ)		*/
/*								*/
/****************************************************************/

NSQ	*newNSQ(void)
{
    NSQ		*currNsq;
    int		index;
#ifdef POINTER64
    int		tokenVal;
#endif

    /*
    If the pool is empty, fill it up again.
    */

    if (!nsqPool)
	{
	nsqPool = (NSQ *) malloc(NSQ_ALLOC * sizeof(NSQ));
	if (!nsqPool)
	    return (0);
	currNsq = nsqPool;
	index = NSQ_ALLOC;
	while (index--)
	    {
	    if (index)
		currNsq -> pNextNSQ = currNsq + 1;
	    else
		currNsq -> pNextNSQ = 0;
	    currNsq++;
	    }
	}

    /*
    Take the one off the front of the pool and clear it.
    */

#ifdef POINTER64
    if ((tokenVal = TokenFromPointer64 (nsqPool)) == 0)
	return 0;
#endif
    currNsq = nsqPool;
    nsqPool = currNsq -> pNextNSQ;
    memset((void *) currNsq, 0, sizeof(NSQ));
#ifdef POINTER64
    currNsq->nsqToken = tokenVal;
#endif
    return (currNsq);
}


/****************************************************************/
/*								*/
/*	FREE AN NAME SERVER QUEUE ELEMENT (NSQ)			*/
/*								*/
/****************************************************************/

void deleteNSQ(oldNsq)
NSQ	*oldNsq;
{
    /*
    Push this one onto the front of the pool
    */

    oldNsq -> pNextNSQ = (struct NSQUEUE *) nsqPool;
    nsqPool = oldNsq;
}
