/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	TRAVERSE DATABASE, POSTING POINTS THAT MEET CRITERIA	*/
/*								*/
/*	This is a task that makes a single pass over the	*/
/*	database. If the opCode is a 1, then create a trend	*/
/*	entry in the Channel Access ring buffer that is		*/
/*	restricted to the requesting workstation client.	*/
/*								*/
/*	If the opCode is a 2, then create an alarm entry	*/
/*	in the Channel Access ring buffer for every point.	*/
/*	Restrict these entries to the requesting workstation 	*/
/*	client.							*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <types.h>
#include <semLib.h>
#include <taskLib.h>
#include <stdio.h>
#include <string.h>
#include <memLib.h>
#include <logLib.h>
#include <inetLib.h>
#include <tickLib.h>

#include "inet_ntoa.h"
#include "dbDefs.h"
#include "dbRecType.h"
#include "dbRecords.h"
#include "dbCommon.h"
#include "dbRecDes.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "serverControl.h"
#include "restable.h"
#include "pvd.h"
#include "tsDefs.h"
#include "cadef.h"
#include "rbMgt.h"
#include "traverseDb.h"


struct alarmList
    {
    struct dbCommon	*recHdr;
    TS_STAMP		alarmTime;
    struct fldDes	*fldDes;
    int			recType;
    };
static timeCompare();

#define	TREND_THROTTLE	12	/* Default of 5 trend points per second */
#define ALARM_THROTTLE	6	/* Default of 10 alarms per second */
#define SIX_MINUTES	(6*60*60) /* 60 ticks/sec * 60 sec/min * 6 min */

int	trendThrottle = TREND_THROTTLE;
int	alarmThrottle = ALARM_THROTTLE;
int	traverseCount = 1;
static int firstTime1 = 1;	       	/* these are used to gate the logMsg delay notification */
static int firstTime2 = 1;

SEM_ID	traverseSem;

/******
	External functions & data
******/

extern struct fldDes **valList;		/* Defined in dbserver.c */
extern int startTime;			/* defined in dbserver.c */
int traverseDelay = SIX_MINUTES;	/* this is global so it can be tuned in the startup.cmd file */

/***
	New stuff for trending, some point types can trend more than
	just the VAL field
	ems 7/14/98
***/
static char *valFldNames []      = {"VAL ", NULL};
static char *biFldNames []       = {"VAL ", "TCNT", NULL};
static char *fanoutFldNames []   = {"VAL ", "SELM", NULL};
static char *pidFldNames []      = {"VAL ", "ASTP", "MVAL", NULL};
static char *selFldNames []      = {"VAL ", "SELM", NULL};
static char *pumpFldNames []     = {"VAL ", "ACRT", NULL};
static char *accumFldNames []    = {
    "VAL ", "FSMP", "FTOT", "FAVG", "FMIN", "FMAX", "RSMP", "RTOT", "RAVG", "RMIN", "RMAX", NULL};
static char *deadtimeFldNames [] = {"VAL ", "CAVE", "CMAX", "CMIN", "CTOT", NULL};
static char *udbFldNames []       = {
    "VAL ", "FP01", "FP02", "FP03", "FP04", "FP05", "FP06", "FP07", "FP08", 
    "BL01", "BL02", "BL03", "BL04", "BL05", "BL06", "BL07", "BL08", "NCMD", NULL};
static char *patFldNames []       = {"VAL ", "ASTP", NULL};

static struct fldDes ***fldList;
static void buildTrendFldList (void);
static int  firstCall = 1;


void traverseDb(opCode, 
		arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10)
int	opCode;
int	arg2;
int	arg3;
int	arg4;
int	arg5;
int	arg6;
int	arg7;
int	arg8;
int	arg9;
int	arg10;
{
    short		i;		/* Indexes point type */
    CCB			*pCCB;
    int			numPoints = 0;
    int			timePoints = 0;
    struct recLoc	*recLoc;
    struct alarmList	*timeVec = 0;
    struct alarmList	*currVec = 0;

    /*
      The following delay loop is used to hold off any database traverses for 
      6 minutes after the completion of a database download. startTime is defined
      in dbserver.c and is set immediatelyafter a database download.

      GCT 2-Oct-1998
    */
    if(firstTime1 > 0)
      {
	firstTime1--;
	logMsg("Waiting %d minutes to begin traverseDb\n",traverseDelay/3600,0,0,0,0,0); 
      }
    while((tickGet() - startTime) < traverseDelay)
      {
	taskDelay(60);
      }
    if(firstTime2 > 0)
      {
	firstTime2--;
	logMsg("traverseDb enabled after %d min. delay\n",traverseDelay/3600,0,0,0,0,0); 
      }
    /*
    Create the counting semaphore if necessary.
    */

    taskLock();
    if (!traverseSem)
	traverseSem = semCCreate(SEM_Q_PRIORITY, traverseCount);
    taskUnlock();

    /*
    Make sure that there are not too many traversals running at once.
    */

    semTake(traverseSem, WAIT_FOREVER);
    pCCB = (CCB *) arg2;
    if (opCode == 1)
	{
	if (firstCall)
	{
	    firstCall = 0;
	    buildTrendFldList ();
	}

	/*
	Wait 2 minutes if this is a trend traversal. The reason for this is to make
	sure that the drivers have a chance to grab the resource in case this is the
	first trend traversal after a boot.
	*/

	taskDelay(7200);

	logMsg("TREND traverseDb started for wkstn %s\n",
		(int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
	logMsg("Started at: ",0,0,0,0,0,0);
	date();
	}

    else if (opCode == 2)
	{
	logMsg("ALARM traverseDb started for wkstn %s\n", 
		(int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
	logMsg("Started at: ",0,0,0,0,0,0);
	date();

	/*
	Count the number of points in the database and allocate a vector
	to store the address of the record and its alarm time. We'll use
	that for sorting.
	*/

	for (i = 0; i < dbRecords -> number; i++)
	    {
	    if (!(recLoc = dbRecords -> papRecLoc[i]))
		continue;	/* No points ever defined this type */
	    numPoints += recLoc -> no_records;
	    }

	if (numPoints)
	    {
	    timeVec = (struct alarmList *) calloc(numPoints, sizeof(struct alarmList));
	    currVec = timeVec;
	    }
	}

    /*
    Traverse the database.
    */

    for (i = 0; i < dbRecords -> number; i++)
	{
	struct recChain		*chain;
	struct fldDes		*valFld;
	struct fldDes		**flds;
	int			j;

	if (!(recLoc = dbRecords -> papRecLoc[i]))
	    continue;		/* No points ever defined this type */

	valFld = valList[i];
	if (opCode == 2)
	    {

	    /*
	    For alarm "catch up," we want the last alarm value, not
	    the current value!
	    */

	    valFld = (struct fldDes *) pvdGetFld(i, "LALM");
	    if (!valFld)
		{
		char	*pntTypName;

		pntTypName = GET_PRECTYPE(i);
		if ((strcmp(pntTypName, "pump") != 0) &&
			(strcmp(pntTypName, "valve") != 0))
		    continue;
		valFld = valList[i];
		}
	    }

	for (chain = (struct recChain *) (recLoc -> pFirst);
		chain; chain = chain -> next)
	    {
	    short		ownsResource;
	    struct dbCommon	*recHdr;

	    if (!(recHdr = (struct dbCommon *) (chain -> pRecord)))
		continue;	/* No record allocated */

	    if (!(recHdr -> name[0]))
		continue;	/* Deleted point */

	    resGetCaAnswer(recHdr -> radr, &ownsResource);

	    if (opCode == 1)
		{

		/*
		Send trend data to the client specified in arg2 if we own the
		resource and the point is enabled for trending.
		*/

		if (!(pCCB -> trendHandler))
		    {
		    logMsg("TREND traverseDB terminated because %s is no longer a trend handler\n",
			    (int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
		    pCCB -> trendTravPending = 0;
		    semGive(traverseSem);
		    exit(1);
		    }
		if (!ownsResource)
		    continue;

		if (recHdr -> tenb)
		    {
		    flds = fldList[i];
		    for (j = 0; flds[j] != NULL; j++)
			{
			caRbSinglePut(i, recHdr, flds[j], DBE_LOG, (CCB *)arg2);

			/*
			Watch out for zero delays. They have the potential to hang you for
			a clock tick even though that isn't what you really want.
			*/

			if (trendThrottle)
			    taskDelay(trendThrottle);
			}
		    }
		}

	    else if (opCode == 2)
		{

		/*
		Put the address of the record and the last alarm time into the vector.
		This will get sorted at the end of the loop.
		*/

		if (!ownsResource)
		    continue;
		timePoints++;
		if (timePoints > numPoints)
		    {
		    logMsg("traverseDb: Points found is greater than points defined!\n"
					,0,0,0,0,0,0);
		    semGive(traverseSem);
		    exit (0);
		    }
		currVec -> recHdr = recHdr;
		currVec -> alarmTime.secPastEpoch = recHdr -> latm.secPastEpoch;
		currVec -> alarmTime.nsec = recHdr -> latm.nsec;
		currVec -> fldDes = valFld;
		currVec -> recType = i;
		currVec++;
		}

	    /*
	    Add else if... clauses here for additional features
	    */

	    }
	}

    if (opCode == 1)
	{
	logMsg("TREND traverseDb for workstation %s normal completion.\n",
		(int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
	pCCB -> trendTravPending = 0;
	}
    else if (opCode == 2)
	{
	if (timePoints)
	    {
	    qsort(timeVec, timePoints, sizeof(struct alarmList), timeCompare);
	    for (currVec = timeVec; timePoints; currVec++, timePoints--)
		{
		/*
		Send an alarm message to the client identified in pCCB.
		*/

		if (!(pCCB -> alarmHandler))
		    {
		    logMsg("ALARM traverseDB terminated because %s is no longer an alarm handler\n",
			    (int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
		    free (timeVec);
		    pCCB -> alarmTravPending = 0;
		    semGive(traverseSem);
		    exit(1);
		    }
		if (currVec -> recHdr -> algp[0] == 0)
		    continue;
		caRbSinglePut(currVec -> recType, currVec -> recHdr, currVec -> fldDes, DBE_ALARM, pCCB);

		/*
		Zero delays are ambiguous and certainly not necessary. Be defensive.
		*/

		if (alarmThrottle)
		    taskDelay(alarmThrottle);
		}
	    }

	if (timeVec)
	    free(timeVec);

	logMsg("ALARM traverseDb for workstation %s normal completion.\n",
		(int)INET_NTOA(pCCB -> clientInet),0,0,0,0,0);
	pCCB -> alarmTravPending = 0;
	}
    semGive(traverseSem);
    exit(1);
}

/****************************************************************/
/*								*/
/*		SET TREND THROTTLING RATE			*/
/*								*/
/****************************************************************/

int setTrendThrottle(ticks)
int	ticks;
{
    printf("Trend throttle changed from %d to %d ticks.\n",
	    trendThrottle, ticks);
    trendThrottle = ticks;
    return (0);
}


/****************************************************************/
/*								*/
/*		SET ALARM THROTTLING RATE			*/
/*								*/
/****************************************************************/

int setAlarmThrottle(ticks)
int	ticks;
{
    printf("Alarm throttle changed from %d to %d ticks.\n",
	    alarmThrottle, ticks);
    alarmThrottle = ticks;
    return (0);
}


/****************************************************************/
/*								*/
/*			COMPARE TIME STAMPS			*/
/*								*/
/*	For sorting by time, the record with the later time 	*/
/*	is "less" and we should return -1.			*/
/*								*/
/****************************************************************/

static int timeCompare(rec1, rec2)
struct alarmList	*rec1;
struct alarmList	*rec2;
{
    if (rec1 -> alarmTime.secPastEpoch > rec2 -> alarmTime.secPastEpoch)

	/*
	Record one later than record 2.
	*/

	return (-1);

    else if (rec1 -> alarmTime.secPastEpoch < rec2 -> alarmTime.secPastEpoch)

	/*
	Record one earlier than record 2.
	*/

	return (+1);

    else if (rec1 -> alarmTime.nsec > rec2 -> alarmTime.nsec)

	/*
	Record one later than record 2.
	*/

	return (-1);

    else if (rec1 -> alarmTime.nsec < rec2 -> alarmTime.nsec)

	/*
	Record one earlier than record 2.
	*/

	return (+1);

    else

	/*
	A tie.
	*/

	return (0);
}

/****************************************************************/
/*								*/
/* utility function						*/
/*								*/
/****************************************************************/

int setTraverseCount(newCount)
int	newCount;
{
    printf("Traversal count changed from %d to %d\n", traverseCount, newCount);
    traverseCount = newCount;
    return (0);
}

/****************************************************************/
/*								*/
/* buildTrendFldList  function					*/
/*	used for sending trends msgs				*/
/*								*/
/****************************************************************/

static void buildTrendFldList(void)
{
    short	  recIndex;
    int		  i;
    int		  size;
    char	  *recType;
    char	  **fldNames;
    struct fldDes **flds;

    fldList = (struct fldDes ***)(calloc (dbRecords -> number, sizeof (struct fldDes *)));

    for (recIndex = 0; recIndex < dbRecords -> number; recIndex++)
    {
	recType = GET_PRECTYPE (recIndex);
	if (recType == NULL)
	{
	    continue;
	}
	if (strcmp (recType, "bi") == 0)
	{
	    fldNames = biFldNames;
	    size = sizeof (biFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "accum") == 0)
	{
	    fldNames = accumFldNames;
	    size = sizeof (accumFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "deadtime") == 0)
	{
	    fldNames = deadtimeFldNames;
	    size = sizeof (deadtimeFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "pid") == 0)
	{
	    fldNames = pidFldNames;
	    size = sizeof (pidFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "sel") == 0)
	{
	    fldNames = selFldNames;
	    size = sizeof (selFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "fanout") == 0)
	{
	    fldNames = fanoutFldNames;
	    size = sizeof (fanoutFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "udb") == 0)
	{
	    fldNames = udbFldNames;
	    size = sizeof (udbFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "pat") == 0)
	{
	    fldNames = patFldNames;
	    size = sizeof (patFldNames) / sizeof (char *);
	}
	else if (strcmp (recType, "pump") == 0)
	{
	    fldNames = pumpFldNames;
	    size = sizeof (pumpFldNames) / sizeof (char *);
	}
	else
	{
	    fldNames = valFldNames;
	    size = sizeof (valFldNames) / sizeof (char *);
	}
	flds = (struct fldDes **)(calloc (size, sizeof (struct fldDes *)));
	for (i = 0; fldNames[i] != NULL; i++)
	{
	    flds[i] = pvdGetFld (recIndex, fldNames[i]);
	}
	fldList[recIndex] = flds;
    }
}
