/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		REDUNDANT DRIVER DATA UPDATE TASK		*/
/*								*/
/*	This task provides a redundant update mechanism for	*/
/*	drivers. For a given resource, if we are master then	*/
/*	we periodically send over a buffer full of data. If	*/
/*	we are slave, we read the buffer and call a routine	*/
/*	to process that data.					*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <sockLib.h>
#include <in.h>
#include <taskLib.h>
#include <ioLib.h>
#include <semLib.h>
#include <stdio.h>

#include "restable.h"

/****** TUNING/SIZING PARAMETERS ******/

#define	REDUPDT_PRI	110		/* Just below periodic scan */
#define REDUPDT_STACK	4096

/****** FUNCTION DECLARATIONS ******/

void drvRedUpdt (RES_TABLE *redundRes, int sockFd, short port, char *buffPtr, int *buffSize,
		 void (*callBack)(), void *callBackArg, int delay, SEM_ID redUpdtSem);

int drvRedTake (SEM_ID redUpdtSem);
int drvRedGive (SEM_ID redUpdtSem);


/****************************************************************/
/*								*/
/*			INITIALIZE				*/
/*								*/
/****************************************************************/

int
initRedDriverData (resName, buffPtr, buffSize, callBack, callBackArg, delay, port, redUpdtSem)
char		*resName;		/* Resource name */
char		*buffPtr;		/* User buffer */
int		*buffSize;		/* Buffer length */
void		(*callBack)();		/* User's processing routine */
void		*callBackArg;		/* Argument to user's routine */
int		delay;			/* Ticks between cycles */
short		port;			/* UDP port number for socket */
SEM_ID		*redUpdtSem;		/* ME semaphore to protect buffer */
{
    RES_TABLE		*redundRes;
    int			sockFd;
    struct sockaddr_in	portAddr;

    /*
    First, translate the resource name to an entry in the resource table.
    */

    if (resGetByName(resName, &redundRes) != RES_OK)
	{
	printErr("initRedDriverData can't find resource name %s.\n");
	return (ERROR);
	}

    /*
    Now create a UDP socket and name it.
    */

    sockFd = socket(PF_INET, SOCK_DGRAM, 0);
    if (sockFd == ERROR)
	{
	perror("initRedDriverData");
	return (ERROR);
	}
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    portAddr.sin_port = htons(port);
    if (bind(sockFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == ERROR)
	{
	perror("initRedDriverData");
	return (ERROR);
	}

    /*
    Create the mutual exclusion semaphore to protect the buffer.
    */

    if ((*redUpdtSem = semMCreate(SEM_Q_PRIORITY | SEM_INVERSION_SAFE)) == NULL)
	return (ERROR);

    /*
    All preparation done. Now, spawn the redundant update task.
    */

    if (taskSpawn(resName, REDUPDT_PRI, VX_FP_TASK, REDUPDT_STACK, (FUNCPTR)drvRedUpdt,
	    (int)redundRes, sockFd, port, (int)buffPtr, (int)buffSize, (int)callBack, (int)callBackArg, delay,
	    (int)*redUpdtSem, 0) == ERROR)
	{
	perror("initRedDriverData");
	return (ERROR);
	}
    return (!ERROR);
}


/****************************************************************/
/*								*/
/*			REDUNDANT UPDATE TASK			*/
/*								*/
/****************************************************************/

void drvRedUpdt(redundRes, sockFd, port, buffPtr, buffSize, callBack, callBackArg, delay, redUpdtSem)
RES_TABLE	*redundRes;
int		sockFd;
short		port;
char		*buffPtr;
int		*buffSize;
void		(*callBack)();
void		*callBackArg;
int		delay;
SEM_ID		redUpdtSem;		/* ME semaphore to protect buffer */
{
    struct sockaddr_in	portAddr;
    unsigned short	currState;
    int			bytesGot;
    int			addrSize;
    char		localBuff[1024];

    for (;;)
	{
	taskDelay(delay);

	/*
	Look at the state in the resource table to determine if we are sending 
	or receiving.
	*/

	currState = redundRes -> localState;
	currState &= ~RES_CHANGED;
	if ((currState == RES_MSTR) || (currState == RES_SLVWANT))
	    {

	    /*
	    Master, we are sending
	    */

	    if (redundRes -> partnerInet.s_addr)
		{
		drvRedTake(redUpdtSem);
		portAddr.sin_family = AF_INET;
		portAddr.sin_addr = redundRes -> partnerInet;
		portAddr.sin_port = htons(port);
		sendto(sockFd, buffPtr, *buffSize, 0, (struct sockaddr *) &portAddr, sizeof(portAddr));
		drvRedGive(redUpdtSem);
		}
	    }
	
	else if (currState == RES_SLV)
	    {

	    /*
	    Slave, we are receiving. See if there's anything there.
	    */

	    for (;;)
		{
	        ioctl(sockFd, FIONREAD, (int)&bytesGot);
	        if (bytesGot)
		    {
	            addrSize = sizeof(portAddr);
	            bytesGot = recvfrom(sockFd, localBuff, sizeof(localBuff), 0, 
			(struct sockaddr *) &portAddr, &addrSize);
		    if (bytesGot > 0)
		        (*callBack)(callBackArg, localBuff, bytesGot);
		    continue;
		    }
		break;
		}
	    continue;
	    }

	/*
	Either master or not slave. In this case, drain any inputs we may have out
	of the socket and throw all the data on the floor. Use our own buffer for
	this because we don't want to trash what the master has been doing.
	*/

	for (;;)
	    {
	    ioctl(sockFd, FIONREAD, (int)&bytesGot);
	    if (!bytesGot)
		break;
	    addrSize = sizeof(portAddr);
	    recvfrom(sockFd, localBuff, sizeof(localBuff), 0, (struct sockaddr *) &portAddr,
		    &addrSize);
	    }
	}
}

/****************************************************************/
/*								*/
/*		MAKE THE SHARED BUFFER AREA PRIVATE		*/
/*								*/
/****************************************************************/


int
drvRedTake(redUpdtSem)
SEM_ID	redUpdtSem;
{
    semTake(redUpdtSem, WAIT_FOREVER);
    return (!ERROR);
}

/****************************************************************/
/*								*/
/*		RELEASE THE SHARED BUFFER AREA			*/
/*								*/
/****************************************************************/

int
drvRedGive(redUpdtSem)
SEM_ID	redUpdtSem;
{
    semGive(redUpdtSem);
    return (!ERROR);
}
