/* HiaserialSio.c - HIASERIAL serial driver */

/* Copyright 1984-1995 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01g,23may97,db	 added hardware options and modem control(SPR #7542).
                 fixed bugs reported in SPRs 9404, 9223.
01f,12oct95,dgf  added bauds > 38k
01e,03aug95,myz  fixed the warning messages
01d,12jul95,myz  fixed the baud rate problem.
01c,20jun95,ms	 fixed comments for mangen.
01b,15jun95,ms   updated for new driver structure
01a,15mar95,myz  written (using HiaserialSerial.c + the VxMon polled driver).
*/

/*
DESCRIPTION
This is the driver for the Intel 8250 UART Chip used on the PC 386.
It uses the SCCs in asynchronous mode only.

USAGE
An HIASERIAL_CHAN structure is used to describe the chip.
The BSP's sysHwInit() routine typically calls sysSerialHwInit()
which initializes all the register values in the HIASERIAL_CHAN structure
(except the SIO_DRV_FUNCS) before calling HiaserialHrdInit().
The BSP's sysHwInit2() routine typically calls sysSerialHwInit2() which
connects the chips interrupt handler (HiaserialInt) via intConnect().

IOCTL FUNCTIONS
This driver responds to all the same ioctl() codes as a normal serial driver;
for more information, see the comments in sioLib.h.  As initialized, the
available baud rates are 110, 300, 600, 1200, 2400, 4800, 9600, 19200, and
38400.

This driver handles setting of hardware options such as parity(odd, even) and
number of data bits(5, 6, 7, 8). Hardware flow control is provided with the
handshakes RTS/CTS. The function HUPCL(hang up on last close) is available.

INCLUDE FILES: drv/sio/HiaserialSio.h
*/

#include "vxWorks.h"
#include "iv.h"
#include "intLib.h"
#include "errnoLib.h"
#include "Hiaserial.h"
#include "pc.h"

/* defines */

#ifndef HIASERIAL_DEFAULT_BAUD
#       define  HIASERIAL_DEFAULT_BAUD      9600
#endif

#ifndef SIO_HUP
#   define SIO_OPEN	0x100A
#   define SIO_HUP	0x100B
#endif

/* locals */

/* baudTable is a table of the available baud rates, and the values to write
 * to the UART's baud rate divisor {high, low} register. the formula is
 * 1843200(source) / (16 * baudrate)
 */

static BAUD baudTable [] =
    {
    {50, 2304}, {75, 1536}, {110, 1047}, {134, 857}, {150, 768},
    {300, 384}, {600, 192}, {1200, 96}, {2000, 58}, {2400, 48},
    {3600,32}, {4800, 24}, {7200,16}, {9600, 12}, {19200, 6}, {38400, 3},
    {57600, 2}, {115200, 1}
    };

static SIO_DRV_FUNCS HiaserialSioDrvFuncs;

/* forward declarations */

LOCAL 	void 	HiaserialInitChannel (HIASERIAL_CHAN *);
LOCAL 	int 	HiaserialIoctl(HIASERIAL_CHAN *, int, int);
LOCAL 	int 	HiaserialStartup(HIASERIAL_CHAN *pChan);
LOCAL 	int 	HiaserialPRxChar (HIASERIAL_CHAN *pChan, char *ch);
LOCAL 	int 	HiaserialPTxChar(HIASERIAL_CHAN *pChan, char ch);
LOCAL 	STATUS 	HiaserialOptsSet (HIASERIAL_CHAN *, UINT);
LOCAL 	STATUS  HiaserialModeSet (HIASERIAL_CHAN *, UINT);
LOCAL   STATUS  HiaserialBaudSet (HIASERIAL_CHAN *, UINT);

/******************************************************************************
*
* HiaserialCallbackInstall - install ISR callbacks to get put chars.
*
* This routine installs the callback functions for the driver
*
* RETURNS: OK on success or ENOSYS on unsupported callback type.
*/ 

static int HiaserialCallbackInstall
    (
    SIO_CHAN *  pSioChan,
    int         callbackType,
    STATUS      (*callback)(),
    void *      callbackArg
    )
    {
    HIASERIAL_CHAN * pChan = (HIASERIAL_CHAN *)pSioChan;

    switch (callbackType)
        {
        case SIO_CALLBACK_GET_TX_CHAR:
            pChan->getTxChar    = callback;
            pChan->getTxArg     = callbackArg;
            return (OK);
        case SIO_CALLBACK_PUT_RCV_CHAR:
            pChan->putRcvChar   = callback;
            pChan->putRcvArg    = callbackArg;
            return (OK);
        default:
            return (ENOSYS);
        }
    }


/******************************************************************************
*
* HiaserialHrdInit - initialize the chip
*
* This routine is called to reset the chip in a quiescent state.
* RETURNS: N/A
*/

void HiaserialHrdInit
    (
    HIASERIAL_CHAN *  pChan		/* pointer to device */
    )
    {
    (*pChan->outByte) (pChan->ier,0x0);  /* disable interrupts */

    if (HiaserialSioDrvFuncs.ioctl == NULL)
	{
	HiaserialSioDrvFuncs.ioctl		= (int (*)())HiaserialIoctl;
	HiaserialSioDrvFuncs.txStartup	= (int (*)())HiaserialStartup;
	HiaserialSioDrvFuncs.callbackInstall = HiaserialCallbackInstall;
	HiaserialSioDrvFuncs.pollInput	= (int (*)())HiaserialPRxChar;
	HiaserialSioDrvFuncs.pollOutput	= (int (*)(SIO_CHAN *,char))HiaserialPTxChar;
	}

    pChan->pDrvFuncs = &HiaserialSioDrvFuncs;
    }

/*******************************************************************************
*
* HiaserialInitChannel  - initialize a single channel
*/

static void HiaserialInitChannel
    (
    HIASERIAL_CHAN *  pChan		/* pointer to device */
    )
    {
    int oldLevel;

    oldLevel = intLock ();

    /* set channel baud rate */

    (void) HiaserialBaudSet(pChan, HIASERIAL_DEFAULT_BAUD);

    /* 8 data bits, 1 stop bit, no parity */

    (*pChan->outByte) (pChan->lcr, (HIASERIAL_LCR_CS8 | HIASERIAL_LCR_1_STB));

    (*pChan->outByte) (pChan->mdc, (HIASERIAL_MCR_RTS | HIASERIAL_MCR_DTR | \
    					HIASERIAL_MCR_OUT2));

    (*pChan->inByte) (pChan->data);	/* clear the port */

    /* enables interrupts  */

    (*pChan->outByte) (pChan->ier, HIASERIAL_IER_RXRDY);

    pChan->options = (CLOCAL | CREAD | CS8);

    intUnlock (oldLevel);
    }

/*******************************************************************************
*
* HiaserialHup - hang up the modem control lines 
*
* Resets the RTS and DTR signals.
*
* RETURNS: OK
*/

LOCAL STATUS HiaserialHup
    (
    HIASERIAL_CHAN * pChan 		/* pointer to channel */
    )
    {
    FAST int     oldlevel;	/* current interrupt level mask */

    oldlevel = intLock ();

    (*pChan->outByte) (pChan->mdc, HIASERIAL_MCR_OUT2);

    intUnlock (oldlevel);

    return (OK);

    }    

/*******************************************************************************
*
* HiaserialOpen - Set the modem control lines 
*
* Set the modem control lines(RTS, DTR) TRUE if not already set.  
* It also clears the receiver. 
*
* RETURNS: OK
*/

LOCAL STATUS HiaserialOpen
    (
    HIASERIAL_CHAN * pChan 		/* pointer to channel */
    )
    {
    FAST int     oldlevel;	/* current interrupt level mask */
    char mask;

    /* read modem control register */

    mask = ((*pChan->inByte) (pChan->mdc)) & (HIASERIAL_MCR_RTS | HIASERIAL_MCR_DTR);

    if (mask != (HIASERIAL_MCR_RTS | HIASERIAL_MCR_DTR)) 
    	{
    	/* RTS and DTR not set yet */

    	oldlevel = intLock ();

	/* set RTS and DTR TRUE */

    	(*pChan->outByte) (pChan->mdc, (HIASERIAL_MCR_RTS | HIASERIAL_MCR_DTR | \
    					HIASERIAL_MCR_OUT2));

    	intUnlock (oldlevel);
	}

    return (OK);
    }

/******************************************************************************
*
* HiaserialBaudSet - change baud rate for channel
*
* This routine sets the baud rate for the UART. The interrupts are disabled
* during chip access.
*
* RETURNS: OK to indicate success, otherwise ERROR is returned
*/

LOCAL STATUS  HiaserialBaudSet
    (
    HIASERIAL_CHAN *   pChan,	/* pointer to channel */
    UINT	   baud		/* requested baud rate */
    )
    {
    int   	oldlevel;
    STATUS	status;
    FAST	int	ix;
    UINT8	lcr;

    /* disable interrupts during chip access */

    oldlevel = intLock ();

    status = ERROR;

    for (ix = 0; ix < NELEMENTS (baudTable); ix++)
	{
	if (baudTable [ix].rate == baud)	/* lookup baud rate value */
	    {
	    lcr = (*pChan->inByte) (pChan->lcr);   
    	    (*pChan->outByte) (pChan->lcr, (char)(HIASERIAL_LCR_DLAB | lcr));
    	    (*pChan->outByte) (pChan->brdh, MSB (baudTable[ix].preset));
    	    (*pChan->outByte) (pChan->brdl, LSB (baudTable[ix].preset));
    	    (*pChan->outByte) (pChan->lcr, lcr);
	    status = OK;
	    break;
	    }
	}

    intUnlock(oldlevel);

    return (status);
    }

/*******************************************************************************
*
* HiaserialModeSet - change channel mode setting
*
* This driver supports both polled and interrupt modes and is capable of
* switching between modes dynamically. 
*
* If interrupt mode is desired this routine enables the channels receiver and 
* transmitter interrupts. If the modem control option is TRUE, the Tx interrupt
* is disabled if the CTS signal is FALSE. It is enabled otherwise. 
*
* If polled mode is desired the device interrupts are disabled. 
*
* RETURNS:
* Returns a status of OK if the mode was set else ERROR.
*/

LOCAL STATUS HiaserialModeSet
    (
    HIASERIAL_CHAN * pChan,		/* pointer to channel */
    UINT	newMode		/* mode requested */
    )
    {
    FAST int     oldlevel;	/* current interrupt level mask */
    char ier, mask;

    if ((newMode != SIO_MODE_POLL) && (newMode != SIO_MODE_INT))
	return (ERROR);

    oldlevel = intLock();
           
    /* only reset the channel if it hasn't done yet */

    if (!pChan->channelMode)
    	HiaserialInitChannel(pChan);

    if (newMode == SIO_MODE_POLL)
    	ier = 0x00;
    else  
	{
    	if (pChan->options & CLOCAL) 
         	ier = HIASERIAL_IER_RXRDY;
	else  
		{
    		mask = ((*pChan->inByte) (pChan->msr)) & HIASERIAL_MSR_CTS;

   		/* if the CTS is asserted enable Tx interrupt */

   		if (mask & HIASERIAL_MSR_CTS)
         		ier = (HIASERIAL_IER_TBE | HIASERIAL_IER_MSI);
		else
         		ier = (HIASERIAL_IER_MSI);

		}	
	}

    (*pChan->outByte) (pChan->ier, ier);
            
    pChan->channelMode = newMode;  
          
    intUnlock(oldlevel);

    return (OK);
   }

/*******************************************************************************
*
* HiaserialOptsSet - set the serial options
*
* Set the channel operating mode to that specified.  All sioLib options
* are supported: CLOCAL, HUPCL, CREAD, CSIZE, PARENB, and PARODD.
*
* RETURNS:
* Returns OK to indicate success, otherwise ERROR is returned
*/

LOCAL STATUS HiaserialOptsSet
    (
    HIASERIAL_CHAN * pChan,		/* pointer to channel */
    UINT options		/* new hardware options */
    )
    {
    FAST int     oldlevel;		/* current interrupt level mask */
    char lcr = 0; 
    char mcr = HIASERIAL_MCR_OUT2;
    char ier = (*pChan->inByte) (pChan->ier);
    char mask;

    if (pChan == NULL || options & 0xffffff00)
	return ERROR;

    switch (options & CSIZE)
	{
	case CS5:
	    lcr = HIASERIAL_LCR_CS5; break;
	case CS6:
	    lcr = HIASERIAL_LCR_CS6; break;
	case CS7:
	    lcr = HIASERIAL_LCR_CS7; break;
	default:
	case CS8:
	    lcr = HIASERIAL_LCR_CS8; break;
	}

    if (options & STOPB)
	lcr |= HIASERIAL_LCR_2_STB;
    else
	lcr |= HIASERIAL_LCR_1_STB;
    
    switch (options & (PARENB | PARODD))
	{
	case PARENB|PARODD:
	    lcr |= HIASERIAL_LCR_PEN; break;
	case PARENB:
	    lcr |= (HIASERIAL_LCR_PEN | HIASERIAL_LCR_EPS); break;
	default:
	case 0:
	    lcr |= HIASERIAL_LCR_PDIS; break;
	}

    (*pChan->outByte) (pChan->ier, 0x00);

    if (!(options & CLOCAL))
	{
	/* !clocal enables hardware flow control(DTR/DSR) */

	mcr |= (HIASERIAL_MCR_DTR | HIASERIAL_MCR_RTS);
	ier |= HIASERIAL_IER_MSI;    /* enable modem status interrupt */

    	mask = ((*pChan->inByte) (pChan->msr)) & HIASERIAL_MSR_CTS;

   	/* if the CTS is asserted enable Tx interrupt */

   	if (mask & HIASERIAL_MSR_CTS)
        	ier |= HIASERIAL_IER_TBE;
	else
    		ier &= (~HIASERIAL_IER_TBE); 

	}
    else 
        ier &= ~HIASERIAL_IER_MSI; /* disable modem status interrupt */ 

    oldlevel = intLock ();

    (*pChan->outByte) (pChan->lcr, lcr);
    (*pChan->outByte) (pChan->mdc, mcr);

    /* now clear the port */

    (*pChan->inByte) (pChan->data);

    if (options & CREAD)  
	ier |= HIASERIAL_IER_RXRDY;

    if (pChan->channelMode == SIO_MODE_INT)
	{
    	(*pChan->outByte) (pChan->ier, ier);
        }

    intUnlock (oldlevel);

    pChan->options = options;

    return OK;
    }

/*******************************************************************************
*
* HiaserialIoctl - special device control
*
* Includes commands to get/set baud rate, mode(INT, POLL), hardware options(
* parity, number of data bits), and modem control(RTS/CTS and DTR/DSR).
* The ioctl commands SIO_HUP and SIO_OPEN are implemented to provide the 
* HUPCL( hang up on last close) function.
*
* RETURNS: OK on success, EIO on device error, ENOSYS on unsupported
*          request.
*/

static int HiaserialIoctl
    (
    HIASERIAL_CHAN *  pChan,	/* device to control */
    int request,		/* request code */
    int arg			/* some argument */
    )
    {
    int ix;
    int baudH;
    int baudL;
    int oldlevel;
    UINT8 lcr,msr,mdc;
    int status = OK;

    switch (request)
	{
	case SIO_BAUD_SET:
	    status = (HiaserialBaudSet(pChan, arg) == OK) ? OK : EIO;
	    break;

        case SIO_BAUD_GET:
            
            oldlevel = intLock();

            status = EIO;

            lcr = (*pChan->inByte) (pChan->lcr);
            (*pChan->outByte) (pChan->lcr, (char)(HIASERIAL_LCR_DLAB | lcr));
            baudH = (*pChan->inByte)(pChan->brdh);
            baudL = (*pChan->inByte)(pChan->brdl);
            (*pChan->outByte) (pChan->lcr, lcr);

            for (ix = 0; ix < NELEMENTS (baudTable); ix++)
                {
                if ( baudH  == MSB (baudTable[ix].preset) &&
                     baudL  == LSB (baudTable[ix].preset) )
                    {
                    *(int *)arg = baudTable[ix].rate;
                    status = OK;
                    }
                }

            intUnlock(oldlevel);
            break;


	case SIO_MODE_SET:
	    status = (HiaserialModeSet (pChan, arg) == OK) ? OK : EIO;
	    break;

        case SIO_MODE_GET:
            *(int *)arg = pChan->channelMode;
            return (OK);

        case SIO_AVAIL_MODES_GET:
            *(int *)arg = SIO_MODE_INT | SIO_MODE_POLL;
            return (OK);

        case SIO_HW_OPTS_SET:
    	    status = (HiaserialOptsSet (pChan, arg) == OK) ? OK : EIO;
    	    break;

        case SIO_HW_OPTS_GET:
            *(int *)arg = pChan->options;
            break;

        case SIO_HUP:
            /* check if hupcl option is enabled */

    	    if (pChan->options & HUPCL) 
	    	status = HiaserialHup (pChan);
            break;
	
	case SIO_OPEN:
            /* check if hupcl option is enabled */

    	    if (pChan->options & HUPCL) 
	    	status = HiaserialOpen (pChan);
	    break;
	case SIO_SET_RTS:
            mdc = (*pChan->inByte) (pChan->mdc);
	    mdc |= HIASERIAL_MCR_RTS;
            (*pChan->outByte) (pChan->mdc, mdc);
	    break;
	case SIO_SET_DTR:
            mdc = (*pChan->inByte) (pChan->mdc);
	    mdc |= HIASERIAL_MCR_DTR;
            (*pChan->outByte) (pChan->mdc, mdc);
	    break;
	case SIO_CLR_RTS:
            mdc = (*pChan->inByte) (pChan->mdc);
	    mdc &= (~HIASERIAL_MCR_RTS);
            (*pChan->outByte) (pChan->mdc, mdc);
	    break;
	case SIO_CLR_DTR:
            mdc = (*pChan->inByte) (pChan->mdc);
	    mdc &= (~HIASERIAL_MCR_DTR);
            (*pChan->outByte) (pChan->mdc, mdc);
	    break;
	case SIO_GET_CTS:
@!
            msr = (*pChan->inByte) (pChan->msr);
            *(int *)arg = msr & HIASERIAL_MSR_ICTS;
	    break;
	case SIO_GET_DSR:
            msr = (*pChan->inByte) (pChan->msr);
            *(int *)arg = msr & HIASERIAL_MSR_IDSR;
	    break;
	case SIO_GET_DCD:
            msr = (*pChan->inByte) (pChan->msr);
            *(int *)arg = msr & HIASERIAL_MSR_ICD;
	    break;
	case SIO_GET-RI:
            msr = (*pChan->inByte) (pChan->msr);
            *(int *)arg = msr & HIASERIAL_MSR_IRI;
	    break;

	default:
	    status = ENOSYS;
	    break;
	}

    return (status);
    }

/*******************************************************************************
*
* HiaserialInt - handle a receiver/transmitter interrupt
*
* This routine handles four sources of interrupts from the UART. They are
* prioritized in the following order by the Interrupt Identification Register:
* Receiver Line Status, Received Data Ready, Transmit Holding Register Empty
* and Modem Status.
*
* If there is another character to be transmitted, it sends it.  If
* not, or if a device has never been created for this channel, just
* disable the interrupt.
* disable the interrupt.
* When a modem status interrupt occurs, the transmit interrupt is enabled if
* the CTS signal is TRUE.
*
* RETURNS: N/A
*/

void HiaserialInt
    (
    HIASERIAL_CHAN  *  inpChan 
    )
    {

    char chanid;
    int  ix = 0;
    HIASERIAL_CHAN  *  pChan; 

    if (pChan->int_vec == COM3TO10_INT_VEC) {
	for (ix = 0; (ix < 128); ix++) {

	    chanid = (*pChan->inByte) (COM3TO10CHIPID);
	    if (chanid == 0) 
		return;
	    pChan = &HiaserialChan[3+chanid];
	    HiaserialIntChan(pChan);
	}
    } else {
	HiaserialIntChan(inpChan);
    }

/*******************************************************************************
*
* HiaserialIntChan - handle a receiver/transmitter interrupt
*
* This routine handles four sources of interrupts from the UART. They are
* prioritized in the following order by the Interrupt Identification Register:
* Receiver Line Status, Received Data Ready, Transmit Holding Register Empty
* and Modem Status.
*
* If there is another character to be transmitted, it sends it.  If
* not, or if a device has never been created for this channel, just
* disable the interrupt.
* disable the interrupt.
* When a modem status interrupt occurs, the transmit interrupt is enabled if
* the CTS signal is TRUE.
*
* RETURNS: N/A
*/

void HiaserialIntChan
    (
    HIASERIAL_CHAN  *  Chan 
    )
    {

    char outChar;
    char interruptID;
    char lineStatus;
    int  ix = 0;
    char ier;
 
    interruptID = (*pChan->inByte) ( pChan->iid );

    do {

	interruptID &= HIASERIAL_IIR_MASK;

        if (interruptID == HIASERIAL_IIR_SEOB)
            lineStatus = (*pChan->inByte) (pChan->lst);
        else if (interruptID == HIASERIAL_IIR_RBRF)
	    {
            if (pChan->putRcvChar != NULL)
	        (*pChan->putRcvChar)( pChan->putRcvArg,
                               (*pChan->inByte) (pChan->data) );
	    else
	        (*pChan->inByte) (pChan->data);
	    }
        else if (interruptID == HIASERIAL_IIR_THRE) 
	    {
            if ((pChan->getTxChar != NULL) && 
               (*pChan->getTxChar) (pChan->getTxArg, &outChar) == OK )
	        (*pChan->outByte) (pChan->data, outChar);
            else
	        (*pChan->outByte) (pChan->ier, 0x01);
	    }
	else if (interruptID == HIASERIAL_IIR_MSTAT) /* modem status register */
	   {
    	   char msr = (*(pChan)->inByte) (pChan->msr);
           ier = (HIASERIAL_IER_RXRDY | HIASERIAL_IER_MSI);

	   /* if CTS is asserted by modem, enable tx interrupt */

	   if ((msr & HIASERIAL_MSR_CTS) && (msr & HIASERIAL_MSR_DCTS)) 
	        (*pChan->outByte) (pChan->ier, (HIASERIAL_IER_TBE | ier));
           else
	        (*pChan->outByte) (pChan->ier, ier);
	   }

        interruptID = (*pChan->inByte) (pChan->iid);

	} while (((interruptID & 0x01) == 0) && (ix++ < 10));
    }

/*******************************************************************************
*
* HiaserialStartup - transmitter startup routine
*
* Call interrupt level character output routine and enable interrupt if it is
* in interrupt mode with no hardware flow control.
* If the option for hardware flow control is enabled and CTS is set TRUE,
* then the Tx interrupt is enabled.
*
* RETURNS: OK
*/

LOCAL int HiaserialStartup
    (
    HIASERIAL_CHAN *  pChan		/* tty device to start up */
    )
    {
    char  ier = HIASERIAL_IER_RXRDY;
    char  mask;

    if (pChan->channelMode == SIO_MODE_INT)
	{
	if (pChan->options & CLOCAL)
		{
		/* No modem control */

		ier |= HIASERIAL_IER_TBE;
		}
	else
		{
    		mask = ((*pChan->inByte) (pChan->msr)) & HIASERIAL_MSR_CTS;

   		/* if the CTS is asserted enable Tx interrupt */

   		if (mask & HIASERIAL_MSR_CTS)
         		ier = (HIASERIAL_IER_TBE | HIASERIAL_IER_MSI);
		else
         		ier = (HIASERIAL_IER_MSI);
		}

    	(*pChan->outByte) (pChan->ier, ier);
    	}

    return (OK);
    }
/******************************************************************************
*
* HiaserialPRxChar - poll the device for input.
*
* RETURNS: OK if a character arrived, EIO on device error, EAGAIN
* if the input buffer if empty.
*/

LOCAL int HiaserialPRxChar
    (
    HIASERIAL_CHAN *    pChan,	/* pointer to channel */
    char *          pChar 	/* pointer to char */
    )
    {

    char pollStatus = ( *(pChan)->inByte) (pChan->lst);

    if ((pollStatus & HIASERIAL_LSR_RXRDY) == 0x00)
        return (EAGAIN);

    /* got a character */

    *pChar = ( *(pChan)->inByte) (pChan->data);

    return (OK);
    }

/******************************************************************************
*
* HiaserialPTxChar - output a character in polled mode.
* 
* Checks if the transmitter is empty. If empty, a character is written to
* the data register. 
*
* If the hardware flow control is enabled the handshake signal CTS has to be
* asserted in order to send a character.
*
* RETURNS: OK if a character arrived, EIO on device error, EAGAIN
* if the output buffer if full.
*/

LOCAL int HiaserialPTxChar
    (
    HIASERIAL_CHAN *  pChan,	/* pointer to channel */
    char          outChar	/* char to send */
    )
    {
    char pollStatus = ( *(pChan)->inByte) (pChan->lst);
    char msr = (*(pChan)->inByte) (pChan->msr);

    /* is the transmitter ready to accept a character? */

    if ((pollStatus & HIASERIAL_LSR_TEMT) == 0x00)
        return (EAGAIN);

    if (!(pChan->options & CLOCAL))	 /* modem flow control ? */
    	{
    	if (msr & HIASERIAL_MSR_CTS)
    		(*(pChan)->outByte) ((pChan)->data, outChar);
	else
		return (EAGAIN);
	}
    else
    	(*(pChan)->outByte) ((pChan)->data, outChar);

    return (OK);
    }
