/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		ROUTINES USED TO ISOLATE MEMORY LEAKS		*/
/*								*/
/*	These routines are used as replacements for the		*/
/*	standard routines malloc(), calloc(), and free().	*/
/*	These routines are basically the same but they		*/
/*	take an additional argument which is an ascii		*/
/*	string to be used as an indentifier. This string	*/
/*	should be unique for each call. A few routines are	*/
/*	also provided to traverse and report the memory		*/
/*	used in a couple of different formats.			*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#if defined(AXP)
#include <stdlib.h>
#endif
#if defined(WIN32)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define NOLEAKTRACK
#endif

struct memLeak
    {
    char		name[48];
    int			lineNumber;
    int			size;
    struct memLeak	*next;
    struct memLeak	*prev;
    };

/******
	Global pointer to the head of the chain
******/

struct memLeak	*firstMemLeak = 0;


/****************************************************************/
/*								*/
/*		TISMALLOC, EQUIVALENT TO MALLOC()		*/
/*								*/
/****************************************************************/

void *tisMalloc(id, lineNumber, nbytes)
char	*id;
int	lineNumber;
int	nbytes;
{
#if defined(NOLEAKTRACK) || defined(WIN32)
    return ((void *)malloc (nbytes));
#else
    struct memLeak	*newLeak;

    nbytes += sizeof(struct memLeak);
    if (!(newLeak = (struct memLeak *) malloc(nbytes)))
	return (newLeak);
    memset (newLeak, 0, nbytes);

#ifdef vxWorks
    taskLock();
#endif vxWorks

    newLeak -> next = firstMemLeak;
    if (newLeak -> next)
    	newLeak -> next -> prev = newLeak;
    newLeak -> prev = 0;
    firstMemLeak = newLeak;

#ifdef vxWorks
    taskUnlock();
#endif vxWorks

    strncpy(newLeak -> name, id, sizeof(newLeak -> name));
    newLeak -> lineNumber = lineNumber;
    newLeak -> size = nbytes - sizeof(struct memLeak);
    return (((char *) newLeak) + sizeof(struct memLeak));
#endif
}


/****************************************************************/
/*								*/
/*		TISCALLOC, EQUIVALENT TO CALLOC()		*/
/*								*/
/****************************************************************/

void *tisCalloc(id, lineNumber, nitems, itemSize)
char	*id;
int	lineNumber;
int	nitems;
int	itemSize;
{
#if defined(NOLEAKTRACK) || defined(WIN32)
    return ((void *)calloc (nitems, itemSize));
#else
    void	*newMem;

    if (!(newMem = tisMalloc(id, lineNumber, nitems * itemSize)))
	return (newMem);
    memset(newMem, 0, nitems * itemSize);
    return (newMem);
#endif
}


/****************************************************************/
/*								*/
/*		TISFREE, EQUIVALENT TO FREE()			*/
/*								*/
/****************************************************************/

void tisFree(addr)
char	*addr;
{
#if defined(NOLEAKTRACK) || defined(WIN32)
    free (addr);
    return;
#else
    struct memLeak	*oldLeak;
    struct memLeak	*currLeak;

    /*
    Don't bother processing a null pointer.
    */

    if (!addr)
	return;

    /*
    Figure out the address and take it out of the list
    */

    oldLeak = (struct memLeak *) (addr - sizeof(struct memLeak));

#ifdef vxWorks
    taskLock();
#endif

    currLeak = firstMemLeak;
    while (currLeak && (currLeak != oldLeak))
	currLeak = currLeak -> next;
    if (!currLeak)
	{

#ifdef vxWorks
	taskUnlock();
#if 0
	printf("Attempt to free nonallocated memory in %s\n", 
		taskName(taskIdSelf()));
#endif
#else
#if 0
	printf("Attempt to free nonallocated memory\n");
#endif
#endif

	return;
	}

    if (currLeak -> prev)
	currLeak -> prev -> next = currLeak -> next;
    else
	firstMemLeak = currLeak -> next;

    if (currLeak -> next)
	currLeak -> next -> prev = currLeak -> prev;

#ifdef vxWorks
    taskUnlock();
#endif

    free(currLeak);
#endif
}


/****************************************************************/
/*								*/
/*		TISREALLOC, EQUIVALENT TO REALLOC()		*/
/*								*/
/****************************************************************/

void *tisRealloc(id, lineNumber, addr, newSize)
char	*id;
char	*addr;
int	lineNumber;
int	newSize;
{
#if defined(NOLEAKTRACK) || defined(WIN32)
    return ((void *)realloc (addr, newSize));
#else
    struct memLeak	*oldLeak;
    struct memLeak	*currLeak;

    /*
    Figure out the address and take it out of the list
    */

    oldLeak = (struct memLeak *) (addr - sizeof(struct memLeak));

#ifdef vxWorks
    taskLock();
#endif

    currLeak = firstMemLeak;
    while (currLeak && (currLeak != oldLeak))
	currLeak = currLeak -> next;
    if (!currLeak)
	{

#ifdef vxWorks
	taskUnlock();
#if 0
	printf("Attempt to realloc nonallocated memory in %s\n", 
		taskName(taskIdSelf()));
#endif
#else
#if 0
	printf("Attempt to realloc nonallocated memory: %s/%d\n", 
		id, lineNumber);
#endif
#endif

	return (0);
	}

    if (currLeak -> prev)
	currLeak -> prev -> next = currLeak -> next;
    else
	firstMemLeak = currLeak -> next;

    if (currLeak -> next)
	currLeak -> next -> prev = currLeak -> prev;

#ifdef vxWorks
    taskUnlock();
#endif

    /*
    Reallocate the original memory buffer and put it at the start
    of the chain.
    */

    newSize += sizeof(struct memLeak);
    if (!(currLeak = (struct memLeak *) realloc(currLeak, newSize)))
	return (currLeak);

#ifdef vxWorks
    taskLock();
#endif

    currLeak -> next = firstMemLeak;
    if (currLeak -> next)
    	currLeak -> next -> prev = currLeak;
    currLeak -> prev = 0;
    firstMemLeak = currLeak;

#ifdef vxWorks
    taskUnlock();
#endif

    strncpy(currLeak -> name, id, sizeof(currLeak -> name));
    currLeak -> lineNumber = lineNumber;
    currLeak -> size = newSize - sizeof(struct memLeak);
    return (((char *) currLeak) + sizeof(struct memLeak));
#endif
}


/****************************************************************/
/*								*/
/*			REPORT MEMORY USAGE			*/
/*								*/
/****************************************************************/

tisMemoryUsage()
{
    struct memLeak	*currLeak;

    currLeak = firstMemLeak;

    while (currLeak)
	{
	printf("%s/%d: %d\n", currLeak -> name, currLeak -> lineNumber, 
		currLeak -> size);
	currLeak = currLeak -> next;
	}
    return (0);
}
