/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		DPA PROGRAM TO PING NODES			*/
/*								*/
/*	This program is a DPA that will ping nodes. It learns	*/
/*	the names to ping from the DPA open requests. Once a	*/
/*	node is put into the list, it stays there forever.	*/
/*								*/
/*	There are two tuning parameters:			*/
/*								*/
/*		PING_RATE:	How often to send ICMP packets	*/
/*								*/
/*		PING_TMOUT:	How fast to expect the answer	*/
/*								*/
/*	NOTE: This program must run as root to be able to use	*/
/*	raw socket I/O. Therefore, the program file has to	*/
/*	be owned by root and have the suid bit set if it is	*/
/*	not going to be directly started by root.		*/
/*								*/
/*	Author:	Gordon Uchenick & Greg Tomalesky       		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netdb.h>
#include <signal.h>

#include "dataTypes.h"
#include "cadef.h"
#include "alarm.h"
#include "msgMon.h"

#define PING_STRING_SIZE 20
#define PING_NAME_SIZE 40
#define	V_DEBUG		3
#define	V_STATUS	2
#define	V_INFO		1
#define V_NOVERBOSE	0
#define	VERBOSE(x)		if(pngVerbose >= x) tisSysLog
#define	PING_RATE	120		/* Seconds */
#define PING_TMOUT	20/*5*/		/* Seconds */
#define PING_CA_TIMEOUT 5		/* seconds */
#define PING_MAX_RETRYS 3		/* number of ping failures needed to declare node down */
#define RECV_PACKET_SIZE 1000
#define	CA_VALUE_SIZE	sizeof(CA_VALUE)

typedef enum {NODE_DOWN=0,NODE_UP,NODE_UNDEFINED} PNG_NODE_STATE;
typedef	union
{
  char char_value;				/* union of possible values */
  int long_value;				/* based on type from event handler */
  short short_value;
  float float_value;
  double double_value;
  char string_value[PING_STRING_SIZE];
} CA_VALUE;
struct pingNode
{
  struct pingNode *next;
  char nodeName[PING_NAME_SIZE];
  char monitorName[PING_NAME_SIZE];
  char statName[PING_NAME_SIZE];
  unsigned long	inetAddr;
  unsigned long	lastTxTime;
  unsigned long	lastRxTime;
  unsigned long lastCATime;
  int pngRetry;
  chid pngPingChannel;
  chid pngMonChannel;
  chid pngStatChannel;
  CA_VALUE pngValue;
  CA_VALUE pngLastValue;
  char pngIsValid;		/* zero means data not valid...one means OK */
  short	status;
  short pngMonStatus;
  short	stat;
  short	sevr;
  short pngDebouncer;		/* de-bounce counter. */
};
char *msgLevel[] =
{
  "Silent",
  "Informational",
  "Status",
  "DEBUG!"
};
char *stateString[] =
{
  "DOWN",
  "UP"
};
struct pingNode	*pingList = 0;
unsigned long	ipAddr;

void pngOpen(char *name,void *handle);
void pngRead(void *handle);
void pngClose(void *handle);
void	pngSignalHandler(int whichone);
char *timesToString(unsigned long *t1,unsigned long *t2);
int    	sockFd,caSocket,maxFd;
fd_set 	fdMask;
int  pngVerbose;
void pngCARead(struct event_handler_args arg);
void pngCAMonitor(struct event_handler_args arg);
void pngCAConnect(struct connection_handler_args arg);
void pngRegisterCA(void *arg, int fd, int cond);	/* Reconnect to CA */
void pngPing(struct pingNode *node);
void pngSetStatus(struct pingNode *node,PNG_NODE_STATE new);

char	*addrToString(unsigned int addr);

/****************************************************************/
/*								*/
/*			TUNING PARAMETERS			*/
/*								*/
/****************************************************************/

int pingRate,pingTimeOut,pingRetry,pingRetryTime,pingCATimeout;	/* all in seconds */
int connectToCa;
main(int argc,char *argv[])
{
  struct protoent      	*proto;
  struct timeval       	timeOut = {1,0};	/* One second */
  int		       	readSize;
  unsigned long		inetAddr;
  char			packet[RECV_PACKET_SIZE];
  unsigned long		timeNow;
  struct pingNode      	*currPing;
  short       	newStat;
  struct icmp	       	*icp;
  struct ip *ipMsg = (struct ip *)&packet[0];	/* first part of ANY message is the IP header */
  extern char	*optarg;
  int cmd;
  int i;
  int caInitialized;
  int count;
  int status;
  struct timeval maxWait,exitTime,currTime;

  tisOpenLog("png2",6,2,NULL,NULL);

  pingRate = PING_RATE;
  pingRetry = PING_MAX_RETRYS;
  pingRetryTime = pingRate/(pingRetry * 2);
  pingTimeOut = (pingRetryTime *3)/4;
  pingCATimeout = PING_CA_TIMEOUT;
  maxFd = 0;
  sockFd = -1;
  caSocket = -1;
#ifdef SOLARIS
  sigset(SIGHUP,pngSignalHandler);
  sigset(SIGINT,pngSignalHandler);
  sigset(SIGUSR1,pngSignalHandler);
  sigset(SIGUSR2,pngSignalHandler);
#else
  signal(SIGHUP,pngSignalHandler);
  signal(SIGINT,pngSignalHandler);
  signal(SIGUSR1,pngSignalHandler);
  signal(SIGUSR2,pngSignalHandler);
#endif
  while((cmd = getopt(argc,argv,"p:v:r:m:")) != -1)
    {
      switch(cmd)
	{
	case 'v':
	  if(*optarg != '-')
	    {
	      pngVerbose = abs(atoi(optarg));	/* set verbosity level */
	    }
	  else
	    {
	      fprintf(stderr,"Debug level must have a value.\n");
	      exit(1);
	    }
	  break;
	case 'p':
	  if(*optarg != '-')
	    {
	      pingRate = abs(atoi(optarg));	/* set ping rate */
	      pingRetryTime = pingRate/(pingRetry * 2);
	      pingTimeOut = (pingRetryTime *3)/4;
	    }
	  else
	    {
	      fprintf(stderr,"Ping rate must have a value.\n");
	      exit(1);
	    }
	  break;
	case 'r':
	  if(*optarg != '-')
	    {
	      pingRetry = abs(atoi(optarg));	/* set ping retry rate */
	      if(pingRetry < 1) pingRetry = 1;	/* never zero or negative */
	      pingRetryTime = pingRate/(pingRetry * 2);
	      pingTimeOut = (pingRetryTime *3)/4;
	    }
	  else
	    {
	      fprintf(stderr,"Retry count must have a value.\n");
	      exit(1);
	    }
	  break;
	case 'm':
	  if(*optarg != '-')
	    {
	      pingCATimeout = abs(atoi(optarg));	/* set CA timeout value */
	    }
	  else
	    {
	      fprintf(stderr,"CA monitor timeout must have a value.\n");
	      exit(1);
	    }
	  break;
	case '?':
	  fprintf(stderr,"Usage is:\n");
	  fprintf(stderr,"\tpng -p interval -t timeout\n");
	  fprintf(stderr,"\tinterval -- seconds between pings(default %d)\n",PING_RATE);
	  fprintf(stderr,"\ttimeout -- seconds until declared offline(default %d).\n",PING_TMOUT);
	  exit(1);
	}
    }
    /*
    Create the socket to send and receive the ICMP packets.
    */

  if (!(proto = getprotobyname("icmp")))
    {
      VERBOSE(V_NOVERBOSE)(0,0,"Can't find icmp protocol: %m");
      exit(1);
    }
  if ((sockFd = socket(AF_INET, SOCK_RAW, proto -> p_proto)) < 0)
    {
      VERBOSE(V_NOVERBOSE)(0,0,"Can't create socket: %m");
       exit(1);
    }
  if(sockFd > maxFd) maxFd = sockFd;
  /*
    Now process all events.
  */
  connectToCa = 1;						/* first time connect */
  caInitialized = 0;
  for (;;)
    {
      if(connectToCa)
	{
	  if(caInitialized)
	    ca_task_exit();
	  if(ca_task_initialize() != ECA_NORMAL)		/* try Channel Access */
	    {
	      VERBOSE(V_NOVERBOSE)(0,0,"Channel Access is not available");
	      exit(1);
	    }
	  caInitialized = 1;
	  ca_add_fd_registration (pngRegisterCA, NULL);
	  if(pngInit() == -1)  /* Read in file of nodes to ping and build ping list. */
	    {
	      VERBOSE(V_NOVERBOSE)(0,0,"Error reading node list file: %m");
	      exit(1);
	    }
	  VERBOSE(V_NOVERBOSE)(0,0,"Initialization complete...png2 starting");
	  connectToCa = 0;
	}
      /*
	Drain all socket I/O
      */
      FD_ZERO(&fdMask);
      if(sockFd >= 0)   FD_SET(sockFd,  &fdMask);
      if(caSocket >= 0) FD_SET(caSocket,&fdMask);
      maxWait.tv_sec = 1;
      maxWait.tv_usec = 0;
      while((maxWait.tv_sec >= 0) && (maxWait.tv_usec >= 0))
	{
	  /*
	    refresh FD mask for next time around
	  */
	  FD_ZERO(&fdMask);
	  if(sockFd >= 0)   FD_SET(sockFd,  &fdMask);
	  if(caSocket >= 0) FD_SET(caSocket,&fdMask);
	  /*
	    Drain away any incoming packets from our own socket.
	  */
	  gettimeofday(&currTime,0);
	  count = select(maxFd + 1, &fdMask, NULL, NULL, &timeOut);
	  VERBOSE(V_DEBUG)(0,0,"SEL -- Select return %d, %08X ca=%d,fd=%d,max=%d",
			   count,fdMask.fds_bits[0],caSocket,sockFd,maxFd);
	  if(count == 0)
	    break;
	  else if(count < 0)
	    {
	      VERBOSE(V_DEBUG)(0,0,"Error from select...%m");
	    }
	  /*
	    A socket has something for us to read. Find out which one and
	    read it.
	  */
	  else
	    {
	      if(FD_ISSET(sockFd,&fdMask))
		{
		  if ((readSize = recv(sockFd, (char *)&packet, RECV_PACKET_SIZE, 0)) > 0)
		    {
		      /*
			Got a packet. Make sure it is a ICMP_ECHOREPLY. If not the ignore it
			*/
		      icp = (struct icmp *)((char *)ipMsg + (ipMsg->ip_hl * sizeof(int)));		/* point to ICMP message */
		      if(icp -> icmp_type == ICMP_ECHOREPLY)		/* must be ICMP_ECHOREPLY! */
			{
			  inetAddr = ntohl(ipMsg->ip_src.S_un.S_addr);	/* get source address of this message from IP header */
			  for (currPing = pingList; currPing; currPing = currPing -> next)
			    {
			      if (currPing -> inetAddr == inetAddr)
				{
				  currPing -> lastRxTime = time(NULL);
				  currPing -> pngRetry = 0;
				  VERBOSE(V_STATUS)(0,0, "IN  -- inetAddr=%s,%s,\"%s\"",
						    addrToString(currPing -> inetAddr),
						    timesToString(&currPing -> lastRxTime,&currPing->lastTxTime),
						    currPing->nodeName);
				  if((currPing -> status != NODE_UP) && (currPing -> pngMonChannel == 0))
				    pngSetStatus(currPing,NODE_UP);
				}
			    }
			}
		    }
		}
	      if((caSocket >= 0) && FD_ISSET(caSocket,&fdMask))
		{
		  VERBOSE(V_DEBUG)(0,0,"CA -- Data available on FD %d",caSocket);
		  status = ca_pend_io(0.01);		/* drain socket to get data from CA */
		  for(i = 0;i < 100;i++)
		    {
		      if(status != ECA_TIMEOUT)
			status = ca_pend_io(0.01);
		      else
			break;
		    }
		}
	    }
	/*
	  adjust timer
	  */
	  gettimeofday(&exitTime,0);
	  exitTime.tv_sec -= currTime.tv_sec;
	  exitTime.tv_usec -= currTime.tv_usec;
	  if(exitTime.tv_usec < 0)
	    {
	      exitTime.tv_usec += 1000000;
	      exitTime.tv_sec--;
	    }
	  maxWait.tv_sec -= exitTime.tv_sec;
	  maxWait.tv_usec -= exitTime.tv_usec;
	  if(maxWait.tv_usec < 0)
	    {
	      maxWait.tv_usec += 1000000;
	      maxWait.tv_sec--;
	    }
	} /* end of while(select...) */
      /*
	Now check to see if any status has changed or if it is time to send out more packets. I got here
	as the result of a select timeout
      */
      timeNow = time(0);
      for (currPing = pingList; currPing; currPing = currPing -> next)
	{
	  /*
	    See if there has been a change in status.
	  */
	  if(currPing -> pngMonChannel != 0)
	     {
 	       newStat = currPing -> pngMonStatus;
	       VERBOSE(V_STATUS)(0,0, "MON -- %s,\"%s\",old=%lf,new=%lf,status=%d",
			       timesToString(&currPing->lastCATime,&timeNow),
			       currPing->nodeName,
			       currPing->pngLastValue.double_value,
			       currPing->pngValue.double_value,newStat);
	       if((currPing->pngLastValue.double_value == currPing->pngValue.double_value) && ((timeNow - currPing->lastCATime) > pingCATimeout))
		 newStat = NODE_DOWN;
	       else
		 {
		   newStat = NODE_UP;
		   currPing->pngLastValue.double_value = currPing->pngValue.double_value;
		 }
	       if(currPing->pngMonStatus != newStat)
		 pngSetStatus(currPing,newStat);
	     }
	  else
	    {
	      newStat = currPing -> status;
	      if((currPing -> pngRetry >= pingRetry) && (timeNow > (currPing->lastTxTime + pingTimeOut)) && (currPing -> lastRxTime < currPing -> lastTxTime))
		newStat = NODE_DOWN;
	      if (currPing -> status != newStat)
		pngSetStatus(currPing,newStat);
	      if((currPing->lastRxTime < currPing->lastTxTime) &&
		 ((timeNow - currPing->lastTxTime) > pingRetryTime) &&
		 (currPing->pngRetry < pingRetry))
		{
		  currPing->pngRetry++;
		  pngPing(currPing);
		}
	      if ((timeNow - currPing->lastTxTime) > pingRate)
		pngPing(currPing);		/* Send out the ping packet. */
	    }
	}
    }	/* end of BIG for loop */
}


/****************************************************************/
/*								*/
/*		COMPUTE THE ICMP HEADER CHECKSUM		*/
/*								*/
/****************************************************************/

in_cksum(addr, len)
u_short *addr;
int len;
{
  register int nleft = len;
  register unsigned short *w = addr;
  register int sum = 0;
  unsigned short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
  while (nleft > 1)  {
    sum += *w++;
    nleft -= 2;
  }

	/* mop up an odd byte, if necessary */
  if (nleft == 1) {
    *(unsigned char *)(&answer) = *(unsigned char *)w ;
    sum += answer;
  }

	/* add back carry outs from top 16 bits to low 16 bits */
  sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum += (sum >> 16);			/* add carry */
  answer = ~sum;				/* truncate to 16 bits */
  return(answer);
}
/*
  Handle signals
*/
void	pngSignalHandler(int whichone)
{
switch(whichone)
  {
  case SIGINT:
    VERBOSE(V_NOVERBOSE)(0,0,"Exiting due to ^C or -INT\n");
    ca_task_exit();
    exit(0);
    break;
  case SIGUSR1:
    if(++pngVerbose > V_DEBUG) pngVerbose = V_DEBUG;
    VERBOSE(V_NOVERBOSE)(0,0,"Message level now at %s\n",msgLevel[pngVerbose]);
    break;
  case SIGUSR2:
    if(--pngVerbose < V_NOVERBOSE) pngVerbose = V_NOVERBOSE;
    VERBOSE(V_NOVERBOSE)(0,0,"Message level now at %s\n",msgLevel[pngVerbose]);
    break;
  case SIGHUP:
    break;
  default:
    tisSysLog(0,0,"Unknown signal %d",whichone);
  }
}
char	*addrToString(unsigned int addr)
{
  static char *inetString = 0;
  if(inetString == 0)
    inetString = (char *)calloc(1,32);
  else
    memset(inetString,0,32);
  sprintf(inetString,"%03d.%03d.%03d.%03d\0",
	  ((addr >> 24) & 0xff),       	/* hi-order nibble first 31 - 24 */
	  ((addr >> 16) & 0xff),       	/* 23 - 16 next */
	  ((addr >>  8) & 0xff),       	/* 15 - 8 next */
	  (addr & 0xff));	       	/*  7 - 0 at last */
  return(inetString);
}
char *timesToString(unsigned long *t1,unsigned long *t2)
{
  static char timeString[60];
  char ts1[28],ts2[28];
  strcpy(ts1,(char *)ctime((time_t *)t1));
  ts1[strlen(ts1)-6] = 0;		/* drop line-feeds & year */
  strcpy(ts2,(char *)ctime((time_t *)t2));
  ts2[strlen(ts2)-6] = 0;
  sprintf(timeString,"RX= %s,TX= %s\0",&ts1[11],&ts2[11]);	/* ONLY Time */
  return(timeString);
}

int pngInit()
{
  FILE *fd;
  char buff[81],*nodeName,caName[50],*monName;
#ifdef	USE_STAT
  char *statName;
#endif	USE_STAT
  struct hostent *hostent;
  struct pingNode *currPing;
  int status;
  evid evID;

  if((fd = fopen("/tis4000db/pingList.txt","r")) != 0)
    {
      while(fgets(buff,80,fd) != 0)
	{
	  nodeName = (char *)strtok(buff,",\n");	/* get the name to ping */
#ifdef USE_STAT
	  statName = (char *)strtok(NULL,",\n");       	/* and the place to send stat values */
#endif
	  monName = (char *)strtok(NULL,",\n");		/* and the name to monitor */
	  /*
	    Get the Internet address for this node name.
	    */
	  if ((hostent = gethostbyname(nodeName)) == 0)
	    continue;
	  ipAddr = ntohl(**((long **)(hostent -> h_addr_list)));
	  /*
	    Look to see if we already have this node.
	    */
	  for (currPing = pingList; currPing; currPing = currPing -> next)
	    if (ipAddr == currPing -> inetAddr)
	      break;
	  if (!currPing)
	    {				/* This node is new. */
	      currPing = (struct pingNode *) calloc(1, sizeof(struct pingNode));
	      currPing -> next = pingList;
	      pingList = currPing;
	      strcpy(currPing->nodeName,nodeName);		/* 13-Mar-97 save name for verbosity later */
	      currPing->monitorName[0] = 0;
	      if(monName)
		strcpy(currPing->monitorName,monName);
	      currPing -> inetAddr = ipAddr;
	    }
	  currPing -> status = 2;
	  currPing -> pngRetry = 0;
	  currPing -> pngIsValid = 1;
	  /*
	    Establish CA channel to the DPA part of this node. The point name looks
	    like: dpa:nodeName.VAL
	  */
	  sprintf(caName,"dpa:%s.VAL\0",nodeName);
	  if((status = ca_build_read_and_monitor(caName,pngCAConnect,currPing,
						 pngCARead,currPing,
						 pngCAMonitor,currPing,
						 &currPing->pngPingChannel,&evID)) != ECA_NORMAL)
	    {
	      currPing->pngPingChannel = 0;
	      VERBOSE(V_NOVERBOSE)(0,0, "Error from ca_build_and_monitor for %s...%s",
				   caName,ca_message(status));
	    }
	  else
	    ca_pend_io(0.1);
#ifdef USE_STAT
	  if(statName)
	    {
	      sprintf(caName,"dpa:%s.VAL\0",statName);
	      if((status = ca_build_read_and_monitor(caName,pngCAConnect,currPing,
						     pngCARead,currPing,
						     pngCAMonitor,currPing,
						     &currPing->pngStatChannel,&evID)) != ECA_NORMAL)
		{
		  currPing->pngStatChannel = 0;
		  VERBOSE(V_NOVERBOSE)(0,0, "Error from ca_build_and_monitor for %s...%s",
				       caName,ca_message(status));
		}
	      else
		ca_pend_io(0.1);
	    }
	  else
	    currPing->pngStatChannel = 0;		/* can't get channel or no monitor name */
#endif
	  if(monName)
	    {
	      currPing -> pngMonStatus = NODE_UNDEFINED;
	      if((status = ca_build_read_and_monitor(monName,pngCAConnect,currPing,
						     pngCARead,currPing,
						     pngCAMonitor,currPing,
						     &currPing->pngMonChannel,&evID)) != ECA_NORMAL)
		{
		  currPing->pngMonChannel = 0;
		  VERBOSE(V_NOVERBOSE)(0,0, "Error from ca_build_and_monitor for %s...%s",
				       monName,ca_message(status));
		}
	      else
		ca_pend_io(0.1);
	    }
	  else
	    {
	      currPing -> pngMonChannel = 0;		/* can't get channel or no monitor name */
	    }
	  /*
	    And also, ping him right away!
	  */
	  if(currPing -> pngMonChannel == 0)
	    pngPing(currPing);
	}
    }
  else
    return(-1);
  return(0);
}
void pngRegisterCA(void *arg, int fd, int cond)	/* Reconnect to CA */
{
  struct pingNode *node;
  if (cond == 0)
    {
      VERBOSE(V_NOVERBOSE)(0,0, "Connection to CA lost!\n");
      FD_CLR(caSocket,&fdMask);
      caSocket = -1;
      connectToCa = 1;
      for (node = pingList; node; node = node -> next)
	node ->pngMonStatus = NODE_UNDEFINED;
    }
  else
    {
      caSocket = fd;				/* save CA fd here */
      if(caSocket > maxFd) maxFd = caSocket;	/* add to group I'm interested in */
      FD_SET(caSocket,&fdMask);
      VERBOSE(V_NOVERBOSE)(0,0, "Connection to CA established on %d maxFd=%d!",
			   caSocket,maxFd);
    }
}
void pngCAConnect(struct connection_handler_args arg)
{
struct pingNode  *node;
node = (struct pingNode *)arg.chid->puser;	/* address of the point info node for this point */

if(arg.op == CA_OP_CONN_UP)
  {
    ca_complete_bld_read_mon(arg.chid);
    if(node -> pngMonChannel == arg.chid)
      {
	VERBOSE(V_INFO)(0,0,"Got CA connect for %s",node->nodeName);
	if(node->pngMonStatus != NODE_UNDEFINED)
	  {
	    node-> pngMonStatus = NODE_UNDEFINED;
	    node -> lastCATime = time(NULL);			/* now is the time I got connected */
	    node -> pngIsValid = 1;
	  }
      }
  }
else
  {
    if(node -> pngMonChannel == arg.chid)
      {
	VERBOSE(V_INFO)(0,0,"Lost CA connect for %s",node->nodeName);
	node-> pngMonStatus = NODE_DOWN;
	node -> pngIsValid = 0;
      }
  }
    
}

void pngCAMonitor(struct event_handler_args arg)
{
  struct pingNode  *node;
  node = arg.usr;					/* address of the point info node for this point */
  if(node -> pngMonChannel == arg.chid)
    {
      node -> pngIsValid = 1;
      node -> lastCATime = time(NULL);			/* now is the time I heard from CA */
      switch(arg.type)
	{
	case OLD_DBR_STS_DOUBLE:
	  node->pngValue.double_value = *((double *)&(((struct dbr_sts_double *)arg.dbr)->value));
	  break;
	case OLD_DBR_STS_FLOAT:
	  node->pngValue.float_value = *((float *)&(((struct dbr_sts_float *)arg.dbr)->value));
	  break;
	case OLD_DBR_STS_CHAR:
	  node->pngValue.char_value = *((char *)&(((struct dbr_sts_char *)arg.dbr)->value));
	  break;
	case OLD_DBR_STS_LONG:
	  node->pngValue.long_value = *((long *)&(((struct dbr_sts_long *)arg.dbr)->value));
	  break;
	case OLD_DBR_STS_SHORT:
	  /*   case OLD_DBR_STS_ENUM: */
	  node->pngValue.short_value = *((short *)&(((struct dbr_sts_short *)arg.dbr)->value));
	  break;
	case OLD_DBR_CTRL_STRING:
	case OLD_DBR_STS_STRING:
	  strcpy(&node->pngValue.string_value[0],((struct dbr_sts_string *)arg.dbr)->value);
	  break;
	case OLD_DBR_CTRL_DOUBLE:
	  node->pngValue.double_value = *((double *)&(((struct dbr_sts_double *)arg.dbr)->value));
	  break;
	case OLD_DBR_CTRL_FLOAT:
	  node->pngValue.float_value = *((float *)&(((struct dbr_sts_float *)arg.dbr)->value));
	  break;
	case OLD_DBR_CTRL_CHAR:
	  node->pngValue.char_value = *((char *)&(((struct dbr_sts_char *)arg.dbr)->value));
	  break;
	case OLD_DBR_CTRL_LONG:
	  node->pngValue.long_value = *((long *)&(((struct dbr_sts_long *)arg.dbr)->value));
	  break;
	case OLD_DBR_CTRL_SHORT:
	  /*   case OLD_DBR_CTRL_ENUM: */
	  node->pngValue.short_value = *((short *)&(((struct dbr_sts_short *)arg.dbr)->value));
	  break;
	}
    }
}

void pngCARead(struct event_handler_args arg)
{
  pngCAMonitor(arg);
}

void pngPing(struct pingNode *node)
{
  struct icmp icmpBuff;
  struct icmp *icp = &icmpBuff;
  struct sockaddr_in destination;

  icp -> icmp_type = ICMP_ECHO;
  icp -> icmp_code = 0;
  icp -> icmp_cksum = 0;
  icp -> icmp_seq = 1;
  icp -> icmp_id = 0;
  icp -> icmp_cksum = in_cksum((unsigned short *) icp, ICMP_MINLEN);
  destination.sin_family = AF_INET;
  destination.sin_addr.s_addr = node -> inetAddr;
  sendto(sockFd, (char *) icp, ICMP_MINLEN, 0, (struct sockaddr *)&destination, sizeof(destination));
  node->lastTxTime =  time(NULL);
  VERBOSE(V_STATUS)(0,0,"OUT -- inetAddr=%s,%s,\"%s\"",
		   addrToString(node -> inetAddr),
		   timesToString(&node -> lastRxTime,&node->lastTxTime),
		   node->nodeName);
}
/*
  Send status changes to DPA
*/
void pngSetStatus(struct pingNode *node,PNG_NODE_STATE new)
{
  if(node -> pngMonChannel != 0)
    {
      node -> pngMonStatus = new;
      ca_put(OLD_DBF_ENUM,node->pngPingChannel,&node->pngMonStatus);
      ca_pend_io(0.1);		/* send data to CA */
    }
  else
    {
      node -> status = new;
      ca_put(OLD_DBF_ENUM,node->pngPingChannel,&node->status);
      ca_pend_io(0.1);		/* send data to CA */
    }
  VERBOSE(V_STATUS)(0,0,"CHG -- status=%d,inetAddr=%s,%s,\"%s\"",
		    new,
		    addrToString(node->inetAddr),
		    timesToString(&node -> lastRxTime,&node->lastTxTime),
		    node->nodeName);
}
