/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*			LDFW1 RECORD PROCESSING			*/
/*								*/
/*	Record support for the LDFW1 block type. This block	*/
/*	controls the start, stop, and sequence from one to	*/
/*	twenty pumps. Timing parameters control behavior	*/
/*	in runtime selection mode. The operator can also	*/
/*	enter a designated sequence.				*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <lstLib.h>
#include <string.h>
#include <tickLib.h>

#include "alarm.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbAddr.h"
#include "dbRecType.h"
#include "recSup.h"
#include "special.h"
#include "rec/ldfw1Record.h"
#include "rec/pumpRecord.h"
#include "redflds.h"
#include "redque.h"
#include "restable.h"
#include "tankfarm.h"
#include "db_addr.h"
#include "dbEvent.h"
#include "recGbl.h"
#include "ldfw.h"
#include "ldfw1redfldidx.h"

/******
	Definitions for the RSET
******/

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
#define get_alarm_double NULL

struct rset ldfw1RSET = {

    RSETNUMBER,
    report,
    initialize,
    init_record,
    process,
    special,
    get_value,
    cvt_dbaddr,
    get_array_info,
    put_array_info,
    get_units,
    get_precision,
    get_enum_str,
    get_enum_strs,
    put_enum_str,
    get_graphic_double,
    get_control_double,
    get_alarm_double
    };
    
/****************************************************************/
/*								*/
/*			EXTERNAL DATA				*/
/*								*/
/****************************************************************/

/* defined as global in ../db/iocInit.c */
extern long *process_cnts;
/* defined as global in ../db/procstat.c */
extern long *process_stats;
extern unsigned short DBC_CRCRecord();


/****************************************************************/
/*								*/
/*			PRIVATE DATA				*/
/*								*/
/****************************************************************/

static unsigned short type;
static void alarm ();
static void monitor();
static struct fldDes **mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"PREQ",
		"LPST",
		"LPCT",
		"LPCM",
		"PC01",
		"PC02",
		"PC03",
		"PC04",
		"PC05",
		"PC06",
		"PC07",
		"PC08",
		"PC09",
		"PC10",
		"PC11",
		"PC12",
		"PC13",
		"PC14",
		"PC15",
		"PC16",
		"PC17",
		"PC18",
		"PC19",
		"PC20",
		0
		};

#define CONFIGURED	0x100
#define MAX_PUMPS	20


/****************************************************************/
/*								*/
/*			INITIALIZE A RECORD			*/
/*								*/
/****************************************************************/

static long init_record (pldfw1)
struct ldfw1Record	*pldfw1;
{
    struct dbAddr	*pdbAddr;
    struct dbCommon	*recHdr;
    short		master;
    short		i;
    char		*pstr;
    char		redbuf [512];

    if (!(mylist))
	{

	/* find record type for ldfw1 */

	for (type = 1, i = 0; type < 100; type++)
	    {
	    pstr = GET_PRECTYPE(type);
	    if (strncmp(pstr, "ldfw1", 5) == 0)
		{
		i = 1;
		break;
		}
	    }
	
	if (i == 0)
	    {
	    recGblRecordError(S_db_noRecType, (struct dbCommon *)pldfw1, "ldfw1:init_record");
	    return (S_db_noRecType);
	    }

	
    	/*
    	Initialize redundancy stuff here.
    	*/

	mylist = redFldAlloc (type, NUMLDFW1FLDS);

	if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pldfw1,
				   "ldfw1:init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	for (i = 0; i < NUMLDFW1FLDS; i++)
	    redFldAdd (type, i, namelist[i]);
	}

    /*
    Initialization stuff typical to all point types.
    */

    pldfw1 -> init = TRUE;
    pldfw1 -> nsta = NO_ALARM;
    pldfw1 -> nsev = NO_ALARM;

    if (pldfw1 -> latm.secPastEpoch == 0)
	tsLocalTime(&pldfw1 -> latm);
    if (pldfw1 -> lptm.secPastEpoch == 0)
	tsLocalTime(&pldfw1 -> lptm);


    /*
    See if all the pumps we need to work with are fully initialized.
    */

    resGetCaAnswer(pldfw1 -> radr, &master);
    if (!ldfwCheckInit(&(pldfw1 -> pm01), &(pldfw1 -> ps01)))
	{
	if (master)
	    {
	    pldfw1 -> nsta = NO_INIT_ALARM;
	    pldfw1 -> nsev = VALID_ALARM;
	    monitor(pldfw1, 0, 0, 0, 0, pldfw1 -> lpst);
	    return (0);
	    }
	}

    /*
    See if the suspend mode input has fully initialized.
    */

    if (pldfw1 -> susp.type != CONSTANT)
	{
	if (pldfw1 -> susp.type != DB_LINK)
	    {
	    if (master)
		{
		pldfw1 -> nsta = NO_INIT_ALARM;
		pldfw1 -> nsev = VALID_ALARM;
		monitor(pldfw1, 0, 0, 0, 0, pldfw1 -> lpst);
	    	return (0);
		}
	    }
	pdbAddr = (struct dbAddr *) (pldfw1 -> susp.value.db_link.pdbAddr);
	recHdr = (struct dbCommon *) (pdbAddr -> precord);
	if ((recHdr -> init == 0) || (recHdr -> ini2 == 0))
	    {
	    if (master)
		{
		pldfw1 -> nsta = NO_INIT_ALARM;
		pldfw1 -> nsev = VALID_ALARM;
		monitor(pldfw1, 0, 0, 0, 0, pldfw1 -> lpst);
	    	return (0);
		}
	    }
	}

    /*
    If we are on the master machine, set our mode to MANUAL and set
    the requested number of pumps to the number running.
    */

    if (master)
	{

    	pldfw1 -> mode = 1;		/* Manual mode */
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> mode), DBE_VALUE);
    	pldfw1 -> ini2 = TRUE;
	pldfw1 -> rcrc = DBC_CRCRecord(pldfw1, type);

	/*
	Force redundant update to get master disk copy in sync now!
	*/

	redQInit ((struct dbCommon *)pldfw1, redbuf, sizeof (redbuf), type);

	redQField (redbuf, mylist [INI2]);

	redQSend (redbuf);
	}

    /*
    We've initialized -- await first processing pass to get rest going.
    */

    return (1);
}


/****************************************************************/
/*								*/
/*			PROCESS A RECORD			*/
/*								*/
/****************************************************************/

static long process(paddr)
struct dbAddr		*paddr;
{
    struct ldfw1Record		*pldfw1 = (struct ldfw1Record *)(paddr->precord);
    short			lowerLimit;
    short			upperLimit;
    short			runningPumps;
    short			stoppingPumps;
    short			pumpsRequested;
    short			lastDesigStart;
    short			status;
    short			susm;
    unsigned short		stat = pldfw1 -> stat;
    unsigned short		sevr = pldfw1 -> sevr;
    short			val  = pldfw1 -> val;
    short			lalm = pldfw1 -> lalm;
    TS_STAMP			latm = pldfw1 -> latm;
    short			mlst = pldfw1 -> mlst;
    short			preq = pldfw1 -> preq;
    short			lpst = pldfw1 -> lpst;
    unsigned long		lpct = pldfw1 -> lpct;
    unsigned long		lpcm = pldfw1 -> lpcm;
    unsigned long		pc01 = pldfw1 -> pc01;
    unsigned long		pc02 = pldfw1 -> pc02;
    unsigned long		pc03 = pldfw1 -> pc03;
    unsigned long		pc04 = pldfw1 -> pc04;
    unsigned long		pc05 = pldfw1 -> pc05;
    unsigned long		pc06 = pldfw1 -> pc06;
    unsigned long		pc07 = pldfw1 -> pc07;
    unsigned long		pc08 = pldfw1 -> pc08;
    unsigned long		pc09 = pldfw1 -> pc09;
    unsigned long		pc10 = pldfw1 -> pc10;
    unsigned long		pc11 = pldfw1 -> pc11;
    unsigned long		pc12 = pldfw1 -> pc12;
    unsigned long		pc13 = pldfw1 -> pc13;
    unsigned long		pc14 = pldfw1 -> pc14;
    unsigned long		pc15 = pldfw1 -> pc15;
    unsigned long		pc16 = pldfw1 -> pc16;
    unsigned long		pc17 = pldfw1 -> pc17;
    unsigned long		pc18 = pldfw1 -> pc18;
    unsigned long		pc19 = pldfw1 -> pc19;
    unsigned long		pc20 = pldfw1 -> pc20;
    char			redbuf[512];

    /*
    Locate the record.
    */

/*    pldfw1 = (struct ldfw1Record *) (paddr -> precord); */

    /*
    Do nothing if we didn't init.
    */

    if ((pldfw1 -> init == 0) || (pldfw1 -> ini2 == 0))
	return (0);

    /*
    Get last process time.
    */

    tsLocalTime(&(pldfw1 -> lptm));

    /*
    Init new running/stopping pump counts to previous values.
    */

    runningPumps = pldfw1 -> val - pldfw1 -> nstp;
    stoppingPumps = pldfw1 -> nstp;
    lastDesigStart = pldfw1 -> lpst;

    /*
    Copy current limits so we can detect change later.
    */

    lowerLimit = pldfw1 -> llim;
    upperLimit = pldfw1 -> ulim;

    /*
    Check for tagged or off scan.
    */

    if (pldfw1 -> tagd)
	{
	pldfw1 -> nsta = TAGGED_ALARM;
	pldfw1 -> nsev = VALID_ALARM;
	}
    else if (pldfw1 -> oscn == OFF_SCAN)
	{
	pldfw1 -> nsta = SCAN_ALARM;
	pldfw1 -> nsev = VALID_ALARM;
	}
    else
	{
	pldfw1 -> nsta = NO_ALARM;
	pldfw1 -> nsev = NO_ALARM;

	/*
	This record must process. Get the number of pumps that are
	currently running. This will also update all our timers for
	pumps in case we are in manual or suspended mode.
	*/

	ldfwPumpsRunning(&(pldfw1 -> pm01), &lowerLimit, &upperLimit, &(pldfw1 -> ps01),
		&(pldfw1 -> pc01), &runningPumps, &stoppingPumps, pldfw1 -> nrpm,
		&(pldfw1 -> lpct), &(pldfw1 -> lpcm));

	/*
	If defined, get the suspended mode input.
	*/

	if (pldfw1 -> susp.type == DB_LINK)
	    {
	    long	options = 0;
	    long	nRequest = 1;

	    status = dbGetLink(&pldfw1 -> susp.value.db_link, 
			       (struct dbCommon *)pldfw1, 
			       DBR_SHORT, (caddr_t) &susm, 
			       &options, &nRequest);
	    }

	/*
	Continue processing only if we are not suspended and also are in automatic.
	*/

	if ((susm != pldfw1 -> susv) && (pldfw1 -> mode == 0))
	    {

	    /*
	    Look at the number that have been requested and handle upper and lower limit.
	    If the block is in both upper and lower (i.e., HIGH and LOW) alarm, then the
	    condition with the higher severity will prevail. If they are the same, then
	    the HIGH alarm will be reported.
	    */

	    pumpsRequested = pldfw1 -> preq;
	    if (pumpsRequested > upperLimit)
		{
		pumpsRequested = upperLimit;
		if (pldfw1 -> usvr > pldfw1 -> nsev)
		    {
		    pldfw1 -> nsev = pldfw1 -> usvr;
		    pldfw1 -> nsta = HIGH_ALARM;
		    }
		}
	    if (pumpsRequested < lowerLimit)
		{
		pumpsRequested = lowerLimit;
		if (pldfw1 -> lsvr > pldfw1 -> nsev)
		    {
		    pldfw1 -> nsev = pldfw1 -> lsvr;
		    pldfw1 -> nsta = LOW_ALARM;
		    }
		}

	    /*
	    Compare the number of pumps that are running to the number that are
	    supposed to be running. If they are different, then we have to do
	    something about it.
	    */

	    if (pumpsRequested > (runningPumps + stoppingPumps))
		{

		/*
		Start another pump.
		*/

		if (ldfwStartPump(&(pldfw1 -> lpct), pldfw1 -> strs, &(pldfw1 -> pm01),
			&(pldfw1 -> ps01), &(pldfw1 -> pc01), pldfw1 -> ospt,
			pldfw1 -> psmd, &lastDesigStart, &(pldfw1 -> ds01), &(pldfw1 -> lpcm)))
		    runningPumps++;
		}

	   else if (pumpsRequested < runningPumps)
		{

		/*
		Stop a pump.
		*/

		if (ldfwStopPump(&(pldfw1 -> lpct), pldfw1 -> stps, &(pldfw1 -> pm01), 
			&(pldfw1 -> ps01), &(pldfw1 -> pc01), pldfw1 -> psmd, 
			&(pldfw1 -> ds01), &(pldfw1 -> lpcm)))
		    {
		    runningPumps--;
		    stoppingPumps++;
		    }
		}
	    }

	    alarm (pldfw1);
	}

    /*
    Call the monitor routine to let the rest of the world know what's going on.
    */

    monitor(pldfw1, lowerLimit, upperLimit, runningPumps, stoppingPumps, lastDesigStart, susm);

    /*
    Handle redundancy here
    */

    redQInit ((struct dbCommon *)pldfw1, redbuf, sizeof (redbuf), type);

    if (stat != pldfw1 -> stat)
	redQField (redbuf, mylist [STAT]);

    if (sevr != pldfw1 -> sevr)
	redQField (redbuf, mylist [SEVR]);

    if (val != pldfw1 -> val)
	redQField (redbuf, mylist [VAL]);

    if (lalm != pldfw1 -> lalm)
	redQField (redbuf, mylist [LALM]);

    if (TsCmpStampsNE (&latm, &pldfw1 -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (mlst != pldfw1 -> mlst)
	redQField (redbuf, mylist [MLST]);

    if (preq != pldfw1 -> preq)
	redQField (redbuf, mylist [PREQ]);

    if (lpst != pldfw1 -> lpst)
	redQField (redbuf, mylist [LPST]);

    if (lpct != pldfw1 -> lpct)
	redQField (redbuf, mylist [LPCT]);

    if (lpcm != pldfw1 -> lpcm)
	redQField (redbuf, mylist [LPCM]);

    if (pc01 != pldfw1 -> pc01)
	redQField (redbuf, mylist [PC01]);

    if (pc02 != pldfw1 -> pc02)
	redQField (redbuf, mylist [PC02]);

    if (pc03 != pldfw1 -> pc03)
	redQField (redbuf, mylist [PC03]);

    if (pc04 != pldfw1 -> pc04)
	redQField (redbuf, mylist [PC04]);

    if (pc05 != pldfw1 -> pc05)
	redQField (redbuf, mylist [PC05]);

    if (pc06 != pldfw1 -> pc06)
	redQField (redbuf, mylist [PC06]);

    if (pc07 != pldfw1 -> pc07)
	redQField (redbuf, mylist [PC07]);

    if (pc08 != pldfw1 -> pc08)
	redQField (redbuf, mylist [PC08]);

    if (pc09 != pldfw1 -> pc09)
	redQField (redbuf, mylist [PC09]);

    if (pc10 != pldfw1 -> pc10)
	redQField (redbuf, mylist [PC10]);

    if (pc11 != pldfw1 -> pc11)
	redQField (redbuf, mylist [PC11]);

    if (pc12 != pldfw1 -> pc12)
	redQField (redbuf, mylist [PC12]);

    if (pc13 != pldfw1 -> pc13)
	redQField (redbuf, mylist [PC13]);

    if (pc14 != pldfw1 -> pc14)
	redQField (redbuf, mylist [PC14]);

    if (pc15 != pldfw1 -> pc15)
	redQField (redbuf, mylist [PC15]);

    if (pc16 != pldfw1 -> pc16)
	redQField (redbuf, mylist [PC16]);

    if (pc17 != pldfw1 -> pc17)
	redQField (redbuf, mylist [PC17]);

    if (pc18 != pldfw1 -> pc18)
	redQField (redbuf, mylist [PC18]);

    if (pc19 != pldfw1 -> pc19)
	redQField (redbuf, mylist [PC19]);

    if (pc20 != pldfw1 -> pc20)
	redQField (redbuf, mylist [PC20]);

    redQSend (redbuf);

    return (0);
}


/****************************************************************/
/*								*/
/*			SPECIAL PROCESSING			*/
/*								*/
/****************************************************************/

static long special (paddr, after)
    struct	dbAddr	*paddr;
    int			after;
{
    int				special_type = paddr -> special;
    long			status = 0;

    if (!after)
    {
	return (status);
    }

    switch (special_type) 
    {
	case SPC_LDFW_CTCK:
	    *((unsigned long *)paddr -> pfield) = tickGet ();
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "ldfw1: special");
	    status = S_db_badChoice;
	    break;
    }

    return (status);
}


/****************************************************************/
/*								*/
/*		CHECK FOR AND REPORT ALARM CONDITIONS		*/
/*								*/
/****************************************************************/

static void alarm (pldfw1)
    struct ldfw1Record	*pldfw1;
{
    int	adis;

    adis = AlarmDisableLink ((void *)pldfw1);
    if ( (pldfw1 -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	pldfw1 -> nsta = INHIBIT_ALARM;
	pldfw1 -> nsev = NO_ALARM;
    }
}


/****************************************************************/
/*								*/
/*		POST MONITOR EVENTS AS NECESSARY		*/
/*								*/
/****************************************************************/

static void monitor(pldfw1, lowerLimit, upperLimit, runningPumps, stoppingPumps, 
		lastDesigStart, susm)
struct ldfw1Record	*pldfw1;
short			lowerLimit;
short			upperLimit;
short			runningPumps;
short			stoppingPumps;
short			lastDesigStart;
short			susm;
{
    unsigned short	eventMask = 0;
    short		stat, sevr, nsta, nsev;
    int			set_ack = 1;

    /*
    Check lower limit.
    */

    if (lowerLimit != pldfw1 -> llim)
	{
	pldfw1 -> llim = lowerLimit;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> llim), DBE_VALUE);
	}

    /*
    Check upper limit.
    */

    if (upperLimit != pldfw1 -> ulim)
	{
	pldfw1 -> ulim = upperLimit;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> ulim), DBE_VALUE);
	}

    /*
    Check last designated start index.
    */

    if (lastDesigStart != pldfw1 -> lpst)
	{
	pldfw1 -> lpst = lastDesigStart;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> lpst), DBE_VALUE);

	/*
	LPST is a prompted field. Recalculate CRC and perform redundant update.
	*/

	}

    /*
    Check number stopping pumps.
    */

    if (stoppingPumps != pldfw1 -> nstp)
	{
	pldfw1 -> nstp = stoppingPumps;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> nstp), DBE_VALUE);
	}

    /*
    Check suspend mode input value.
    */

    if (susm != pldfw1 -> susm)
	{
	pldfw1 -> susm = susm;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> susm), DBE_VALUE);
	}

    /*
    If we are in manual, then make # pumps requested equal to
    the number that are running.

    Also do this if suspended!
    */

    if (((susm == pldfw1 -> susv) || (pldfw1 -> mode == 1)) && (pldfw1 -> preq != runningPumps))
	{
	pldfw1 -> preq = runningPumps;
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> preq), DBE_VALUE);
	}

    /*
    Check for change is status or severity, remembering if either
    one of them changed.
    */
    stat = pldfw1 -> stat;
    sevr = pldfw1 -> sevr;
    nsta = pldfw1 -> nsta;
    nsev = pldfw1 -> nsev;

    /* set current stat and sevr */
    pldfw1 -> stat = nsta;
    pldfw1 -> sevr = nsev;
    pldfw1 -> nsta = 0;
    pldfw1 -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pldfw1 -> psta) && (nsev == pldfw1 -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and our previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pldfw1 -> psta = stat;
	pldfw1 -> psev = sevr;
	db_post_events ((struct dbCommon *)pldfw1, (caddr_t)&pldfw1 -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pldfw1, (caddr_t)&pldfw1 -> psev, DBE_VALUE);
    }

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	eventMask = DBE_ALARM;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)pldfw1, (caddr_t)&pldfw1 -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pldfw1, (caddr_t)&pldfw1 -> sevr, DBE_VALUE);
	if ( (pldfw1 -> sevr != NO_ALARM) && (set_ack) &&
	     (pldfw1 -> sevr != EVENT_ALARM) )
	{
	    if (pldfw1 -> ack == 0) 
	    {
		pldfw1 -> ack = 1;           /* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)pldfw1, (caddr_t)&pldfw1 -> ack, DBE_VALUE);
	    }
	}
    }

    /*
    Now look for a change in value.
    */

    if ((runningPumps + stoppingPumps) != pldfw1 -> val)
	{
	pldfw1 -> val = runningPumps + stoppingPumps;
	eventMask |= DBE_VALUE;
	if (pldfw1 -> tenb)
	    eventMask |= DBE_LOG;
	}

    if (eventMask & DBE_ALARM)
	{
	pldfw1 -> lalm = pldfw1 -> val;
	tsLocalTime(&(pldfw1 -> latm));
	}

    if (eventMask)
	db_post_events((struct dbCommon *)pldfw1, (caddr_t)&(pldfw1 -> val), eventMask);
}


/****************************************************************/
/*								*/
/*		DATABASE ACCESS UTILITY ROUTINES		*/
/*								*/
/****************************************************************/

static long get_precision(paddr, precision)
struct dbAddr	*paddr;
long		*precision;
{
    /*
    You can't have a fractional pump. Also assume that you can't
    have fractional seconds, at least for now.
    */

    *precision = 0;
    return (0);
}

static long get_graphic_double(paddr, pgd)
struct dbAddr		*paddr;
struct dbr_grDouble	*pgd;
{
    short		pumpsConfigured;
    short		*pumpStatus;
    short		i;

    struct ldfw1Record	*pldfw1 = (struct ldfw1Record *) (paddr -> precord);

    switch ( ((struct fldDes *)(paddr -> pfldDes)) -> offset )
	{
	case ( (int)&(((struct ldfw1Record *)0) -> strs) ):
	case ( (int)&(((struct ldfw1Record *)0) -> stps) ):
	case ( (int)&(((struct ldfw1Record *)0) -> ospt) ):
	    pgd -> upper_disp_limit = 99999;
	    break;
	default:
	    pumpsConfigured = 0;
	    pumpStatus = &(pldfw1 -> ps01);
	    for (i = 0; i < MAX_PUMPS; i++)
		if (pumpStatus[i] & CONFIGURED)
		    pumpsConfigured++;
	    pgd -> upper_disp_limit = pumpsConfigured;
	    break;
	}
    pgd -> lower_disp_limit = 0;
    return (0);
}

static long get_control_double(paddr, pcd)
struct dbAddr		*paddr;
struct dbr_ctrlDouble	*pcd;
{
    short		pumpsConfigured;
    short		*pumpStatus;
    short		i;

    struct ldfw1Record	*pldfw1 = (struct ldfw1Record *) (paddr -> precord);

    switch ( ((struct fldDes *)(paddr -> pfldDes)) -> offset )
	{
	case ( (int)&(((struct ldfw1Record *)0) -> strs) ):
	case ( (int)&(((struct ldfw1Record *)0) -> stps) ):
	case ( (int)&(((struct ldfw1Record *)0) -> ospt) ):
	    pcd -> upper_ctrl_limit = 99999;
	    break;
	default:
	    pumpsConfigured = 0;
	    pumpStatus = &(pldfw1 -> ps01);
	    for (i = 0; i < MAX_PUMPS; i++)
		if (pumpStatus[i] & CONFIGURED)
		    pumpsConfigured++;
	    pcd -> upper_ctrl_limit = pumpsConfigured;
	}
    pcd -> lower_ctrl_limit = 0;
    return (0);
}
