/***************************************************************************
*
*    rbTask
*
*	This task will run every x seconds and calls a function for the
*	AO, BO and MBBO that will readback the output and compare it to
*	the points current VAL to see if they agree, and if not update VAL
*
****************************************************************************/

/*	Modification Log
*       ----------------
*	01	03-25-97	ems	creation date
*
*/


#include <vxWorks.h>
#include <inetLib.h>
#include <string.h>
#include <task_params.h>
#include <logLib.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "recSup.h"
#include "dbRecords.h"
#include "dbRecType.h"
#include "tsDefs.h"
#include "alarm.h"
#include "rec/aoRecord.h"
#include "rec/boRecord.h"
#include "rec/mbboRecord.h"
#include "restable.h"

#define	DEFAULT_TD	30	/* default task delay time */
#define	MINREADBACK	10	/* minimum readback delay time */

enum {
    AO_IDX = 0,
    BO_IDX,
    MBBO_IDX,
    LAST_IDX
    };

static int	delay_time = 0;

static int	checkEligable (int, caddr_t);
static int	checkTime (char *, unsigned short, TS_STAMP);
int		getRbDelay (void);
void		rbTask (void);
int		setRbDelay (int);

/*******************************************************************************/

static int	checkEligable (int idx, caddr_t pRecord)
{
    int			status = 1;

    switch (idx)
    {
	case AO_IDX:
	{
	    struct aoRecord	*pao;

	    pao = (struct aoRecord *)pRecord;

	    /* if not hardware point */
	    if ( (pao -> dtyp <= 1) &&
		 ( (pao -> out.type == CONSTANT)||
		   (pao -> out.type == PV_LINK) ||
		   (pao -> out.type == DB_LINK) ||
		   (pao -> out.type == CA_LINK)) )
		 break;

	    /* if record is already being processed */
	    if (pao -> pact)
		break;

	    /* if we are tagged, or off scan, we dont 
	       want to do this
	    */
	    if ( (pao -> tagd) || (pao -> oscn == OFF_SCAN) )
		break;

	    /* if we are going to get our input from another 
	       point we dont want to do this
	    */
	    if ( (pao -> dol.type == DB_LINK) && 
		 (pao -> omsl == CLOSED_LOOP) )
		break;

	    if (checkTime (pao -> name, pao -> rbdt, pao -> lptm))
		break;

	    status = 0;
	    break;
	}
	case BO_IDX:
	{
	    struct boRecord	*pbo;

	    pbo = (struct boRecord *)pRecord;

	    /* if not hardware point */
	    if ( (pbo -> dtyp <= 1) &&
		 ( (pbo -> out.type == CONSTANT)||
		   (pbo -> out.type == PV_LINK) ||
		   (pbo -> out.type == DB_LINK) ||
		   (pbo -> out.type == CA_LINK)) )
		 break;

	    /* if record is already being processed */
	    if (pbo -> pact)
		break;

	    /* if we are tagged, or off scan, we dont 
	       want to do this
	    */
	    if ( (pbo -> tagd) || (pbo -> oscn == OFF_SCAN) )
		break;

	    /* if we are going to get our input from another 
	       point we dont want to do this
	    */
	    if ( (pbo -> dol.type == DB_LINK) && 
		 (pbo -> omsl == CLOSED_LOOP) )
		break;

	    if (checkTime (pbo -> name, pbo -> rbdt, pbo -> lptm))
		break;

	    status = 0;
	    break;
	}
	case MBBO_IDX:
	{
	    struct mbboRecord	*pmbbo;

	    pmbbo = (struct mbboRecord *)pRecord;

	    /* if not hardware point */
	    if ( (pmbbo -> dtyp <= 1) &&
		 ( (pmbbo -> out.type == CONSTANT)||
		   (pmbbo -> out.type == PV_LINK) ||
		   (pmbbo -> out.type == DB_LINK) ||
		   (pmbbo -> out.type == CA_LINK)) )
		 break;

	    /* if record is already being processed */
	    if (pmbbo -> pact)
		break;

	    /* if we are tagged, or off scan, we dont 
	       want to do this
	    */
	    if ( (pmbbo -> tagd) || (pmbbo -> oscn == OFF_SCAN) )
		break;

	    /* if we are going to get our input from another 
	       point we dont want to do this
	    */
	    if ( (pmbbo -> dol.type == DB_LINK) && 
		 (pmbbo -> omsl == CLOSED_LOOP) )
		break;

	    if (checkTime (pmbbo -> name, pmbbo -> rbdt, pmbbo -> lptm))
		break;

	    status = 0;
	    break;
	}
	default:
	    break;
    }

    return (status);
}


/*******************************************************************************/

static int	checkTime (char *pvname, unsigned short rbdt, TS_STAMP lptm)
{
    TS_STAMP		now;
    unsigned int	deltaTime;

    /* get the current epics time */
    tsLocalTime (&now);

    /* get the time in seconds since this point was last processed */
    deltaTime = now.secPastEpoch - lptm.secPastEpoch;

    /* if our last process time was greater than the deadtime choice
       we want to process this record
    */
    if (deltaTime > rbdt)
	return (0);
    return (1);
}


/*******************************************************************************/

int	getRbDelay (void)
{
    int		delaytime;

    delaytime = delay_time / vxTicksPerSecond;
    logMsg ("readback task delay is %d seconds\n", (int)delaytime,0,0,0,0,0);
    return (0);
}


/*******************************************************************************/

void	rbTask (void)
{
    struct recLoc	*recLoc;
    struct recChain	*recChains[3];
    struct recChain	*chain;
    char		*pStr;
    int			idx;
    int			aoidx = 0;
    int			boidx = 0;
    int			mbboidx = 0;
    int			foundIt;
    short		master;

    if (delay_time == 0)
	delay_time = DEFAULT_TD * vxTicksPerSecond;

    for (aoidx = 1, foundIt = 0; aoidx < dbRecords -> number; aoidx++) 
    {
	pStr = GET_PRECTYPE (aoidx);
	if ((strcmp (pStr, "ao")) == 0) 
	{
	    foundIt = 1;
	    break;
	}
    }
    if (foundIt == 0)
    {
	logMsg ("rbTask: can not find ao index\n",0,0,0,0,0,0);
	taskSuspend (taskIdSelf());
    }

    for (boidx = 1, foundIt = 0; boidx < dbRecords -> number; boidx++) 
    {
	pStr = GET_PRECTYPE (boidx);
	if ((strcmp (pStr, "bo")) == 0) 
	{
	    foundIt = 1;
	    break;
	}
    }
    if (foundIt == 0)
    {
	logMsg ("rbTask: can not find bo index\n",0,0,0,0,0,0);
	taskSuspend (taskIdSelf());
    }

    for (mbboidx = 1, foundIt = 0; mbboidx < dbRecords -> number; mbboidx++) 
    {
	pStr = GET_PRECTYPE (mbboidx);
	if ((strcmp (pStr, "mbbo")) == 0) 
	{
	    foundIt = 1;
	    break;
	}
    }
    if (foundIt == 0)
    {
	logMsg ("rbTask: can not find mbbo index\n",0,0,0,0,0,0);
	taskSuspend (taskIdSelf());
    }
    logMsg ("readback task initialized with task delay of %d\n",
		(int)(delay_time / vxTicksPerSecond),0,0,0,0,0);
    while (1)
    {
	if ((recLoc = dbRecords -> papRecLoc[aoidx]) == 0)
	{
	    logMsg ("rbTask: failed to find ao chain\n",0,0,0,0,0,0);
	    taskSuspend (taskIdSelf());
	}
	recChains[AO_IDX] = (struct recChain *)recLoc -> pFirst;

	if ((recLoc = dbRecords -> papRecLoc[boidx]) == 0)
	{
	    logMsg ("rbTask: failed to find bo chain\n",0,0,0,0,0,0);
	    taskSuspend (taskIdSelf());
	}
	recChains[BO_IDX] = (struct recChain *)recLoc -> pFirst;

	if ((recLoc = dbRecords -> papRecLoc[mbboidx]) == 0)
	{
	    logMsg ("rbTask: failed to find mbbo chain\n",0,0,0,0,0,0);
	    taskSuspend (taskIdSelf());
	}
	recChains[MBBO_IDX] = (struct recChain *)recLoc -> pFirst;
	for (idx = AO_IDX; idx < LAST_IDX; idx++)
	{
	    for (chain = recChains[idx]; chain; chain = chain -> next)
	    {
		struct dbCommon		*recHdr;
		struct aoRecord		*pao;
		struct boRecord		*pbo;
		struct mbboRecord	*pmbbo;

		recHdr = (struct dbCommon *)chain -> pRecord;
		if (recHdr == 0)
		    continue;
		if (recHdr -> name[0] == 0)
		    continue;
		/* no lset, then initDatabase has not done its thing yet */
		if (recHdr -> lset == 0)
		    continue;
		/* point is not fully init'd */
		if ( (recHdr -> init == 0) || (recHdr -> ini2 == 0))
		    continue;
		/* not if we are not master */
		resGetCaAnswer (recHdr -> radr, &master);
		if (master == 0)
		    continue;
		/* execute the readback function */
		switch (idx)
		{
		    case AO_IDX:
			pao = (struct aoRecord *)chain -> pRecord;

			/* if  readback disabled */
			if (pao -> rbdt == 0)
			    break;

			if (checkEligable (idx, chain -> pRecord))
			    break;

			if (pao -> rbfn)
			    (*(pao -> rbfn))(pao);

			break;

		    case BO_IDX:
			pbo = (struct boRecord *)chain -> pRecord;
			/* if  readback disabled */
			if (pbo -> rbdt == 0)
			    break;

			if (checkEligable (idx, chain -> pRecord))
			    break;

			if (pbo -> rbfn)
			    (*(pbo -> rbfn))(pbo);

			break;

		    case MBBO_IDX:
			pmbbo = (struct mbboRecord *)chain -> pRecord;
			/* if  readback disabled */
			if (pmbbo -> rbdt == 0)
			    break;

			if (checkEligable (idx, chain -> pRecord))
			    break;

			if (pmbbo -> rbfn)
			    (*(pmbbo -> rbfn))(pmbbo);

			break;
		    default:
			break;
		}
	    }
	}
	if (delay_time == 0)
	    delay_time = DEFAULT_TD * vxTicksPerSecond;
	taskDelay (delay_time);
    }
}

/*******************************************************************************/

int	setRbDelay (int taskdelay)
{
    if (taskdelay < MINREADBACK)
    {
	logMsg ("rbDelay: task delay minimum is %d seconds\n", MINREADBACK,0,0,0,0,0);
	return (1);
    }
    else
    {
	delay_time = taskdelay * vxTicksPerSecond;
    }
    return (0);
}


/*******************************************************************************/

int	startRbTask (void)
{
    if (taskSpawn(RBTASK_NAME, RBTASK_PRI, RBTASK_OPT, RBTASK_STACK,
			(FUNCPTR)rbTask,0,0,0,0,0,0,0,0,0,0)
			== ERROR)
    {
	logMsg("Can't spawn readback task.\n",0,0,0,0,0,0);
	return (-1);
    }
    return (0);
}

