/*recBi.c,v 1.1.1.1 1999/08/03 14:44:18 tis4000 Exp */

/* recBi.c - Record Support Routines for Binary Input records */
/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            7-14-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 *
 * Modification Log:
 * -----------------
 * .01  12-12-88        lrd     lock the record on entry and unlock on exit
 * .02  12-15-88        lrd     Process the forward scan link
 * .03  12-23-88        lrd     Alarm on locked MAX_LOCKED times
 * .04  01-13-89        lrd     delete db_read_bi
 * .05  03-17-89        lrd     database link inputs
 * .06  03-29-89        lrd     make hardware errors MAJOR
 *                              remove hw severity spec from database
 * .07  04-06-89        lrd     add monitor detection
 * .08  05-03-89        lrd     removed process mask from arg list
 * .09  01-31-90        lrd     add the plc_flag arg to the ab_bidriver call
 * .10  03-21-90        lrd     add db_post_events for RVAL
 * .11  04-11-90        lrd     make local variables static
 * .12  10-10-90	mrk	Made changes for new record and device support
 * .13  03-12-93	ems	deleted ack from redundancy
 * .14  01-03-94	ems	ack = 0 in init_record
 *				in process, if not init'd do nothing
 * .15  01-05-94	ems	added special func to handle SPC_CNST
 * .16  03-01-94	ems	added support for process stats
 * .17	03-29-94	ems	make forward frunc def agree with actual func def
 * .18  04-21-94	ems	add INI2 to namelist, chg NUMFLDS to NUMBIFLDS
 *				rewrite init_record
 * .19	06-10-94	ems	moved call to tsLocalTime from after call to device
 *				to before if on scan
 * .20	06-13-94	ems	added support for maximize severity
 * .21	08-18-94	ems	monitor(), rm DBE_LOG from ack, only set DBE_LOG on val
 *				if tenb (trending) enabled
 * .22	10-12-94	ems	add TIME to redundancy, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 *				added lval for use to test COS alarm, used to use lalm
 *				but it is now used as above
 * .23	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .24	12-05-94	ems	added support for transition counter TCNT
 * .25  01-13-95        wep     add mlst to redundancy, init_record(), init lptm and latm
 * .26	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .27	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .28	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .29	10-23-95	ems	update TCNT in process not in alarm so dont need alarms
 *				enabled, also update LVAL in process not in alarm
 *				need to update LVAL like TCNT
 * .30	04-11-96	ems	added code for alarm delay
 * .31	06-19-96	ems	added code for alarm disable ADIS
 *
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<strLib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/biRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"biredfldidx.h"
#include	"restable.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
#define get_precision NULL
static long get_enum_str ();
static long get_enum_strs ();
static long put_enum_str ();
#define get_graphic_double NULL
#define get_control_double NULL
#define get_alarm_double NULL

struct rset biRSET = {
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

struct bidset {		/* binary input dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;	/*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_bi;	/*(0,1,2)=> success and */
                        /*(convert,don't continue, don't convert)*/
                        /* if convert then raw value stored in rval */
	};

/* The following definitions are required for redundancy */

static struct	fldDes	**mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"LVAL",
		"TCNT",
		"MLST",
		"HHLL",
		"RTCT",
		0
		};

static	unsigned	short	type;

static void alarm ();
static void monitor ();


/*------------------------------------------------------------------------*/

static long init_record(pbi)
    struct biRecord	*pbi;
{
    struct	bidset	*pdset;
    long		status;
    unsigned	short	i;
    char		*pstr;
    char		redbuf [256];
    int			ini2_flag = 0;
    int			do_convert;

	if (!(mylist)) 
	{
	    for (type = 1; type < 100; type++) 
	    {
		pstr = GET_PRECTYPE (type);
		if (strcmp (pstr, "bi") == 0) 
		    break;
	    }
	    mylist = redFldAlloc (type, NUMBIFLDS);

	    if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pbi,
				   "bi, init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	    for (i = 0; i < NUMBIFLDS; i++) 
		redFldAdd (type, i, namelist [i]);
	}

        if (!(pdset = (struct bidset *) (pbi->dset))) 
	{
	    recGblRecordError (S_dev_noDSET, (struct dbCommon *)pbi, "bi: init_record");
	    return(S_dev_noDSET);
        }

        /* must have read_bi function defined */

        if ( (pdset->number < 5) || (pdset->read_bi == NULL) ) 
	{
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pbi, "bi: init_record");
	    return (S_dev_missingSup);
        }

	pbi -> init = TRUE;
	pbi -> fpps = 1;
	pbi -> nsta = NO_ALARM;
	pbi -> nsev = NO_ALARM;
	pbi -> msev = NO_ALARM;
	pbi -> tcnt = 0.0;

	/* init alarm delay variables */
	pbi -> larf = 0;
	pbi -> hhll = 0;
	pbi -> hilo = 0;
	pbi -> dvct = 0;
	pbi -> rtct = 0;

	/* init last process time and last alarm time */
	if (pbi -> lptm.secPastEpoch == 0)
	    tsLocalTime (&pbi -> lptm);
	if (pbi -> latm.secPastEpoch == 0)
	    tsLocalTime (&pbi -> latm);

	do_convert = 1;

	/* do not allow conversion of rval to val */

	if ( (pbi -> dtyp <= 1) && (pbi -> inp.type == CONSTANT) )
	    do_convert = 0;

        if( pdset->init_record ) 
	{
	    status = (*pdset->init_record) (pbi, process);
	    if (status)
	    {
	        if ( (!pbi -> tagd) && (pbi -> oscn != OFF_SCAN) &&
		     (pbi -> mode == NORMAL) )
		{
		    pbi -> nsta = NO_INIT_ALARM;
		    pbi -> nsev = VALID_ALARM;
		}
		else
		    ini2_flag = 1;
	    }
	    else
	    {
		ini2_flag = 1;
		if (do_convert)
		{
		    if (pbi->rval == 0) 
		        pbi->val =0;
		    else 
		        pbi->val = 1;

		    /* Handle inversion */

		    if (pbi -> invt)
		    {
			if (pbi -> val)
			    pbi -> val = 0;
			else
			    pbi -> val = 1;
		    }
		}
		pbi->udf = FALSE;
	    }
        }
	else
	    ini2_flag = 1;

	/* init last value, use by alarm() to test COS alarm */

	pbi -> lval = pbi -> val;

	if (resGetCaAnswer(pbi -> radr, &i) == RES_OK)
	{
	    if (i)
	    {
		/* post STAT, SEVR and VAL (monitor), if we will finally init, 
		   because we have read a value from the input device, set INI2,
		   send all redund fields to slave, when slave receives INI2, 
		   it will be fully init'd also
		*/

	    	/* check for maximize severity */

	    	if (pbi -> msev > pbi -> nsev)
	    	{
		    pbi -> nsev = pbi -> msev;
		    pbi -> nsta = LINK_ALARM;
	    	}

		monitor (pbi);

		if (ini2_flag)
		{
		    pbi -> ini2 = 1;
		    redQInit ((struct dbCommon *)pbi, redbuf, sizeof (redbuf), type);
		    redQField (redbuf, mylist [STAT]);
		    redQField (redbuf, mylist [SEVR]);
		    redQField (redbuf, mylist [VAL]);
		    redQField (redbuf, mylist [LALM]);
		    redQField (redbuf, mylist [INI2]);
		    redQField (redbuf, mylist [LVAL]);
		    redQSend (redbuf);
		}
	    }
	}
        return(0);
}


/*------------------------------------------------------------------------*/

static long process (paddr)

    struct dbAddr	*paddr;
{
    struct biRecord	*pbi=(struct biRecord *)(paddr->precord);
    struct bidset	*pdset = (struct bidset *)(pbi->dset);
    char		redbuf [256];
    long		status = 0;
    unsigned short	stat   = pbi -> stat;
    unsigned short	sevr   = pbi -> sevr;
    unsigned short	val    = pbi -> val;
    unsigned short	lalm   = pbi -> lalm;
    unsigned short	lval   = pbi -> lval;
    float		tcnt   = pbi -> tcnt;
    TS_STAMP		latm   = pbi -> latm;
    double		mlst   = pbi -> mlst;
    short		hhll   = pbi -> hhll;
    short		rtct   = pbi -> rtct;
    unsigned short	monitor_mask;

	   /* if not init'd do nothing */

	if ((!pbi -> init) || (!pbi -> ini2))
	    return (0);

	if((pdset == NULL) || (pdset->read_bi == NULL)) 
	{
	    pbi->pact=TRUE;
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pbi, "read_bi");
	    return (S_dev_missingSup);
	}

	/* get time stamp for last time record was processed */

	tsLocalTime (&pbi -> lptm);

	pbi->msev = NO_ALARM;

	if (pbi -> tagd)
	{
	    pbi -> nsta = TAGGED_ALARM;
	    pbi -> nsev = VALID_ALARM;
	}
        else if (pbi->oscn == OFF_SCAN) 
	{
            pbi->nsta = SCAN_ALARM;
	    pbi->nsev = VALID_ALARM;
	    status = 0;
        } 
	else 
	{
	    pbi -> nsta = NO_ALARM;
	    pbi -> nsev = NO_ALARM;

	    /* update number of process times */
	    if (process_stats)
	       process_cnts[type]++;	

	    if (pbi -> mode == NORMAL)
	    {
	        status = (*pdset->read_bi) (pbi); /* read the new value */
	        pbi->pact = TRUE;

	        if (status == 1) 
		    return(0);
	        
	        if (status == -2) 
		{
		    pbi->pact = FALSE;
		    return (0);
		}
	        
	        if (status == 0) 
		{
		    /* convert rval to val */
		    if (pbi->rval == 0) 
		        pbi->val =0;
		    else 
		        pbi->val = 1;
		    
		    pbi->udf = FALSE;

		    /* Handle inversion */

		    if (pbi -> invt)
		    {
			if (pbi -> val)
			    pbi -> val = 0;
			else
			    pbi -> val = 1;
		    }

	        /* status is one if an asynchronous record is being processed*/

	        } 
		else if (status == 2) 
		    status=0;
	    }

	    /* check for alarms */

	    alarm (pbi);

	    /* check for maximize severity */

	    if (pbi -> msev > pbi -> nsev)
	    {
		pbi -> nsev = pbi -> msev;
		pbi -> nsta = LINK_ALARM;
	    }
	}

	/* if the time stamp is 0, init and post val with
	   alarm bit set, this makes sure that proper alarm
	   messages are set if IOC was down and alarm status
	   may have chg'd
	*/
	if (pbi -> latm.secPastEpoch == 0)
	{
	    tsLocalTime (&pbi -> latm);
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi -> val, DBE_VALUE | DBE_ALARM);
	}

	/* if it is the 1st process pass 
	   init last value when alarm chg'd (lalm)
	*/

	if (pbi -> fpps)
	{
	    pbi -> fpps = 0;
	    pbi -> lalm = pbi -> val;
	}

	/* check event list */

	monitor (pbi);

	redQInit ((struct dbCommon *)pbi, redbuf, sizeof (redbuf), type);

	if (stat != pbi->stat) 
	    redQField (redbuf, mylist [STAT]);
	
	if (sevr != pbi->sevr) 
	    redQField (redbuf, mylist [SEVR]);
	
	if (val != pbi->val) 
	    redQField (redbuf, mylist [VAL]);
	
	if (lalm != pbi->lalm) 
	    redQField (redbuf, mylist [LALM]);
	
	if (TsCmpStampsNE (&latm, &pbi -> latm) )
	    redQField (redbuf, mylist [LATM]);
	
	/* update last value for COS alarm */
	if (pbi -> val != lval)
	{
	    pbi -> lval = pbi -> val;
	    /* register a change of state in the counter */
	    pbi -> tcnt += 1.0;
	}

	if (lval != pbi->lval) 
	    redQField (redbuf, mylist [LVAL]);
	
	monitor_mask = 0;
	if (tcnt != pbi->tcnt) 
	{
	    redQField (redbuf, mylist [TCNT]);
	    monitor_mask = DBE_VALUE;
	}
	if (pbi -> tenb)
	{
	    if ( (tcnt != pbi -> tcnt) || (stat != pbi -> stat) ||
		 (sevr != pbi -> sevr) )
		monitor_mask |= DBE_LOG;
	}
	if (monitor_mask)
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi -> tcnt, monitor_mask);
	
	if (mlst != pbi -> mlst)
	    redQField (redbuf, mylist [MLST]);
	
	if (hhll != pbi -> hhll)
	    redQField (redbuf, mylist [HHLL]);
	
	if (rtct != pbi -> rtct)
	    redQField (redbuf, mylist [RTCT]);
	
	redQSend (redbuf);

	/* process the forward scan link record */

	if (pbi->flnk.type == DB_LINK) 
	    dbScanPassive ((struct dbAddr *)pbi->flnk.value.db_link.pdbAddr);
	
	pbi->pact = FALSE;
	return (status);
}


/*------------------------------------------------------------------------*/

static long special(paddr,after)
    struct dbAddr *paddr;
    int           after;
{
    struct biRecord     *pbi = (struct biRecord *)(paddr->precord);
    int                 special_type = paddr->special;
    static unsigned short save_cnst;

    /* if dtype and inp are not CONSTANT, do not let them
       update CNST
    */

    if(!after) 
    {
	save_cnst = pbi -> cnst;
	return(0);
    }

    switch(special_type) {

    case(SPC_CNST):
	if ((pbi -> dtyp <= 1 ) && (pbi -> inp.type == CONSTANT))
        {
	    pbi -> inp.value.value = pbi -> cnst;
            pbi -> val = pbi -> cnst;
        }
	else
	    pbi -> cnst = save_cnst;
        return(0);

    default:
        recGblDbaddrError(S_db_badChoice,paddr,"bi: special");
        return(S_db_badChoice);
    }
}


/*------------------------------------------------------------------------*/

static long get_value(pbi,pvdes)
    struct biRecord		*pbi;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_ENUM;
    pvdes->no_elements=1;
    (unsigned short *)(pvdes->pvalue) = &pbi->val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char	  *pstring;
{
    struct biRecord	*pbi=(struct biRecord *)paddr->precord;
    unsigned short      *field = (unsigned short *)paddr -> pfield;

    if((*field) == 0)
    {
	strncpy(pstring,pbi->znam,sizeof(pbi->znam));
	pstring[sizeof(pbi->znam)] = 0;
    } 
    else if((*field) == 1)
    {
	strncpy(pstring,pbi->onam,sizeof(pbi->onam));
	pstring[sizeof(pbi->onam)] = 0;
    } 
    else 
    {
	strcpy(pstring,"Illegal_Value");
    }
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_enum_strs(paddr,pes)
    struct dbAddr *paddr;
    struct dbr_enumStrs *pes;
{
    struct biRecord	*pbi=(struct biRecord *)paddr->precord;

    pes->no_str = 2;
    bzero((char *)(&pes->strs[0]),sizeof(pes->strs));
    strncpy(pes->strs[0],pbi->znam,sizeof(pbi->znam));
    strncpy(pes->strs[1],pbi->onam,sizeof(pbi->onam));
    return(0);
}


/*------------------------------------------------------------------------*/

static long put_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char          *pstring;
{
    struct biRecord     *pbi=(struct biRecord *)paddr->precord;
    unsigned short      *field = (unsigned short *)paddr -> pfield;

    if(strncmp(pstring,pbi->znam,sizeof(pbi->znam))==0) 
	*field = 0;
    else  if(strncmp(pstring,pbi->onam,sizeof(pbi->onam))==0) 
	*field = 1;
    else 
	return(S_db_badChoice);
    return(0);
}


/*------------------------------------------------------------------------*/

static void alarm (pbi)
    struct biRecord	*pbi;
{
	unsigned short	val = pbi->val;
	int		adis = 0;

        /* check for alarm inhibit */
	adis = AlarmDisableLink ((void *)pbi);
        if ( (pbi->aenb == ALARM_DISABLED) || (adis != 0) )
	{
	    if (pbi -> nsta != NO_INIT_ALARM)
	    {
                pbi->nsta = INHIBIT_ALARM;
                pbi->nsev = NO_ALARM;
	    }
            return;
        }

        if(pbi->udf == TRUE)
	{
            if (pbi->nsev < VALID_ALARM)
	    {
                pbi->nsta = UDF_ALARM;
                pbi->nsev = VALID_ALARM;
            }
            return;
        }

	/* check for hardware alarms */

	if (pbi -> nsev == VALID_ALARM)
		return;

	if(val > 1) 
	    return;

        if (val == 0)
	{
            if (pbi->nsev < pbi->zsv) 
	    {
                pbi->nsta = STATE_ALARM;
                pbi->nsev = pbi->zsv;
            }
        } 
	else 
	{
            if (pbi->nsev < pbi->osv)
	    {
                pbi->nsta = STATE_ALARM;
                pbi->nsev = pbi->osv;
            }
        }

	/* check for output of state alarm */

	AlarmStates ((void *)pbi, 0);

        /* check for cos alarm */

	if (val != pbi -> lval)
	{
	    if (pbi -> nsev < pbi -> cosv)
	    {
		pbi -> nsta = COS_ALARM;
		pbi -> nsev = pbi -> cosv;
	    }
	}

	/* check for return to normal */

	AlarmStates ((void *)pbi, 1);

	return;
}


/*------------------------------------------------------------------------*/

static void monitor (pbi)
    struct biRecord	*pbi;
{
	unsigned short	monitor_mask;
        short           stat,sevr,nsta,nsev;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/

        stat = pbi->stat;
        sevr = pbi->sevr;
        nsta = pbi->nsta;
        nsev = pbi->nsev;

        /*set current stat and sevr*/

        pbi->stat = nsta;
        pbi->sevr = nsev;
        pbi->nsta = 0;
        pbi->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == pbi -> psta) && (nsev == pbi -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    pbi -> psta = stat;
	    pbi -> psev = sevr;
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi -> psev, DBE_VALUE);
	}

	monitor_mask = 0;

	/* alarm condition changed this scan */

	/*
	EXTRA PATCH: Some users may have both alarm states
	on a digital input have the same alarm priority. If
	the VAL field is monitored with the mask == DBE_ALARM,
	then they will never see the changes. This is a way
	to make sure that won't happen.
	*/

	if (((nsta == STATE_ALARM) || (nsta == COS_ALARM)) && 
	    (pbi -> mlst != pbi -> val) )
		stat = ~nsta;

	if ((stat != nsta) || (sevr != nsev)) 
	{
	    /* post events for alarm condition change*/

	    monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&pbi -> latm);
	    pbi -> lalm = pbi -> val;

	    /* post stat and sevr fields */

	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi->stat, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi->sevr, DBE_VALUE);

            if ((pbi->sevr != NO_ALARM) && (set_ack) &&
		(pbi->sevr != EVENT_ALARM) )  
	    {
                if (pbi->ack == 0)
		{
                    pbi->ack = 1;           /* new unacknowledged alarm */
                    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi->ack, DBE_VALUE);
                }
            }
	}
        /* check for value change */

        if (pbi->mlst != pbi->val)
	{
            /* post events for value change and archive change */
            monitor_mask |= DBE_VALUE;

        }

	if (pbi -> tenb)
	{
	    /* if VAL has changed or there has been a change in alarm status */
	    if ( (pbi -> mlst != pbi -> val) || (monitor_mask & DBE_ALARM) )
		monitor_mask |= DBE_LOG;
	}
	if (monitor_mask & (DBE_VALUE | DBE_LOG))
	    pbi -> mlst = pbi -> val;


	/* send out monitors connected to the value field */

	if (monitor_mask) 
	{
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi->val, monitor_mask);
	}
	
	if (pbi->oraw != pbi->rval) 
	{
	    db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi->rval, DBE_VALUE);
	    pbi->oraw = pbi->rval;
	}
	return;
}
