/*recSel.c,v 1.1.1.1 1999/08/03 14:44:21 tis4000 Exp */

/* recSel.c - Record Support Routines for Select records */
/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            6-2-89
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (A	T-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  11-16-89        lrd     fixed select algorithms not to compare against
 *                              the previous value
 * .02  10-12-90	mrk	changes for new record support
 * .03  01-04-94	ems	set INIT = 1 and ACK = 0 in init_record
 *				do nothing in process if not init'd
 * .04  02-23-94	ems	added rate of change logic
 * .05  03-01-94	ems	added support for process stats
 *				removed ack = 0
 * .06  03-10-94	ems	alarm chgs, if val is past alarm setting, set that
 *				alarm condition, else if val was in that alarm and
 *				it is still in hyst range of alarm value, then stay
 *				in alarm, remove return from high alarm, let user
 *				determine which is more important low or high
 * .07	03-29-94	ems	make forward func defs agree with actual defs
 * .08  04-04-94	ems	init roci to 0, redund cant handle -1, rework bumping of
 *				roci in alarm, make sure does not exceed MAX_ROC_VALEUS
 * .09  05-06-94	ems	rewrite init_record, add ack to monitor, 
 *				get_control_double, if field is HYST, high = hopr-lopr
 * .10	06-14-94	ems	added support for maximize severity, moved call to 
 *				tsLocalTime inside on scan, moved process_cnts inside
 *				inside on scan
 * .11	08-08-94	ems	added field types to get_graphic_double and
 *				get_control_double
 * .12	08-18-94	ems	monitor(), rm DBE_LOG from ack, archive val only if
 *				tenb (trending) is set, set DBE_VALUE only on input
 *				values
 * .13	10-12-94	ems	add TIME to redundancy, fpas chg'd to frps, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .12	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .13	01-13-95	ems	added mlst to redundancy, init latm and lptm in init_record
 * .14	02-06-95	ems	added LSLN so can trend SELN
 * .15	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .16	09-11-95	ems	archive seln/val if alarm stat/sevr chgs
 * .17	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .18	09-22-95	ems	alarm, if alarms disabled do a return
 * .19	02-06-96	ems	added code for deviation alarm
 *				always update rate of chg buffer as soon as alarms are enabled
 *				    so will always have latest data, also check if rate of chg
 *				    value is > 0.0 before process roc alarm
 *				remove return from HIHI and LOLO alarms, so will always check 
 *				    all alarm sevrs
 * .20	04-04-96	ems	added code to support alarm delay
 * .21	06-20-96	ems	added code for alarm disable ADIS
 * .22	07-02-96	ems	alarm(), update rocb if roc > 0.0 or frps is set
 * .23	07-02-96	ems	process(), compare msev >= to nsev rather than >, so msev
 *				can be expressed if one of the inputs are bad
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>
#include	<stdlib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"redflds.h"
#include	"redque.h"
#include	"rec/selRecord.h"
#include	"selredfldidx.h"
#include	"restable.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"
#include	"okToUse.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/
#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
#define special NULL
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units ();
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
static long get_alarm_double ();

struct rset selRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double };

#define	SEL_MAX	12
#define SELECTED	0
#define SELECT_HIGH	1
#define SELECT_LOW	2
#define SELECT_MEDIAN	3

static unsigned short	type;

/* The folowing definitions are for redundancy */
 
static struct fldDes	**mylist = 0;

static	char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"ROCI",
		"ROCC",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		"DVCT",
		0
		};

static void alarm();
static void monitor();
static int fetch_values();
static int do_sel();



/*------------------------------------------------------------------------*/

static long init_record (psel)
    struct selRecord     *psel;
{
    struct link		*plink;
    unsigned short	i;
    double		*pvalue;
    char		*pstr;
    char		redbuf[256];
    int			ini2_flag = 0;


    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "sel", 3) == 0) 
		break;
	}

	mylist = redFldAlloc (type, NUMSELFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)psel,
				"sel, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMSELFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

	/* allocate rate of change buffer */
    if (!(psel -> rocb))
    {
	psel -> rocb = (char *) calloc (1, (MAX_ROC_VALUES * sizeof (float)));
	if (!psel -> rocb)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)psel,
				   "sel, init_record, rocb failed");
	    return (S_rec_Alloc);
	}
	psel -> rocc = 0;
	psel -> roci = 0;
    }

    psel -> init = 1;
    psel -> fpps = 1;
    psel -> nsta = NO_ALARM;
    psel -> nsev = NO_ALARM;
    psel -> msev = NO_ALARM;
    psel -> frps = TRUE;

    /* init alarm delay variables */
    psel -> larf = 0;
    psel -> hhll = 0;
    psel -> hilo = 0;
    psel -> dvct = 0;
    psel -> rtct = 0;

    /* init last alarm time and last process time */
    if (psel -> latm.secPastEpoch == 0)
	tsLocalTime (&psel -> latm);
    if (psel -> lptm.secPastEpoch == 0)
	tsLocalTime (&psel -> lptm);

    /* get seln initial value if nvl is a constant*/
    if (psel->nvl.type == CONSTANT ) 
	psel->seln = psel->nvl.value.value;

    plink = &psel->inpa;
    pvalue = &psel->a;
    for(i=0; i<psel -> ninp; i++, plink++, pvalue++) 
    {
	if (plink->type==CONSTANT) 
	    *pvalue = plink->value.value;
    }

    /* if we are tagged, off scan or in manual mode we are not
       going to calculate val, just let the block init
    */
    if ( (!psel -> tagd) && (psel -> oscn != OFF_SCAN) &&
	 (psel -> mode == NORMAL) )
    {
	if (fetch_values (psel) == 0)
	{
	    if (do_sel (psel) == 0)
		ini2_flag = 1;
	    else
	    {
		psel -> nsta = NO_INIT_ALARM;
		psel -> nsev = VALID_ALARM;
	    }
	}
	else
	{
	    psel -> nsta = NO_INIT_ALARM;
	    psel -> nsev = VALID_ALARM;
	}
    }
    else
        ini2_flag = 1;

    psel -> lsln = psel -> seln;

    /* if we are MASTER post VAL and send it to redundant partner */

    if (resGetCaAnswer (psel -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* post STAT, SEVR and VAL (monitor), if we will finally init, 
   	       because we have read a value from the input device, set INI2,
	       send all redund fields to slave, when slave receives INI2, 
	       it will be fully init'd also
	    */

	    /* check for maximize severity */

	    if ( (psel -> msev > NO_ALARM) && (psel -> msev >= psel -> nsev) )
	    {
		psel -> nsev = psel -> msev;
		psel -> nsta = LINK_ALARM;
	    }

	    monitor (psel);

	    if (ini2_flag)
	    {
		psel -> ini2 = 1;
		redQInit ((struct dbCommon *)psel, redbuf, sizeof (redbuf), type);

		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [LALM]);
		redQField (redbuf, mylist [ROCI]);
		redQField (redbuf, mylist [ROCC]);
		redQField (redbuf, mylist [INI2]);

		redQSend (redbuf);
	    }
	}
    }

    return(0);
}


/*------------------------------------------------------------------------*/

static long process (paddr)
    struct dbAddr	*paddr;
{
    struct selRecord	*psel=(struct selRecord *)(paddr->precord);
    short		val_sent = 0;
    unsigned short	stat = psel -> stat;
    unsigned short	sevr = psel -> sevr;
    double		val  = psel -> val;
    short		roci = psel -> roci;
    short		rocc = psel -> rocc;
    double		lalm = psel -> lalm;
    TS_STAMP		latm = psel -> latm;
    double		mlst = psel -> mlst;
    short		hhll = psel -> hhll;
    short		hilo = psel -> hilo;
    short		rtct = psel -> rtct;
    unsigned long	larf = psel -> larf;
    short		dvct = psel -> dvct;
    char		redbuf [256];
    unsigned short	monitor_mask;

    /* if not init'd, do nothing */

    if ((!psel -> init) || (!psel -> ini2))
	return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&psel -> lptm);

    psel -> msev = NO_ALARM;

    /* off normal operations */

    if (psel -> tagd)
    {
	psel -> nsta = TAGGED_ALARM;
	psel -> nsev = VALID_ALARM;
    }
    else if (psel -> oscn == OFF_SCAN)
    {
	psel -> nsta = SCAN_ALARM;
	psel -> nsev = VALID_ALARM;
    }
    else
    {
	psel -> nsta = NO_ALARM;
	psel -> nsev = NO_ALARM;

	/* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	if (psel -> mode == NORMAL)
	{
	    psel->pact = TRUE;
	    if(fetch_values(psel) == 0) 
	    {
		if(do_sel(psel) != 0) 
		{
		    if(psel->nsev < VALID_ALARM) 
		    {
			psel->nsta = CALC_ALARM;
			psel->nsev = VALID_ALARM;
		    }
		}
	    }
	}

	/* check for alarms */
	alarm(psel);

	/* check for maximize severity */

	if ( (psel -> msev > NO_ALARM) && (psel -> msev >= psel -> nsev) )
	{
	    psel -> nsev = psel -> msev;
	    psel -> nsta = LINK_ALARM;
	}
    }

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (psel -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&psel -> latm);
	db_post_events ((struct dbCommon *)psel, (caddr_t)&psel -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (psel -> fpps)
    {
	psel -> fpps = 0;
	psel -> lalm = psel -> val;
    }

    /* check event list */
    monitor(psel);

    /* process the forward scan link record */
    if (psel->flnk.type==DB_LINK)
	dbScanPassive((struct dbAddr *)psel->flnk.value.db_link.pdbAddr);

    psel->pact=FALSE;

    monitor_mask = 0;
    if (psel -> lsln != psel -> seln)
    {
	monitor_mask = DBE_VALUE;
    }
    if (psel -> tenb)
    {
	if ( (psel -> lsln != psel -> seln) || (stat != psel -> stat) ||
	     (sevr != psel -> sevr) )
	    monitor_mask |= DBE_LOG;
    }

    if (monitor_mask & (DBE_VALUE | DBE_LOG))
	psel -> lsln = psel -> seln;

    if (monitor_mask)
	db_post_events ((struct dbCommon *)psel, (caddr_t)&psel -> seln, monitor_mask);

    redQInit ((struct dbCommon *)psel, redbuf, sizeof (redbuf), type);

    if (stat != psel->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != psel->sevr) 
	redQField (redbuf, mylist [SEVR]);
	
    if (val != psel->val) 
    {
	val_sent = 1;
	redQField (redbuf, mylist [VAL]);
    }
	
    if (roci != psel->roci) 
    {
	if (!val_sent)
	    redQField (redbuf, mylist [VAL]);
	redQField (redbuf, mylist [ROCI]);
    }

    if (rocc != psel->rocc) 
	redQField (redbuf, mylist [ROCC]);
	
    if (lalm != psel->lalm) 
	redQField (redbuf, mylist [LALM]);

    if (TsCmpStampsNE (&latm, &psel -> latm))
	redQField (redbuf, mylist [LATM]);
	
    if (mlst != psel->mlst) 
	redQField (redbuf, mylist [MLST]);

    if (hhll != psel -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (hilo != psel -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (rtct != psel -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (larf != psel -> larf)
	redQField (redbuf, mylist [LARF]);

    if (dvct != psel -> dvct)
	redQField (redbuf, mylist [DVCT]);
	
    redQSend (redbuf);

    return(0);
}


/*------------------------------------------------------------------------*/

static long get_value (psel,pvdes)
    struct selRecord		*psel;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_DOUBLE;
    pvdes->no_elements=1;
    (double *)(pvdes->pvalue) = &psel->val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_units (paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    struct selRecord	*psel=(struct selRecord *)paddr->precord;

    strncpy(units,psel->egu,sizeof(psel->egu));
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_precision (paddr,precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    struct selRecord	*psel=(struct selRecord *)paddr->precord;

    *precision = psel->prec;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double (paddr,pgd)
    struct dbAddr *paddr;
    struct dbr_alDouble	*pgd;
{
    struct selRecord	*psel=(struct selRecord *)paddr->precord;

    pgd->upper_alarm_limit = psel->hihi;
    pgd->upper_warning_limit = psel->high;
    pgd->lower_warning_limit = psel->low;
    pgd->lower_alarm_limit = psel->lolo;
    return(0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double (paddr,pgd)
    struct dbAddr *paddr;
    struct dbr_grDouble	*pgd;
{
    struct selRecord	*psel=(struct selRecord *)paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct selRecord *)0) -> seln) ):
	    pgd -> upper_disp_limit = 11;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct selRecord *)0) -> hyst) ):
	case ( (int)&(((struct selRecord *)0) -> roc) ):
	case ( (int)&(((struct selRecord *)0) -> adel) ):
	case ( (int)&(((struct selRecord *)0) -> mdel) ):
	case ( (int)&(((struct selRecord *)0) -> devl) ):
	{
	    float    span;

	    span = psel -> hopr - psel -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;
	}

	case ( (int)&(((struct selRecord *)0) -> rocn) ):
	    pgd -> upper_disp_limit = 60;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct selRecord *)0) -> ninp) ):
	    pgd -> upper_disp_limit = 12;
	    pgd -> lower_disp_limit = 2;
	    break;

	default:
	    pgd -> upper_disp_limit = psel -> hopr;
	    pgd -> lower_disp_limit = psel -> lopr;
	    break;
    }

    return(0);
}


/*------------------------------------------------------------------------*/

static long get_control_double (paddr,pcd)
    struct dbAddr *paddr;
    struct dbr_ctrlDouble *pcd;
{
    struct selRecord	*psel=(struct selRecord *)paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct selRecord *)0) -> seln) ):
	    pcd -> upper_ctrl_limit = 11;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct selRecord *)0) -> hyst) ):
	case ( (int)&(((struct selRecord *)0) -> roc) ):
	case ( (int)&(((struct selRecord *)0) -> adel) ):
	case ( (int)&(((struct selRecord *)0) -> mdel) ):
	case ( (int)&(((struct selRecord *)0) -> devl) ):
	{
	    float    span;

	    span = psel -> hopr - psel -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	case ( (int)&(((struct selRecord *)0) -> rocn) ):
	    pcd -> upper_ctrl_limit = 60;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct selRecord *)0) -> ninp) ):
	    pcd -> upper_ctrl_limit = 12;
	    pcd -> lower_ctrl_limit = 2;
	    break;

	default:
	    pcd -> upper_ctrl_limit = psel -> hopr;
	    pcd -> lower_ctrl_limit = psel -> lopr;
	    break;
    }

    return (0);

}


/*------------------------------------------------------------------------*/

static void alarm (psel)
    struct selRecord	*psel;
{
    double		val = psel -> val;
    register float	hval, lval;
    register short	i, inx;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* add value to the rate of change buffer, 
       if there is a rate of chg limit or 1st pass is set
    */

    if ( (psel -> roc > 0.0) || (psel -> frps == FALSE) )
    {
	psel -> roci++;
	if (psel->roci >= MAX_ROC_VALUES) 
	{
	    psel->roci = 0;
	}
	if (psel->rocc < MAX_ROC_VALUES)
	{
	    psel->rocc++;
	}
	if (psel -> frps)
	{
	    psel -> frps = FALSE;
	    *(float *) (psel -> rocb) = val;
	}
	*(float *) (psel->rocb + (psel->roci * sizeof (float))) = val;
    }

    /* check for alarm inihibit */

    adis = AlarmDisableLink ((void *)psel);

    if ( (psel -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (psel -> nsta != NO_INIT_ALARM)
	{
	    psel -> nsta = INHIBIT_ALARM;
	    psel -> nsev = NO_ALARM;
	}
	return;
    }

    /* undefined condition */
    if(psel->udf ==TRUE)
    {
	if (psel->nsev<VALID_ALARM)
	{
	    psel->nsta = UDF_ALARM;
	    psel->nsev = VALID_ALARM; 
	    return;
	}
    }

    /* rate of change alarms */

    if ( (psel->rcsv != NO_ALARM) && (psel->rocb) && (psel -> roc > 0.0) ) 
    {
	if (psel->rocc >= 2) 
	{
	    inx = psel->roci;
	    hval = lval =  *(float *) (psel->rocb + (inx * sizeof (float)));
	    for (i = 0; 
		 i < ((psel -> rocc < psel->rocn) ? psel -> rocc : psel -> rocn); 
		 i++) 
	    {
		if (--inx < 0) 
		{
		    inx = MAX_ROC_VALUES - 1;
		}
		if ((*((float *) (psel->rocb + (inx * sizeof (float))))) > hval) 
		{
		    hval = *((float *) (psel->rocb + (inx * sizeof (float))));
		}
		if ((*((float *) (psel->rocb + (inx * sizeof (float))))) < lval) 
		{
		    lval = *((float *) (psel->rocb + (inx * sizeof (float))));
		}
	    }
	    if ((hval - lval) > psel->roc) 
	    {
		psel->nsev = psel->rcsv;
		psel->nsta = ROC_ALARM;
	    }
	}
    }

    /* set up for limit alarms */

    alarm_limits.hihi = psel -> hihi;
    alarm_limits.high = psel -> high;
    alarm_limits.low  = psel -> low;
    alarm_limits.lolo = psel -> lolo;
    alarm_limits.devl = psel -> devl;
    alarm_limits.devs = psel -> devs;
    alarm_limits.hyst = psel -> hyst;
    alarm_limits.hhsv = psel -> hhsv;
    alarm_limits.hsv  = psel -> hsv;
    alarm_limits.lsv  = psel -> lsv;
    alarm_limits.llsv = psel -> llsv;
    alarm_limits.dvsv = psel -> dvsv;

    AlarmLimits (&alarm_limits, val, (void *)psel);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((psel -> nsta == NO_ALARM) && (psel -> nsev == NO_ALARM)) &&
	 ((psel -> stat == HIHI_ALARM) || (psel -> stat == HIGH_ALARM) ||
	  (psel -> stat == LOW_ALARM)  || (psel -> stat == LOLO_ALARM) ||
	  (psel -> stat == DEVIATION_ALARM) ) )
    {
	if (psel -> aldl > 0)
	{
	    if (psel -> rtct < ALDL_MAX)
		psel -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (psel -> rtct < psel -> aldl)
	    {
		psel -> nsta = psel -> stat; 
		psel -> nsev = psel -> sevr;
	    }
	}
    }
    else
    {
	if (psel -> aldl > 0)
	{
	    psel -> rtct = 0;
	}
    }

    return;
}


/*------------------------------------------------------------------------*/

static void monitor (psel)
    struct selRecord	*psel;
{
	unsigned short	monitor_mask;
	double		delta;
        short           stat,sevr,nsta,nsev;
	double          *pnew;
	double          *pprev;
	int             i;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/
        stat=psel->stat;
        sevr=psel->sevr;
        nsta=psel->nsta;
        nsev=psel->nsev;

        /*set current stat and sevr*/
        psel->stat = nsta;
        psel->sevr = nsev;
        psel->nsta = 0;
        psel->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == psel -> psta) && (nsev == psel -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    psel -> psta = stat;
	    psel -> psev = sevr;
	    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel -> psev, DBE_VALUE);
	}

        monitor_mask = 0;

        /* alarm condition changed this scan */
        if (stat!=nsta || sevr!=nsev) 
	{
	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&psel -> latm);
	    psel -> lalm = psel -> val;

	    /* post events for alarm condition change*/
	    monitor_mask = DBE_ALARM;

	    /* post stat and nsev fields */
	    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel->stat,DBE_VALUE);
	    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel->sevr,DBE_VALUE);
	    if ( (psel -> sevr != NO_ALARM) && (set_ack) &&
		 (psel -> sevr != EVENT_ALARM) )
	    {
		if (psel -> ack == 0)
		{
		    psel -> ack = 1;
		    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel ->ack, DBE_VALUE);
		}
	    }
        }

        /* check for value change */
        delta = psel->mlst - psel->val;
        if(delta<0.0) delta = -delta;
        if (delta > psel->mdel) 
	{
	    /* post events for value change */
	    monitor_mask |= DBE_VALUE;

	    /* update last value monitored */
	    psel->mlst = psel->val;
        }

        /* check for archive change if tenb (trending) is enabled */

	if (psel -> tenb)
	{
            delta = psel->alst - psel->val;
            if(delta<0.0) 
		delta = -delta;
            if ( (delta > psel->adel) || (monitor_mask & DBE_ALARM) )
	    {
	    	/* post events on value field for archive change */

	    	monitor_mask |= DBE_LOG;

	    	/* update last archive value monitored */

	    	psel->alst = psel->val;
	    }
        }

        /* send out monitors connected to the value field */

        if (monitor_mask)
	{
	    db_post_events ((struct dbCommon *)psel, (caddr_t)&psel->val,monitor_mask);

	    /* check all input fields for changes*/

	    for(i=0, pnew=&psel->a, pprev=&psel->la; i<psel -> ninp; i++, pnew++, pprev++) 
	    {
		if(*pnew != *pprev) 
		{
		    db_post_events ((struct dbCommon *)psel, (caddr_t)pnew,DBE_VALUE);
		    *pprev = *pnew;
		}
	    }
	}
        return;
}


/*------------------------------------------------------------------------*/

static int do_sel (psel)
struct selRecord *psel;  /* pointer to selection record  */
{
	double		*pvalue;
	struct link	*plink;
	double		order[SEL_MAX];
	unsigned short	order_inx;
	unsigned short	i,j;

	/* selection mechanism */
	pvalue = &psel->a;
	switch (psel->selm)
	{
	    case (SELECTED):
		psel->val = *(pvalue+psel->seln);
		break;

	    case (SELECT_HIGH):
		psel->val = *pvalue;
		for (i = 0; i < psel -> ninp; i++,pvalue++)
		{
		    if (psel->val < *pvalue)
			psel->val = *pvalue;
		}
		break;

	    case (SELECT_LOW):
		psel->val = *pvalue;
		for (i = 0; i < psel -> ninp; i++,pvalue++)
		{
		    if (psel->val > *pvalue)
			psel->val = *pvalue;
		}
		break;

	    case (SELECT_MEDIAN):
		/* order only those fetched from another record */
		plink = &psel->inpa;
		order_inx = 0;
		for (i = 0; i < psel -> ninp; i++,pvalue++,plink++)
		{
		    if (plink->type == DB_LINK)
		    {
			j = order_inx;
			while ((order[j-1] > *pvalue) && (j > 0))
			{
			    order[j] = order[j-1];
			    j--;
			}
			order[j] = *pvalue;
			order_inx++;
		    }
		}
		psel->val = order[order_inx/2];
		break;

	    default:
		return(-1);
	}

	psel->udf=FALSE;
	return(0);
}


/*------------------------------------------------------------------------*/

/*
 * FETCH_VALUES
 *
 * fetch the values for the variables from which to select
 */
static int fetch_values (psel)
struct selRecord *psel;
{
	long		nRequest;
	long		options=0;
	struct link	*plink;
	double		*pvalue;
	int		i;
	long		status;

	plink = &psel->inpa;
	pvalue = &psel->a;
	/* If select mechanism is SELECTED only get selected input*/
	if(psel->selm == SELECTED) 
	{
            /* fetch the select index */
	    if(psel->nvl.type == DB_LINK )
	    {
		options=0;
		nRequest=1;

		status = dbGetLink(&(psel->nvl.value.db_link),
				   (struct dbCommon *)psel,
				   DBR_USHORT, (caddr_t)&(psel->seln),
				   &options, &nRequest);
		if (status != 0)
		{
		    if (psel->nsev<VALID_ALARM) 
		    {
			psel->nsta = LINK_ALARM;
			psel->nsev = VALID_ALARM;
		    }
		    return(-1);
		}
		else if (!psel -> ini2)
		{
		    if (!okToUse (&psel -> nvl))
			return (-1);
		}
	    }

	    plink += psel->seln;
	    pvalue += psel->seln;

	    if(plink->type==DB_LINK) 
	    {
		nRequest=1;
		status=dbGetLink(&plink->value.db_link,
				 (struct dbCommon *)psel,
				 DBR_DOUBLE, (caddr_t)pvalue,
				 &options,&nRequest);
		if(status != 0) 
		{
		    if(psel->nsev<VALID_ALARM) 
		    {
			psel->nsev=VALID_ALARM;
			psel->nsta=LINK_ALARM;
		    }
		    return(-1);
		}
		else if (!psel -> ini2)
		{
		    if (!okToUse (plink))
			return (-1);
		}
	    }
	    return(0);
	}

	/* fetch all inputs*/
	for(i=0; i<psel -> ninp; i++, plink++, pvalue++) 
	{
	    if(plink->type==DB_LINK) 
	    {
		nRequest=1;
		status = dbGetLink(&plink->value.db_link,
				   (struct dbCommon *)psel,
				   DBR_DOUBLE, (caddr_t)pvalue,
				   &options,&nRequest);
		if(status != 0) 
		{
		    if(psel->nsev<VALID_ALARM) 
		    {
			psel->nsev=VALID_ALARM;
			psel->nsta=LINK_ALARM;
		    }
		    return(-1);
		}
		else if (!psel -> ini2)
		{
		    if (!okToUse (plink))
			return (-1);
		}
	    }
	}
	return(0);
}
