/****
*
*	File name :		l1_digi.c
*	Author :		Steve Johnson
*	Creation date :	September 16, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		digi_init
*		digi_connect
*		digi_publish
*		digi_accept
*		digi_write
*		digi_poll
*		digi_read
*		digi_clear
*		digi_close
*		digi_exit
*
*	Physical I/O layer for Digiboard Acceleport 8E
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslevel1.h>
#include <sstimer.h>
#include <ssl1digi.h>
#include "digibios.h"
#include "digifep.h"

#define LEVEL1_GROUP		"LEVEL1_DIGI"

#define CHANNELS_PER_BOARD	8

#define DATA_BITS_5			0x0000
#define DATA_BITS_6			0x0010
#define DATA_BITS_7			0x0020
#define DATA_BITS_8			0x0030

#define STOP_BITS_1			0x0000
#define STOP_BITS_2			0x0040

#define PARITY_NONE			0x0000
#define PARITY_EVEN			0x0100
#define PARITY_ODD			0x0300

#define BAUD_RATES_FAST		0x0400
#define BAUD_RATES_SLOW		0x0000

#define DTR					0x80
#define CTS					0x20
#define DSR					0x10
#define CD					0x08
#define RTS					0x02

#define DIGI_RELEASE_RESET	0x00
#define DIGI_RESET			0x04
#define DIGI_HOLD_RESET		0x05
#define DIGI_ENABLE_MEMORY	0x06

#define DIGI_MAP_SIZE		0x2000
#define DIGI_EVENTS			0x0800
#define DIGI_BIOS_STATUS	0x0C00
#define DIGI_NUM_CHANNELS	0x0C22
#define DIGI_BIOS_MOVE		0x0C40
#define DIGI_FEP_ACK		0x0D20
#define DIGI_COMMAND_IN		0x0D10
#define DIGI_EVENT_IN		0x0D18
#define DIGI_EVENT_OUT		0x0D1A
#define DIGI_EVENT_DATA		0x0D1C
#define DIGI_EVENT_MASK		0x0D1E
#define DIGI_BIOS_AREA		0x1800
#define DIGI_BUFFER_SPACE	0x0400
#define DIGI_INTERRUPTS		0x0E04

#define I8259EOI		0x20
#define I8259OCW1		0x21
#define I8259OCW1A		0xA1
#define I8259OCW2		0x20
#define I8259OCW2A		0xA0

#define VectorIRQ3		0x0B
#define VectorIRQ5		0x0D
#define VectorIRQ7		0x0F
#define VectorIRQ10		0x72
#define VectorIRQ11		0x73
#define VectorIRQ12		0x74
#define VectorIRQ15		0x76
#define MaskIRQ3		0xF7
#define MaskIRQ5		0xDF
#define MaskIRQ7		0x7F
#define MaskIRQ10		0xFB
#define MaskIRQ11		0xF7
#define MaskIRQ12		0xEF
#define MaskIRQ15		0x7F

#define WRITE_TIMEOUT	3000		/* 3 seconds */

/****
*
*	Device context definitions
*
****/

typedef struct {
	QUEUE_DEF	link;
	UCHAR		*base_memory;
	void		(INTERRUPT *old_irq_int) ();
	short		base_io_port;
	short		base_irq;
	UCHAR		old_value8259;
	UCHAR		irq_vector;
	UCHAR		irq_mask;
	char		name [MAX_STRING];
} BOARD_DEF;

typedef struct {
	QUEUE_DEF	link;
	BOARD_DEF	*board;
	USHORT		*read_in;
	USHORT		*read_out;
	UCHAR		*read_overrun;
	UCHAR		*read_data;
	USHORT		*write_in;
	USHORT		*write_out;
	UCHAR		*write_data;
	ULONG		write_count;
	ULONG		read_count;
	USHORT		read_map;
	USHORT		read_mask;
	USHORT		write_map;
	USHORT		write_mask;
	long		write_timeout;
	UCHAR		channel_number;
	UCHAR		board_number;
	UCHAR		rts_on;
	char		name [MAX_STRING];
} CHANNEL_DEF;

static QUEUE_DEF		root_channel;
static QUEUE_DEF		root_board;
static int				int_count1;
static int				int_count2;
static int				int_channel;
static int				int_event;
static int				int_status;
static int				int_prior;

static CHANNEL_DEF		*add_channel (char *);
static void				delete_channel (CHANNEL_DEF *);
static BOARD_DEF		*add_board (char *);
static void				delete_board (BOARD_DEF *);
static BOARD_DEF		*find_board (char *);
static int				test_memory (UCHAR *);
static int				hardware_init (BOARD_DEF *);
static int				load_bios (short, UCHAR *);
static int				load_fep (short, UCHAR *);
static int				map_buffers (BOARD_DEF *);
static void				delay (int);
static size_t			get_ring_size (CHANNEL_DEF *);
static int				wrong_map (CHANNEL_DEF *, USHORT);
static int				set_baud_rate (CHANNEL_DEF *);
static void				set_modem_control (CHANNEL_DEF *, UCHAR, UCHAR);
static void				set_flow_control (CHANNEL_DEF *);

/****
*
****/

LEVEL1_STATUS digi_init (void)
{
	TRACE ("digi_init, entry point");
	queue_init (&root_channel);
	queue_init (&root_board);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	char			digi_group [MAX_STRING];
	int				board_number;
	int				channel_number;
	USHORT			mem_offset;
	USHORT			*data_buffer;
	UCHAR			*base_memory;
	USHORT			channel_offset;
	BOARD_DEF		*board;
	CHANNEL_DEF		*channel;
	int				channels_per_board;
	USHORT			digi_memory_seg;
	long			write_timeout;
	char			digi_prefix [MAX_STRING];
	char			message [MAX_STRING];

	TRACE ("digi_connect, entry point");
	if (strlen (name) >= MAX_STRING) {
		ERROR_TRACE_S ("digi_connect, invalid channel name = (%s)", name);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_NAME;
	}
	if (!ini_get_str (LEVEL1_GROUP, "DIGI_PREFIX", digi_prefix)) {
		ERROR_TRACE_S ("digi_connect, no DIGI_PREFIX for (%s)", LEVEL1_GROUP);
		return LEVEL1_BAD_NAME;
	}
	write_timeout = ini_get_long (LEVEL1_GROUP, "WRITE_TIMEOUT", WRITE_TIMEOUT);
	channels_per_board = ini_get_int (LEVEL1_GROUP
								    , "CHANNELS_PER_BOARD"
								    , CHANNELS_PER_BOARD);
	digi_memory_seg = (USHORT) ini_get_hex (LEVEL1_GROUP
										  , "MEMORY_SEGMENT"
										  , 0xF000L);
	board_number = ini_get_int (name, "BOARD_NUMBER", 0xFFFF);
	if (board_number == 0xFFFF) {
		ERROR_TRACE_S ("digi_connect, no BOARD_NUMBER = (%s)", name);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_NAME;
	}
	channel_number = ini_get_int (name, "CHANNEL_NUMBER", 0xFFFF);
	if (channel_number == 0xFFFF) {
		ERROR_TRACE_S ("digi_connect, no CHANNEL_NUMBER = (%s)", name);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_NAME;
	}
	if (channel_number >= channels_per_board) {
		ERROR_TRACE_I ("digi_connect, invalid channel = %d", channel_number);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_NAME;
	}
	sprintf (digi_group, digi_prefix, board_number);
	board = find_board (digi_group);
	if (!board) {
		board = add_board (digi_group);
		if (!board) {
			ERROR_TRACE_S ("digi_connect, no board = (%s)", name);
			*user_handle = (LEVEL1_HANDLE) NULL;
			return LEVEL1_NO_MEMORY;
		}
	}
	TRACE_S ("digi_init, digi_group = (%s)", digi_group);
	channel = add_channel (name);
	if (!channel) {
		ERROR_TRACE ("digi_connect, add_channel failed");
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	channel -> channel_number = (UCHAR) channel_number;
	channel -> board_number = (UCHAR) board_number;
	channel -> board = board;
	channel_offset = (USHORT) (0x1000 | (0x80 * channel_number));
	base_memory = board -> base_memory;

	data_buffer = (USHORT *) &base_memory [0x08 + channel_offset];
	mem_offset = (data_buffer [0] - digi_memory_seg) << 4;
	channel -> write_data = &base_memory [mem_offset & 0x1FFF];
	channel -> write_map = mem_offset >> 13;
	channel -> write_in = (USHORT *) &base_memory [0x0A + channel_offset];
	channel -> write_out = (USHORT *) &base_memory [0x0C + channel_offset];
	data_buffer = (USHORT *) &base_memory [0x0E + channel_offset];
	channel -> write_mask = data_buffer [0];
	sprintf (message
		   , "address = 0x%lX, Map = %d, Mask = 0x%X"
		   , (long) channel -> write_data
		   , channel -> write_map
		   , channel -> write_mask);
	TRACE_S ("digi_connect, write %s", message);

	data_buffer = (USHORT *) &base_memory [0x10 + channel_offset];
	mem_offset = (data_buffer [0] - digi_memory_seg) << 4;
	channel -> read_data = &base_memory [mem_offset & 0x1FFF];
	channel -> read_map = mem_offset >> 13;
	channel -> read_in = (USHORT *) &base_memory [0x12 + channel_offset];
	channel -> read_out = (USHORT *) &base_memory [0x14 + channel_offset];
	data_buffer = (USHORT *) &base_memory [0x16 + channel_offset];
	channel -> read_mask = data_buffer [0];
	channel -> read_overrun = (UCHAR *) &base_memory [0x59 + channel_offset];
	sprintf (message
		   , "address = 0x%lX, Map = %d, Mask = 0x%X"
		   , (long) channel -> read_data
		   , channel -> read_map
		   , channel -> read_mask);
	TRACE_S ("digi_connect, read %s", message);
	if (channel -> read_mask != 0x7FF) {
		printf ("digi_connect, read_mask = 0x%X\n", channel -> read_mask);
		ERROR_TRACE_I ("digi_connect, read_mask = 0x%X", channel -> read_mask);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}
	if (channel -> write_mask != 0x7FF) {
		printf ("digi_connect, write_mask = 0x%X\n", channel -> read_mask);
		ERROR_TRACE_I ("digi_connect, write_mask = 0x%X", channel -> read_mask);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}

	if (!set_baud_rate (channel)) {
		ERROR_TRACE ("digi_init, set_baud_rate failed");
		return LEVEL1_IO_ERROR;
	}
	channel -> write_timeout = ini_get_long (name, "WRITE_TIMEOUT", write_timeout);
	*user_handle = (LEVEL1_HANDLE) channel;
	TRACE ("digi_connect, successful");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	ERROR_TRACE_S ("digi_publish, not supported (%s)", name);
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS digi_accept (LEVEL1_HANDLE	*client_handle
						 , LEVEL1_HANDLE	server_handle)
{
	ERROR_TRACE ("digi_accept, not supported");
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS digi_write (LEVEL1_HANDLE		user_handle
					     , size_t			buffer_size
					     , void				*buffer)
{
	int				warned_once;
	USHORT			index;
	USHORT			base_index;
	UCHAR			*user_buffer;
	TIME_DEF		now;
	TIME_DEF		start_time;
	CHANNEL_DEF		*channel;
	
	TRACE_B ("digi_write", buffer_size, buffer);
	if (!user_handle) {
		ERROR_TRACE ("digi_write, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	channel = (CHANNEL_DEF *) user_handle;
	TRACE_S ("digi_write, channel = %s", channel -> name);

	set_modem_control (channel, 0xFF, 0x00);
	channel -> rts_on = TRUE;
	/****
	*	set_modem_control (channel, 0x03, 0x00);
	****/

	warned_once = FALSE;
	time_get (&start_time);
	while (buffer_size >= get_ring_size (channel)) {
		if (!warned_once) {
			ERROR_TRACE ("digi_write, transmit_overflow");
			ERROR_TRACE_I ("digi_write, write_mask = %d", channel -> write_mask + 1);
			ERROR_TRACE_I ("digi_write, write_in = %d", channel -> write_in [0]);
			ERROR_TRACE_I ("digi_write, write_out = %d", channel -> write_out [0]);
			ERROR_TRACE_I ("digi_write, buffer_size = %d", buffer_size);
			warned_once = TRUE;
		}
		time_get (&now);
		if (time_expired (&start_time, &now, channel -> write_timeout)) {
			ERROR_TRACE_L ("digi_write, timedout = %ld", channel -> write_timeout);
			return FALSE;
		}
	}
	base_index = channel -> write_in [0];
	user_buffer = (UCHAR *) buffer;
	DISABLE ();
	OUTPORT (channel -> board -> base_io_port + 1, (char) (0x80 + channel -> write_map));
	for (index = 0; index < buffer_size; ++index) {
		channel -> write_data [base_index] = user_buffer [index];
		base_index = (base_index + 1) & channel -> write_mask;
	}
	OUTPORT (channel -> board -> base_io_port + 1, 0x80);
	ENABLE ();
	channel -> write_in [0] = base_index;
	channel -> write_count += buffer_size;

	/****
	* char			message [MAX_STRING];
	*sprintf (message
	*	   , "OIX=0x%03X,OX=0x%03X,IX=0x%03X, M=0x%X,A=0x%lX,MAP=%d,Tot=%ld"
	*	   , base_index
	*	   , channel -> write_out [0]
	*	   , channel -> write_in [0]
	*	   , channel -> write_mask
	*	   , (long) &channel -> write_data [base_index]
	*	   , (int) channel -> write_map
	*	   , channel -> write_count);
	*TRACE_S ("digi_write, %s", message);
	****/
	/****
	*	set_modem_control (channel, 0xFF, 0x00);
	****/
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	CHANNEL_DEF	*channel;

	if (!user_handle) {
		ERROR_TRACE ("digi_poll, NULL handle");
		*bytes_ready = 0;
		return LEVEL1_BAD_HANDLE;
	}
	channel = (CHANNEL_DEF *) user_handle;
	if (channel -> read_overrun [0]) {
		TRACE ("digi_poll, read_overrun");
		channel -> read_out [0] = channel -> read_in [0];
		channel -> read_overrun [0] = 0;
		*bytes_ready = 0;
		return LEVEL1_OVERFLOW;
	}
	if (channel -> rts_on) {
		if (channel -> write_in [0] == channel -> write_out [0]) {
			/* set_modem_control (channel, 0x01, 0x02); */
			/* set_modem_control (channel, 0xFD, 0x02); doesn't toggle */
			/* set_modem_control (channel, 0x0D, 0xF2); RD,CTS,DSR,CD */
			set_modem_control (channel, 0x0C, 0xF3);
			channel -> rts_on = FALSE;
		}
	}
	if (channel -> read_in [0] == channel -> read_out [0]) {
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	}
	*bytes_ready = (channel -> read_in [0] - channel -> read_out [0])
				 & channel -> read_mask;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_read (LEVEL1_HANDLE	user_handle
					   , size_t			buffer_size
					   , void			*buffer
					   , size_t			*return_length)
{
	int			read_map;
	USHORT		digi_index;
	USHORT		user_index;
	UCHAR		*user_buffer;
	UCHAR		*digi_buffer;
	CHANNEL_DEF	*channel;

	if (!user_handle) {
		ERROR_TRACE ("digi_read, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	channel = (CHANNEL_DEF *) user_handle;
	user_buffer = (UCHAR *) buffer;
	if (channel -> read_overrun [0]) {
		ERROR_TRACE ("digi_read, read_overrun");
		channel -> read_out [0] = channel -> read_in [0];
		channel -> read_overrun [0] = 0;
		return LEVEL1_OVERFLOW;
	}
	user_index = 0;
	while (channel -> read_in [0] != channel -> read_out [0]) {
		digi_index = channel -> read_out [0];
		channel -> read_out [0] = (channel -> read_out [0] + 1)
								& channel -> read_mask;
		digi_buffer = &channel -> read_data [digi_index];
		if (wrong_map (channel, digi_index)) {
			TRACE ("digi_read switch read data map");
			digi_buffer = (UCHAR *) (((ULONG) digi_buffer) - 0x2000L);
			read_map = 1 + channel -> read_map;
		} else {
			read_map = channel -> read_map;
		}
		DISABLE ();
		OUTPORT (channel -> board -> base_io_port + 1, (char) (0x80 + channel -> read_map));
		user_buffer [user_index] = *digi_buffer;
		++user_index;
		OUTPORT (channel -> board -> base_io_port + 1, 0x80);
		ENABLE ();
		++channel -> read_count;
		
		/****
		*	char		message [MAX_STRING];
		* sprintf (message
		*	   , "IIX=0x%03X,OX=0x%03X,IX=0x%03X, M=0x%X,A=0x%lX,MAP=%d,C=0x%02X,Tot=%ld"
		*	   , digi_index
		*	   , channel -> read_out [0]
		*	   , channel -> read_in [0]
		*	   , channel -> read_mask
		*	   , (long) digi_buffer
		*	   , (int) read_map
		*	   , user_buffer [user_index - 1]
		*	   , channel -> read_count);
		*TRACE_S ("digi_read, %s", message);
		****/
		if (user_index >= buffer_size) {
			*return_length = user_index;
			return LEVEL1_NORMAL;
		}
	}
	*return_length = user_index;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_clear (LEVEL1_HANDLE user_handle)
{
	CHANNEL_DEF	*channel;

	TRACE ("digi_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("digi_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	channel = (CHANNEL_DEF *) user_handle;
	channel -> read_out [0] = channel -> read_in [0];
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_shutdown (LEVEL1_HANDLE user_handle)
{
	CHANNEL_DEF	*channel;

	TRACE ("digi_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("digi_shutdown, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	channel = (CHANNEL_DEF *) user_handle;
	delete_channel (channel);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS digi_exit (void)
{
	CHANNEL_DEF		*channel;
	BOARD_DEF		*board;

	TRACE ("digi_exit, entry point");
	while (queue_next (&root_channel)) {
		TRACE ("digi_exit, close next channel");
		channel = (CHANNEL_DEF *) queue_next (&root_channel);
		level1_shutdown ((LEVEL1_HANDLE) channel);
	}
	while (queue_next (&root_board)) {
		TRACE ("digi_exit, delete next board");
		board = (BOARD_DEF *) queue_next (&root_board);
		/****
		*	int_remove (board);
		****/
		delete_board (board);
	}
	TRACE_I ("digi_exit, int_count1 = %d", int_count1);
	TRACE_I ("digi_exit, int_count12 = %d", int_count2);
	TRACE_I ("digi_exit, int_channel = %d", int_channel);
	TRACE_I ("digi_exit, int_event = %d", int_event);
	TRACE_I ("digi_exit, int_status = %d", int_status);
	TRACE_I ("digi_exit, int_prior = %d", int_prior);
	return LEVEL1_NORMAL;
}

/****
*
****/

static CHANNEL_DEF *add_channel (char *channel_name)
{
	CHANNEL_DEF	*channel;

	channel = (CHANNEL_DEF *) malloc (sizeof (CHANNEL_DEF));
	if (!channel) {
		ERROR_TRACE_I ("add_channel, malloc size = %d", sizeof (CHANNEL_DEF));
		return NULL;
	}
	strcpy (channel -> name, channel_name);
	channel -> write_count = 0;
	channel -> read_count = 0;
	queue_append (&root_channel, &channel -> link);
	return channel;
}

/****
*
****/

static void delete_channel (CHANNEL_DEF *channel)
{
	TRACE ("delete_channel, entry point");
	queue_remove (&root_channel, &channel -> link);
	free (channel);
}

/****
*
****/

static BOARD_DEF *add_board (char *board_name)
{
	BOARD_DEF	*board;
	ULONG		flat_address;

	board = (BOARD_DEF *) malloc (sizeof (BOARD_DEF));
	if (!board) {
		ERROR_TRACE_I ("add_board, malloc size = %d", sizeof (BOARD_DEF));
		return (BOARD_DEF *) NULL;
	}
	flat_address = ini_get_hex (board_name, "BASE_MEMORY", 0xFFFFFFFL);
	if (flat_address == 0xFFFFFFFFL) {
		ERROR_TRACE_S ("add_board, no BASE_MEMORY = (%s)", board_name);
		free (board);
		return (BOARD_DEF *) NULL;
	}
	board -> base_memory = (UCHAR *) ADDRESS_ALIAS (flat_address);
	board -> base_io_port = (short) ini_get_hex (board_name
											   , "BASE_IO_PORT"
											   , 0xFFFFFFFFL);
	if (board -> base_io_port == 0xFFFFFFFFL) {
		ERROR_TRACE_S ("add_board, no BASE_IO_PORT = (%s)", board_name);
		free (board);
		return (BOARD_DEF *) NULL;
	}
	board -> base_irq = (short) ini_get_int (board_name, "BASE_IRQ", 0);
	if (!board -> base_irq) {
		ERROR_TRACE_S ("add_board, no BASE_IRQ = (%s)", board_name);
		free (board);
		return (BOARD_DEF *) NULL;
	}
	if (!hardware_init (board)) {
		ERROR_TRACE_S ("add_board, hardware_init failed = (%s)", board_name);
		free (board);
		return (BOARD_DEF *) NULL;
	}
	strcpy (board -> name, board_name);
	queue_append (&root_board, &board -> link);
	return board;
}

/****
*
****/

static void delete_board (BOARD_DEF *board)
{
	TRACE ("delete_board, entry point");
	queue_remove (&root_board, &board -> link);
	free (board);
}

static BOARD_DEF *find_board (char *name)
{
	BOARD_DEF		*board;

	board = (BOARD_DEF *) queue_next (&root_board);
	while (board) {
		if (case_compare (board -> name, name)) {
			return board;
		}
		board = (BOARD_DEF *) queue_next (&board -> link);
	}
	return (BOARD_DEF *) NULL;
}

static int wrong_map (CHANNEL_DEF *channel, USHORT index)
{
	int			base_map;
	int			test_map;
	UCHAR		*calc_buffer;
	char		message [MAX_STRING];

	calc_buffer = &channel -> read_data [0];
	base_map = (int) ((((ULONG) calc_buffer) & 0xE000L) >> 13);
	calc_buffer = &channel -> read_data [index];
	test_map = (int) ((((ULONG) calc_buffer) & 0xE000L) >> 13);
	if (test_map != base_map) {
		sprintf (message
			   , "Index=0x%X, baseM=%d, baseA=0x%lX, testM=%d, testA=0x%lX"
			   , index
			   , base_map
			   , (long) &channel -> read_data [0]
			   , test_map
			   , (long) &channel -> read_data [index]);
		TRACE_S ("wrong_map data = %s", message);
		return TRUE;
	}
	return FALSE;
}

static int hardware_init (BOARD_DEF *board)
{
	short		base_io_port;
	long		start_tick;
	UCHAR		status;
	USHORT		command;
	UCHAR		*base_memory;

	TRACE ("hardware_init, entry point");
	base_io_port = board -> base_io_port;
	base_memory = board -> base_memory;
	OUTPORT (base_io_port, DIGI_RESET);
	delay (100);
	status = (UCHAR) INPORT (base_io_port);
	if ((0x01 & status) == 0x01) {
		ERROR_TRACE_I ("hardware_init, Digiboard Failure #1(%d)", status);
		return FALSE;
	}
	OUTPORT (base_io_port, DIGI_HOLD_RESET);
	delay (100);
	status = (UCHAR) INPORT (base_io_port);
	if (0x44 != status) {
		ERROR_TRACE_I ("hardware_init, Digiboard Failure #2(%d)", status);
		return FALSE;
	}
	command = 0x10;
	switch (board -> base_irq) {
	case 3:
		board -> irq_vector = VectorIRQ3;
		board -> irq_mask = MaskIRQ3;
		command |= 0x01;
		break;
	case 5:
		board -> irq_vector = VectorIRQ5;
		board -> irq_mask = MaskIRQ5;
		command |= 0x02;
		break;
	case 7:
		board -> irq_vector = VectorIRQ7;
		board -> irq_mask = MaskIRQ7;
		command |= 0x03;
		break;
	case 10:
		board -> irq_vector = VectorIRQ10;
		board -> irq_mask = MaskIRQ10;
		command |= 0x04;
		break;
	case 11:
		board -> irq_vector = VectorIRQ11;
		board -> irq_mask = MaskIRQ11;
		command |= 0x05;
		break;
	case 12:
		board -> irq_vector = VectorIRQ12;
		board -> irq_mask = MaskIRQ12;
		command |= 0x06;
		break;
	case 15:
		board -> irq_vector = VectorIRQ15;
		board -> irq_mask = MaskIRQ15;
		command |= 0x07;
		break;
	default:
		ERROR_TRACE_I ("hardware_init, unsupported IRQ = %d", board -> base_irq);
		return FALSE;
	}
	command |= (USHORT) (0x0FFE & ADDRESS_MASK (base_memory));
	TRACE_I ("hardware_init, memory init command= 0x%X", command);
	OUTPORT (base_io_port + 2, (UCHAR) (command & 0xFF));
	OUTPORT (base_io_port + 3, (UCHAR) (command >> 8));
	OUTPORT (base_io_port, DIGI_RESET);
	start_tick = GetTickCount ();
	while (0x44 != (UCHAR) INPORT (base_io_port)) {
		if (start_tick + 1000 < GetTickCount ()) {
			ERROR_TRACE_I ("hardware_init, Digiboard Failure #3(%d)", status);
			return FALSE;
		}
	}
	OUTPORT (base_io_port, DIGI_ENABLE_MEMORY);
	delay (100);
	if (!test_memory (base_memory)) {
		ERROR_TRACE ("hardware_init, Digiboard memory failure");
		return FALSE;
	}
	if (!load_bios (base_io_port, base_memory)) {
		ERROR_TRACE ("hardware_init, load_bios failed");
		return FALSE;
	}
	if (!load_fep (base_io_port, base_memory)) {
		ERROR_TRACE ("hardware_init, load_fep failed");
		return FALSE;
	}
	if (!map_buffers (board)) {
		ERROR_TRACE ("hardware_init, map_buffers failed");
		return FALSE;
	}
	/****
	*	int_install (board);
	*	if (!test_interrupts (base_memory)) {
	*		ERROR_TRACE ("hardware_init, test_interrupt failed");
	*		int_remove (board);
	*		return FALSE;
	*	}
	****/
	TRACE ("Digiboard Initialized successfully");
	return TRUE;
}

/****
*	Not used
*
*	static void				int_install (BOARD_DEF *);
*	static void int_install (BOARD_DEF *board)
*	{
*		USHORT		*mem_command;
*	
*		int_count1 = 0;
*		int_count2 = 0;
*		board -> old_value8259 = (UCHAR) (INPORT (I8259OCW1A) & 0xFF);
*		board -> old_irq_int = GET_VECTOR (board -> irq_vector);
*		SET_VECTOR (board -> irq_vector, int_8250);
*		OUTPORT (I8259OCW1A, board -> old_value8259 & board -> irq_mask);
*		mem_command = (USHORT *) &board -> base_memory [DIGI_INTERRUPTS];
*		mem_command [0] = 1;
*	}
****/

/****
*	Not used
*
*	static void				int_remove (BOARD_DEF *);
*	static void int_remove (BOARD_DEF *board)
*	{
*		USHORT		*mem_command;
*	
*		return;
*		mem_command = (USHORT *) &board -> base_memory [DIGI_INTERRUPTS];
*		mem_command [0] = 0;
*		DISABLE ();
*		OUTPORT (I8259OCW1A, board -> old_value8259);
*		SET_VECTOR (board -> irq_vector, board -> old_irq_int);
*		ENABLE ();
*	}
*****/

static void delay (int milliseconds)
{
	long		start_tick;

	start_tick = GetTickCount ();
	while (start_tick + milliseconds >= GetTickCount ()) {
		taskDelay (1);
	}
}

static int test_memory (UCHAR *base_memory)
{
	long		index;

	for (index = 0; index < DIGI_MAP_SIZE; ++index) {
		base_memory [index] = (UCHAR) (index & 0xFF);
	}
	for (index = 0; index < DIGI_MAP_SIZE; ++index) {
		if (base_memory [index] != (UCHAR) (index & 0xFF)) {
			TRACE_L ("test_memory failed at %ld", index);
			return FALSE;
		}
	}
	return TRUE;
}

static int load_bios (short io_port, UCHAR *base_memory)
{
	long		index;
	long		start_tick;

	base_memory [DIGI_BIOS_STATUS] = 0;
	OUTPORT (io_port + 1, 0x87);
	for (index = 0; index < DIGI_BIOS_SIZE; ++index) {
		base_memory [index + DIGI_BIOS_AREA] = digi_bios_code [index];
	}
	for (index = 0; index < DIGI_BIOS_SIZE; ++index) {
		if (base_memory [index + DIGI_BIOS_AREA] != digi_bios_code [index]) {
			TRACE_L ("load_bios failed at %ld", index);
			return FALSE;
		}
	}
	OUTPORT (io_port + 1, 0x80);
	OUTPORT (io_port, DIGI_RELEASE_RESET);
	start_tick = GetTickCount ();
	while (('G'!= base_memory [DIGI_BIOS_STATUS]) 
	&&     ('D'!= base_memory [DIGI_BIOS_STATUS + 1])) {
		if (start_tick + 10000 < GetTickCount ()) {
			TRACE  ("Unable to load the digiboard bios");
			return FALSE;
		}
	}
	return TRUE;
}

static int load_fep (short io_port, UCHAR *base_memory)
{
	long		index;
	long		start_tick;
	USHORT		command_index;
	USHORT		*in_command;
	USHORT		*fep_command;
	USHORT		*word_command;

	OUTPORT (io_port + 1, 0x81);
	for (index = 0; index < DIGI_FEP_SIZE; ++index) {
		base_memory [index] = digi_fep_code [index];
	}
	for (index = 0; index < DIGI_FEP_SIZE; ++index) {
		if (base_memory [index] != digi_fep_code [index]) {
			ERROR_TRACE_L ("load_fep failed at %ld", index);
			return FALSE;
		}
	}
	OUTPORT (io_port + 1, 0x80);
	fep_command = (USHORT *) &base_memory [DIGI_BIOS_MOVE];
	fep_command [0] = 0x0002;
	fep_command [1] = 0xF200;
	fep_command [2] = 0x0000;
	fep_command [3] = 0x0200;
	fep_command [4] = 0x0000;
	fep_command [5] = 0x2000;
	OUTPORT (io_port, 0x08);
	OUTPORT (io_port, 0x00);
	start_tick = GetTickCount ();
	while (fep_command [0]) {
		if (start_tick + 1000 < GetTickCount ()) {
			ERROR_TRACE  ("Unable to load the fep driver");
			return FALSE;
		}
	}

	fep_command [0] = 0x0001;
	fep_command [1] = 0x0200;
	fep_command [2] = 0x0004;
	base_memory [DIGI_FEP_ACK + 0] = 0;
	base_memory [DIGI_FEP_ACK + 1] = 0;
	OUTPORT (io_port, 0x08);
	OUTPORT (io_port, 0x00);
	start_tick = GetTickCount ();
	while (('O'!= base_memory [DIGI_FEP_ACK]) 
	&&     ('S'!= base_memory [DIGI_FEP_ACK + 1])) {
		if (start_tick + 5000 < GetTickCount ()) {
			ERROR_TRACE ("Unable to execute the fep driver");
			return FALSE;
		}
	}
	in_command = (USHORT *) &base_memory [DIGI_COMMAND_IN];
	command_index = in_command [0];
	TRACE_I ("load_fep, command_index = %d", command_index);
	base_memory [DIGI_BUFFER_SPACE + command_index + 0] = 0x92;
	base_memory [DIGI_BUFFER_SPACE + command_index + 1] = 0x00;
	word_command = (USHORT *) &base_memory [DIGI_BUFFER_SPACE + command_index + 2];
	word_command [0] = 0x8000;
	in_command [0] = (command_index + 4) & 0x3FF;
	return TRUE;
}

static int map_buffers (BOARD_DEF *board)
{
	USHORT		command_index;
	UCHAR		*byte_command;
	USHORT		*word_command;
	USHORT		*in_command;

	in_command = (USHORT *) &board -> base_memory [DIGI_COMMAND_IN];
	command_index = in_command [0];
	TRACE_I ("map_buffers, command_index = %d", command_index);
	byte_command = (UCHAR *) &board -> base_memory [DIGI_BUFFER_SPACE + command_index];
	byte_command [0] = 0xF2;
	byte_command [1] = 0;
	word_command = (USHORT *) &byte_command [2];
	word_command [0] = 32;
	in_command [0] = (command_index + 4) & 0x3FF;
	delay (100);
	return TRUE;
}

/****
*	Not used
*
*	static int				test_interrupts (UCHAR *);
*	static int test_interrupts (UCHAR *base_memory)
*	{
*		int			old_count;
*		long		start_tick;
*		USHORT		command_index;
*		USHORT		*mem_command;
*		USHORT		*event_in;
*		USHORT		*event_out;
*		USHORT		*in_command;
*	
*		old_count = int_count1;
*		in_command = (USHORT *) &base_memory [DIGI_COMMAND_IN];
*		command_index = in_command [0];
*		TRACE_I ("test_interrupts, command_index = %d", command_index);
*		mem_command = (USHORT *) &base_memory [DIGI_BUFFER_SPACE + command_index];
*		mem_command [0] = 0x00FF;
*		mem_command [1] = 0x0000;
*		in_command [0] = (command_index + 4) & 0x3FF;
*		start_tick = GetTickCount ();
*		while (int_count1 == old_count) {
*			if (start_tick + 1000 < GetTickCount ()) {
*				TRACE  ("Failed to get the interrupt");
*				return FALSE;
*			}
*		}
*		event_in = (USHORT *) &base_memory [DIGI_EVENT_IN];
*		event_out = (USHORT *) &base_memory [DIGI_EVENT_OUT];
*		event_out [0] = event_in [0];
*		return TRUE;
*	}
****/

static int set_baud_rate (CHANNEL_DEF *channel)
{
	USHORT		command_index;
	USHORT		*in_command;
	UCHAR		*base_memory;
	USHORT		*word_command;
	USHORT		command;
	long		default_baud_rate;
	long		local_baud_rate;

	default_baud_rate = ini_get_long (LEVEL1_GROUP, "BAUD_RATE", 9600L);
	local_baud_rate = ini_get_long (channel -> name, "BAUD_RATE", default_baud_rate);
	TRACE_L ("set_baud_rate, new baud_rate = %ld", local_baud_rate);
	command = DATA_BITS_8 + PARITY_NONE + STOP_BITS_1;
	switch (local_baud_rate) {
	case 1200L:
		command |= (0x09 + BAUD_RATES_SLOW);
		break;
	case 1800L:
		command |= (0x0A + BAUD_RATES_SLOW);
		break;
	case 2400L:
		command |= (0x0B + BAUD_RATES_FAST);
		break;
	case 4800L:
		command |= (0x0C + BAUD_RATES_FAST);
		break;
	case 14400L:
		command |= (0x04 + BAUD_RATES_FAST);
		break;
	case 19200L:
		command |= (0x0E + BAUD_RATES_FAST);
		break;
	case 28800L:
		command |= (0x0A + BAUD_RATES_FAST);
		break;
	case 38400L:
		command |= (0x0F + BAUD_RATES_FAST);
		break;
	case 57600L:
		command |= (0x05 + BAUD_RATES_FAST);
		break;
	case 115200L:
		command |= (0x03 + BAUD_RATES_FAST);
		break;
	case 230000L:
		command |= (0x06 + BAUD_RATES_FAST);
		break;
	case 9600L:
	default:
		command |= (0x0D + BAUD_RATES_FAST);
	}
	base_memory = channel -> board -> base_memory;
	in_command = (USHORT *) &base_memory [DIGI_COMMAND_IN];
	command_index = in_command [0];
	TRACE_I ("set_baud_rate, channel_number = %d", channel -> channel_number);
	base_memory [DIGI_BUFFER_SPACE + command_index + 0] = 0xF5;
	base_memory [DIGI_BUFFER_SPACE + command_index + 1] = channel -> channel_number;
	word_command = (USHORT *) &base_memory [DIGI_BUFFER_SPACE + command_index + 2];
	word_command [0] = command;
	in_command [0] = (command_index + 4) & 0x3FF;

	set_modem_control (channel, 0xFF, 0x00);
	channel -> rts_on = FALSE;
	/* set_modem_control (channel, 0x01, 0x02); */
	set_flow_control (channel);
	return TRUE;
}

static void set_modem_control (CHANNEL_DEF *channel
							 , UCHAR		set_value
							 , UCHAR		clear_value)
{
	USHORT		command_index;
	USHORT		*in_command;
	UCHAR		*base_memory;

	TRACE ("set_modem_control, entry point");
	base_memory = channel -> board -> base_memory;
	in_command = (USHORT *) &base_memory [DIGI_COMMAND_IN];
	command_index = in_command [0];
	base_memory [DIGI_BUFFER_SPACE + command_index + 0] = 0xE9;
	base_memory [DIGI_BUFFER_SPACE + command_index + 1] = channel -> channel_number;
	base_memory [DIGI_BUFFER_SPACE + command_index + 2] = set_value;
	base_memory [DIGI_BUFFER_SPACE + command_index + 3] = clear_value;
	in_command [0] = (command_index + 4) & 0x3FF;
	TRACE_I ("set_modem_control, set_control = 0x%X", set_value);
	TRACE_I ("set_modem_control, clear_control = 0x%X", clear_value);
}

static void set_flow_control (CHANNEL_DEF *channel)
{
	USHORT		command_index;
	USHORT		*in_command;
	UCHAR		*base_memory;

	TRACE ("set_flow_control, entry point");
	base_memory = channel -> board -> base_memory;
	in_command = (USHORT *) &base_memory [DIGI_COMMAND_IN];
	command_index = in_command [0];
	base_memory [DIGI_BUFFER_SPACE + command_index + 0] = 0xF4;
	base_memory [DIGI_BUFFER_SPACE + command_index + 1] = channel -> channel_number;
	base_memory [DIGI_BUFFER_SPACE + command_index + 2] = 0x00;
	base_memory [DIGI_BUFFER_SPACE + command_index + 3] = 0xB3;
	in_command [0] = (command_index + 4) & 0x3FF;
}

static size_t get_ring_size (CHANNEL_DEF *channel)
{
	if (channel -> write_in [0] < channel -> write_out [0]) {
		return (channel -> write_out [0] - channel -> write_in [0]) - 1;
	} else {
		return (((channel -> write_mask + 1) - channel -> write_in [0])
				+ channel -> write_out [0]) - 1;
	}
}

/****
*	Not used
*	static int 				dump (UCHAR *);
*	static int dump (UCHAR *base_memory)
*	{
*		USHORT		*word_buffer;
*		ULONG		*long_buffer;
*	
*		TRACE_I ("0C00 = 0x%X", base_memory [DIGI_BIOS_STATUS]);
*		TRACE_I ("0C01 = 0x%X", base_memory [DIGI_BIOS_STATUS + 1]);
*		word_buffer = (USHORT *) &base_memory [DIGI_BIOS_STATUS + 2];
*		TRACE_I ("0C02 = 0x%X", word_buffer [0]);
*		TRACE_I ("0C22 = 0x%X", base_memory [DIGI_NUM_CHANNELS]);
*		TRACE_I ("0C30 = 0x%X", base_memory [0xC30]);
*		TRACE_I ("0C31 = 0x%X", base_memory [0xC31]);
*		long_buffer = (ULONG *) &base_memory [0xC34];
*		TRACE_L ("0C34 = 0x%X", long_buffer[0]);
*		word_buffer = (USHORT *) &base_memory [0x0C38];
*		TRACE_I ("0C38 = 0x%X", word_buffer[0]);
*		long_buffer = (ULONG *) &base_memory [0xC3A];
*		TRACE_L ("0C3A = 0x%X", long_buffer[0]);
*		word_buffer = (USHORT *) &base_memory [0x0D10];
*		TRACE_I ("0D10 = 0x%X", word_buffer[0]);
*		TRACE_I ("0D12 = 0x%X", word_buffer[1]);
*		TRACE_I ("0D14 = 0x%X", word_buffer[2]);
*		TRACE_I ("0D16 = 0x%X", word_buffer[3]);
*		TRACE_I ("0D18 = 0x%X", word_buffer[4]);
*		TRACE_I ("0D1A = 0x%X", word_buffer[5]);
*		TRACE_I ("0D1C = 0x%X", word_buffer[6]);
*		TRACE_I ("0D1E = 0x%X", word_buffer[7]);
*		TRACE_I ("0D20 = 0x%X", word_buffer[8]);
*		word_buffer = (USHORT *) &base_memory [0x1000];
*		TRACE_I ("1008 = 0x%X", word_buffer[4]);
*		TRACE_I ("100A = 0x%X", word_buffer[5]);
*		TRACE_I ("100C = 0x%X", word_buffer[6]);
*		TRACE_I ("100E = 0x%X", word_buffer[7]);
*		TRACE_I ("1010 = 0x%X", word_buffer[8]);
*		TRACE_I ("1012 = 0x%X", word_buffer[9]);
*		TRACE_I ("1014 = 0x%X", word_buffer[10]);
*		TRACE_I ("1016 = 0x%X", word_buffer[11]);
*		TRACE_I ("1020 = 0x%X", word_buffer[16]);
*		TRACE_I ("104E = 0x%X", base_memory [0x104E]);
*		TRACE_I ("1050 = 0x%X", base_memory [0x1050]);
*		TRACE_I ("1051 = 0x%X", base_memory [0x1051]);
*		TRACE_I ("1054 = 0x%X", base_memory [0x1054]);
*		return TRUE;
*	}
****/

/****
*
*	interrupt handler for 8250 through the 8259
*
****/

/****
*	Not used
*
*	static void INTERRUPT	int_8250 (void);
*	void INTERRUPT int_8250 (void)
*	{
*		USHORT		index;
*		USHORT		mask;
*		USHORT		*event_mask;
*		USHORT		*event_in;
*		USHORT		*event_out;
*		UCHAR		*base_memory;
*		BOARD_DEF	*board;
*	
*		++int_count1;
*		board = (BOARD_DEF *) queue_next (&root_board);
*		while (board) {
*			base_memory = (UCHAR *) board -> base_memory;
*			event_in = (USHORT *) &base_memory [DIGI_EVENT_IN];
*			event_out = (USHORT *) &base_memory [DIGI_EVENT_OUT];
*			event_mask = (USHORT *) &base_memory [DIGI_EVENT_MASK];
*			if (event_out [0] != event_in [0]) {
*				++int_count2;
*				mask = event_mask [0];
*				index = event_out [0];
*				int_channel = base_memory [DIGI_EVENTS + index + 0];
*				int_event = base_memory [DIGI_EVENTS + index + 1];
*				int_status = base_memory [DIGI_EVENTS + index + 2];
*				int_prior = base_memory [DIGI_EVENTS + index + 3];
*				event_out [0] = (index + 4) & mask;
*			}
*			board = (BOARD_DEF *) queue_next (&board -> link);
*		}
*		OUTPORT (I8259OCW2A, I8259EOI);
*		OUTPORT (I8259OCW2, I8259EOI);
*	}
****/

