/****
*
*	File name :		libvxw.c
*	Author :		Steve Johnson
*	Creation date :	September 10, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points:
*
*		GetTickInit
*		TimeGetFileMod
*		TimeCompare
*		time_get
*		file_find_first
*		file_find_next
*		kbhit
*		getch
*		vx_enable
*		vx_disable
*		vx_getvector
*		vx_setvector
*		vx_inport
*		vx_outport
*		GetTickCount
*		GetLocalTime
*		vx_open
*		vx_fileno
*		vx_close
*		vx_write
*		vx_seek
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <sstimer.h>

static int		lock_key;
static long		base_time;

/****
*
****/

void GetTickInit (void)
{
	struct timespec	time_spec;

	clock_gettime (CLOCK_REALTIME, &time_spec);
	base_time = time_spec . tv_sec;
	TRACE_L ("GetTickinit, base_time = 0x%lX", base_time);
}

int GetTickUnits (LARGE_INTEGER *frequency)
{
	frequency->u.LowPart = 1000;
	frequency->u.HighPart = 0;
	return TRUE;
}

int GetTickCounter (LARGE_INTEGER *counter)
{
	counter->u.LowPart = GetTickCount ();
	counter->u.HighPart = 0;
	return TRUE;
}

/****
*
****/

int TimeGetFileMod (char *file_name, SYSTEMTIME *system_time)
{
	struct stat		file_stat;
	struct tm		*sys_time;
	char			message [MAX_STRING];

	if (stat (file_name, &file_stat)) {
		ERROR_TRACE_S ("TimeGetFileMod, stat failed on (%s)", file_name);
		ERROR_TRACE_I ("TimeGetFileMod, stat errno = (%d)", errno);
		return FALSE;
	}
	sys_time = localtime (&file_stat . st_mtime);
	sprintf (message
		   , "%d/%d/%d %d:%d:.%d"
		   , sys_time -> tm_mday
		   , sys_time -> tm_mon
		   , sys_time -> tm_year
		   , sys_time -> tm_hour
		   , sys_time -> tm_min
		   , sys_time -> tm_sec);
	TRACE_S ("TimeGetFileMod, file time stamp = %s", message);
	system_time -> wMilliseconds = 0;
	system_time -> wSecond = sys_time -> tm_sec;
	system_time -> wMinute = sys_time -> tm_min;
	system_time -> wHour = sys_time -> tm_hour;
	system_time -> wDay = sys_time -> tm_mday;
	system_time -> wMonth = sys_time -> tm_mon;
	system_time -> wYear = 1900 + sys_time -> tm_year;
	return TRUE;
}

int TimeCompare (SYSTEMTIME *first, SYSTEMTIME *second)
{
	TIME_DEF		first_time;
	TIME_DEF		second_time;
	DATE_TIME_DEF	first_date_time;
	DATE_TIME_DEF	second_date_time;

	first_date_time . year = first -> wYear;
	first_date_time . month = first -> wMonth;
	first_date_time . day = first -> wDay;
	first_date_time . hour = first -> wHour;
	first_date_time . minute = first -> wMinute;
	first_date_time . second = first -> wSecond;
	first_date_time . milliseconds = 0;
	time_date_to_time (&first_date_time, &first_time);
	second_date_time . year = second -> wYear;
	second_date_time . month = second -> wMonth;
	second_date_time . day = second -> wDay;
	second_date_time . hour = second -> wHour;
	second_date_time . minute = second -> wMinute;
	second_date_time . second = second -> wSecond;
	second_date_time . milliseconds = 0;
	time_date_to_time (&second_date_time, &second_time);
	return time_compare (&first_time, &second_time);
}

void time_get (TIME_DEF *time)
{
	int				status;
	struct timespec	time_spec;

	status = clock_gettime (CLOCK_REALTIME, &time_spec);
	if (!status) {
		time -> tv_sec = time_spec . tv_sec - base_time;
		time -> tv_usec = time_spec . tv_nsec / 1000;
	} else {
		time -> tv_sec = 0;
		time -> tv_usec = 0;
	}
}

/****
*
*	In vxworks we cannot sit there and poll standard input repeatedly
*	other tasks may need it, so we just return FALSE always
*
****/

int kbhit (void)
{
	return FALSE;
}

int getch (void)
{
	return (int) getchar ();
}

void vx_enable (void)
{
	intUnlock (lock_key);
}

void vx_disable (void)
{
	lock_key = intLock ();
}

void *vx_getvector (int int_vector)
{
	FUNCPTR		*vx_vector;

	vx_vector = (FUNCPTR *) INUM_TO_IVEC (int_vector);
	return intVecGet (vx_vector);
}

void vx_setvector (int int_vector, void (*new_vector))
{
	FUNCPTR		int_function;
	FUNCPTR		*vx_vector;

	vx_vector = (FUNCPTR *) INUM_TO_IVEC (int_vector);
	int_function = intHandlerCreate (new_vector, 0);
	intVecSet (vx_vector, int_function);
}

void vx_outport (USHORT io_port, UCHAR data)
{
	sysOutByte (io_port, data);
}

UCHAR vx_inport (USHORT io_port)
{
	return (UCHAR) sysInByte (io_port);
}

int xkbhit (void)
{
	int				count;
	struct timeval	time_value;
	struct timeval	*time_pointer;
	FD_BIT_MASK		write_mask;
	FD_BIT_MASK		except_mask;
	FD_BIT_MASK		read_mask;

	time_value . tv_sec = 0;
	time_value . tv_usec = 0;
	time_pointer = &time_value;
	FD_ZERO (&write_mask);
	FD_ZERO (&except_mask);
	FD_ZERO (&read_mask);
	FD_SET (0, &read_mask);
	count = select (FD_SETSIZE
				  , &read_mask
				  , &write_mask
				  , &except_mask
				  , time_pointer);
	if (count == -1) {
		return FALSE;
	} else if (count == 0) {
		return FALSE;
	}
	return TRUE;
}

/****
*
*	really returns milliseconds not ticks
*
****/

long GetTickCount (void)
{
	int				status;
	int				current_second;
	struct timespec	time_spec;

	status = clock_gettime (CLOCK_REALTIME, &time_spec);
	if (status) {
		ERROR_TRACE_I ("GetTickCount, clock_gettime = %d", errno);
		return 0;
	}
	current_second = time_spec . tv_sec - base_time;
	return ((current_second * 1000) + (time_spec . tv_nsec / 1000000));
}

void GetLocalTime (SYSTEMTIME *system_time)
{
	time_t			seconds;
	struct timespec	time_spec;
	struct tm		*sys_time;

	time (&seconds);
	sys_time = localtime (&seconds);
	clock_gettime (CLOCK_REALTIME, &time_spec);
	system_time -> wMilliseconds = time_spec . tv_nsec / 1000;
	system_time -> wSecond = sys_time -> tm_sec;
	system_time -> wMinute = sys_time -> tm_min;
	system_time -> wHour = sys_time -> tm_hour;
	system_time -> wDay = sys_time -> tm_mday;
	system_time -> wMonth = sys_time -> tm_mon;
	system_time -> wYear = 1900 + sys_time -> tm_year;
}

void SetLocalTime (SYSTEMTIME *system_time)
{
	ERROR_TRACE ("SetLocalTime, not implemented");
}

/****
*
****/

TFILE vx_open (char *file_name, int sharing, int type)
{
	return (TFILE) open (file_name, sharing, type);
}

TFILE vx_fileno (FILE *file_descr)
{
	return (TFILE) fileno (file_descr);
}

void vx_close (TFILE file)
{
	close (file);
}

void vx_write (TFILE file, void *buffer, size_t length)
{
	write (file, buffer, length);
}

void vx_read (TFILE file, void *buffer, size_t length)
{
	read (file, buffer, length);
}

void vx_seek (TFILE file, long position)
{
	lseek (file, position, SEEK_SET);
}

