/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*		THE TIS4000 WORKSTATION MONITOR			*/
/*								*/
/*	This program reads the tile /tis4000/tasks/taskList.	*/
/*	Upon parsing the list, we start all of the programs	*/
/*	that comprise the system. We then wait for any one	*/
/*	of them to die or to receive a signal.			*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef SOLARIS
#include <sys/time.h>
#include <sys/resource.h>
#endif

/****************************************************************/
/*								*/
/*			FUNCTION PROTOTYPES			*/
/*								*/
/****************************************************************/

void	sigHupHandler();
void	sigIntHandler();
char	*nextToken();

/****************************************************************/
/*								*/
/*			GLOBAL DATA				*/
/*								*/
/****************************************************************/

int		sigHupActive = 0;
int		sigIntActive = 0;

struct child
    {
    int			pid;
    char		*progName;
    char		*scriptName;
    char		**argList;
    int			numCrashes;
    time_t		firstStarted;
    time_t		lastCrashed;
    struct child	*nextChild;
    struct child	*prevChild;
    } *firstChild = 0;
struct child	*lastChild;

/****************************************************************/
/*								*/
/*			MAIN PROGRAM				*/
/*								*/
/****************************************************************/

main()
{
    int			i;
    int			childStatus;
    int			childPid;
    int			signalNumber;
    char		scriptLine[256];
    struct stat		statBuff;
    struct child	*currChild;
    int			timeNow;

    /*
    The very first thing to do is to fork and do all of our
    work in the child. This is just in case the user forgets
    to run us in the background.
    */

    i = fork();
    if (i < 0)
	{
	perror("Can't create subprocesses");
	exit(1);
	}
    if (i)	/* Eventually exit when we are the parent. For now, keep running so we can debug! */
	/*
	exit(0)
	*/
	;
    /*
    Eventually, the entire else clause disappears. For now, trash the child
    */
    else
	exit(0);
	

    /*
    Put our pid into a file so that other people can easily send
    us signals. We process HUP and INT. All others have a default
    action for the time being.
    */

    i = open("/tis4000/tasks/tis4000Mon.pid", O_RDWR | O_CREAT, 0640);
    sprintf(scriptLine, "%d\n", getpid());
    write(i, scriptLine, strlen(scriptLine));

    /*
    The next thing to do is to close all the file descriptors
    that we can and then open the three standard ones to a
    harmless I/O device, namely /dev/null. By doing this, we 
    save all of our child processes the trouble of doing this 
    because they will inherit all of our open files.
    */

/*
Do this when we think we won't need printing any more during development.
#ifdef SOLARIS
    {
    struct rlimit rl;
    if (getrlimit (RLIMIT_NOFILE, &rl))
        {

        for (i = 19; i >= 0; --i)
	    close(i);
        }
    else
        {
        for (i = rl.rlim_cur - 1; i >= 0; --i)
	    close(i);
        }
    }
#else    
    for (i = getdtablesize() - 1; i >= 0; --i)
	close(i);
#endif

    i = open("/dev/null", O_RDWR);
    dup(i);
    dup(i);
*/

    /*
    We use SIGINT as our signal to stop the system. We use SIGHUP
    as our signal to restart the system. Install handlers so that
    we can catch those signals.
    */

    signal(SIGHUP, sigHupHandler);
    signal(SIGINT, sigIntHandler);

    /*
    Now wait for something to happen.
    */

    for (;;)
	{
	childPid = wait(&childStatus);

	/*
	If the pid returned by the system is -1, then we have no remaining
	children. This can be because we haven't started the system yet,
	or because we are stopping or restarting it.
	*/

	if (childPid == -1)
	    {
	    if (sigIntActive)
		{
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: System shutdown complete");
		exit(0);
		}
	    else if (sigHupActive)
		{
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Shutdown complete. Restart proceeding");
		sigHupActive = 0;
		}
	    sysStartup();
	    continue;
	    }
	
	/*
	If we are stopping the system for either permanently or for a restart,
	do nothing except wait for all child processes to die.
	*/

	if (sigIntActive || sigHupActive)
	    continue;

	/*
	A child process died. Print an informative message. First find the PID in our
	list of children that we've launched.
	*/

	currChild = firstChild;
	while (currChild)
	    {

	    if (currChild -> pid != childPid)
		{
		currChild = currChild -> nextChild;
		continue;
		}
	    if ((childStatus & 0xff) == 0)
		{

		/*
		Child exited on his own. Report his exit code.
		*/

		signalNumber = 0;
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: %s exited, code %d",
			currChild -> progName, (childStatus & 0x0000ff00) >> 8);
		}
	    else if ((childStatus & 0x7f) != 0x7f)
		{

		/*
		Child died from a signal. Signal number in 7 lsb. 0x80 bit tells
		us if there was a dump.
		*/

		signalNumber = childStatus & 0x7f;
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: %s terminated due to signal %d, %s",
			currChild -> progName, signalNumber, 
			((childStatus & 0x80) ? "DUMP" : "NODUMP"));
		}
	    else
		{
		signalNumber = (childStatus & 0x0000ff00) >> 8;
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: %s stopped due to signal %d",
			currChild -> progName, (childStatus & 0x0000ff00) >> 8);
		}
	    
	    /*
	    Record the statistics for this crash.
	    */

	    currChild -> numCrashes++;

	    /*
	    If there is a script for this program, invoke it.
	    */

	    if (currChild -> scriptName)
		{
		if (stat(currChild -> scriptName, &statBuff) == -1)
		    {
		    syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Script %s was deleted!",
			    currChild -> scriptName);
		    currChild -> pid = 0;
		    break;
		    }
		timeNow = time(NULL);
		if (currChild -> lastCrashed == 0)
		    currChild -> lastCrashed = timeNow;
		sprintf(scriptLine, "%s %s %d %d %d %d", currChild -> scriptName, 
			currChild -> progName, currChild -> numCrashes, 
			timeNow - currChild -> firstStarted,
			timeNow - currChild -> lastCrashed, signalNumber);
		system(scriptLine);
		}

	    /*
	    Record the time of this crash as the last one and then restart the program.
	    */

	    currChild -> lastCrashed = time(NULL);
	    syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: restarting program %s",
		    currChild -> progName);
	    if (stat(currChild -> progName, &statBuff) == -1)
		{
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Program %s was deleted!",
			currChild -> progName);
		currChild -> pid = 0;
		break;
		}
	    if ((currChild -> pid = fork()) == 0)
		execv(currChild -> progName, currChild -> argList);
	    break;
	    }
	}
}

/****************************************************************/
/*								*/
/*			SIGHUP HANDLER				*/
/*								*/
/****************************************************************/

void sigHupHandler()
{
    sigHupActive = 1;
    sysShutdown();
}

/****************************************************************/
/*								*/
/*			SIGINT HANDLER				*/
/*								*/
/****************************************************************/

void sigIntHandler()
{
    sigIntActive = 1;
    sysShutdown();
}

/****************************************************************/
/*								*/
/*			START UP THE SYSTEM			*/
/*								*/
/****************************************************************/

sysStartup()
{
    FILE		*fd;
    struct child	*currChild;
    struct stat		statBuff;
    int			i;
    int			currLineNum;
    int			nbytes;
    int			numArgs;
    char		buff[256];
    char		*currString;
    char		*nextString;
    char		*tempString;

    /*
    Start by freeing up the old data structures. This list is
    empty the first time that the system is started.
    */

    currChild = firstChild;
    firstChild = 0;
    lastChild = 0;
    while (currChild)
	{
	free(currChild -> progName);
	free(currChild -> scriptName);
	for (i = 0; currChild -> argList[i]; i++)
	    free(currChild -> argList[i]);
	free(currChild -> argList);
	if (!(currChild -> nextChild))
	    {
	    free(currChild);
	    break;
	    }
	if (currChild -> prevChild)
	    free(currChild -> prevChild);
	currChild = currChild -> nextChild;
	}

    /*
    Read the file. We do this every time because it may have changed
    while the system was running.
    */

    fd = fopen("/tis4000/tasks/taskList", "r");
    if (fd == NULL)
	{
	syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: can't open /tis4000/tasks/taskList: %m");
	exit (0);
	}
    currLineNum = 0;

    /*
    Process each line in the file.
    */

    while (fgets(buff, sizeof(buff), fd) != NULL)
	{
	currLineNum++;

	/*
	Ignore blank lines (first byte null, return, or new line)
	Also, allow comments by first byte == '#'
	*/

	if (buff[0] == '\n')
	    continue;
	if (buff[0] == '\r')
	    continue;
	if (buff[0] == 0)
	    continue;
	if (buff[0] == '#')
	    continue;

	/*
	Isolate the first token and compare it for our key words.
	*/

	currString = buff;
	nextString = nextToken(currString);
	if (!strcmp(currString, "pause") || !strcmp(currString, "PAUSE"))
	    sleep(1);

	else if (!strcmp(currString, "null") || !strcmp(currString, "NULL"))
	    {

	    /*
	    This is an immediate shell script invocation. Run along the string
	    and replace all token delimeters with blanks and then invoke the
	    script.
	    */

	    currString = nextString;
	    while (*nextString)
		{
		if (*nextString == ':')
		    *nextString = ' ';
		nextString++;
		}
	    system(currString);
	    }

	else
	    {

	    /*
	    This is a program to run. Allocate a structure member and store the 
	    program name.
	    */

	    if (!firstChild)
		currChild = firstChild = (struct child *) calloc(1, sizeof(struct child));
	    else
		{
		currChild -> nextChild = (struct child *) calloc(1, sizeof(struct child));
		currChild -> nextChild -> prevChild = currChild;
		currChild = currChild -> nextChild;
		}
	    lastChild = currChild;
	    nbytes = strlen(currString);
	    if (nbytes)
		{
		currChild -> progName = (char *) malloc(nbytes + 1);
	        strcpy(currChild -> progName, currString);
		}
	    else
		{
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Null program name in line %d",
			currLineNum);
		exit(0);
		}

	    /*
	    Make sure that the program exists. Bail out if it doesn't
	    */

	    if (stat(currString, &statBuff) == -1)
		{
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Program %s not found (line %d)",
			currString, currLineNum);
		exit(0);
		}

	    /*
	    Now get the script. It's OK if there isn't one.
	    */

	    currString = nextString;		/* Point to script name */
	    nextString = nextToken(currString);
	    nbytes = strlen(currString);
	    if (nbytes)
		{
		currChild -> scriptName = (char *) malloc(nbytes + 1);
		strcpy(currChild -> scriptName, currString);

		/*
		Given that there is a script, make sure that it exists.
		*/

		if (stat(currString, &statBuff) == -1)
		    {
		    syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: Script %s not found (line %d)",
			    currString, currLineNum);
		    exit(0);
		    }
		}
	
	    /*
	    Now pick up the arguments. We can have arguments even if there isn't a script.
	    Figure out how many arguments there are. It is equal to the number of delimeters
	    in the line + 2. This may be one extra if the user put a colon after the last
	    argument, but it won't make much of a difference. We're only temporarily wasting
	    four (8 on the AXP) bytes of memory. We still need two pointers even if there
	    are no arguments to make calls to execv().
	    */

	    currString = nextString;
	    numArgs = 0;
	    tempString = currString;
	    while (*tempString)
		{
		if (*tempString == ':')
		    numArgs++;
		tempString++;
		}
	    numArgs += 3;
	    currChild -> argList = (char **) calloc(numArgs, sizeof(char *));
	    currChild -> argList[0] = (char *) malloc(strlen(currChild -> progName) + 1);
	    strcpy(currChild -> argList[0], currChild -> progName);
	    for (i = 1; i <= numArgs; i++, currString = nextString)
		{
		if (*currString == 0)
		    break;
		nextString = nextToken(currString);
		nbytes = strlen(currString);
		currChild -> argList[i] = (char *) malloc(nbytes + 1);
		strcpy(currChild -> argList[i], currString);
		}

	    /*
	    Now that we've got the arguments, we can run the program.
	    */

	    syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: starting %s", currChild -> progName);
	    if ((currChild -> pid = fork()) == 0)
		{
		execv(currChild -> progName, currChild -> argList);
		syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: can't start %s: %m",
			currChild -> progName);
		exit(0);
		}

	    /*
	    Record time that we started.
	    */

	    currChild -> firstStarted = time(NULL);
	    }
	}
    fclose(fd);
    syslog(LOG_ALERT | LOG_DAEMON, "tis4000Mon: System startup complete.");
}

/****************************************************************/
/*								*/
/*			SHUT THE SYSTEM DOWN			*/
/*								*/
/****************************************************************/

sysShutdown()
{
    struct child	*currChild;

    /*
    Shut down the system by sending an INT signal to each process
    in the reverse order from which they were started. This is
    why the list of child structures has backward links.
    */

    currChild = lastChild;
    while (currChild)
	{
	if (currChild -> pid)
	    kill(currChild -> pid, SIGINT);
	currChild = currChild -> prevChild;
	}
}

/****************************************************************/
/*								*/
/*	SMASH DELIMETER, RETURN POINTER TO THE NEXT BYTE	*/
/*								*/
/*	If we see a new line, set that to a null. When we	*/
/*	encounter a null, return the addr of the null.		*/
/*								*/
/****************************************************************/

char *nextToken(string)
char	*string;
{
    for (;;)
	{
	if (*string == '\n')
	    *string = 0;
	else if (*string == '\r')
	    *string = 0;
	if (*string == 0)
	    return (string);
	if (*string == ':')
	    {
	    *string = 0;
	    return (string + 1);
	    }
	string++;
	}
}
